/**
 * \file msg.h <gre/msg.h>
 * \brief GRE::MSG class definitions
 *
 * \if NODOC
 * $Id: msg.h_v 1.19 2011/10/31 16:16:35 mju Exp $
 *
 * $Log: msg.h_v $
 * Revision 1.19  2011/10/31 16:16:35  mju
 * Add action_SetDrawStereo.
 *
 * Revision 1.18  2011/10/05 15:22:52  mju
 * Fewer changeViewPoint3d actions.
 *
 * Revision 1.17  2011/10/05 13:34:36  mju
 * Add changeViewPoint3d actions.
 *
 * Revision 1.16  2011/09/06 16:40:35  mju
 * Add action_Reclassify.
 *
 * Revision 1.15  2010/07/15 16:10:53  mju
 * Add action_UserRefresh.
 *
 * Revision 1.14  2008/02/01 15:02:40  mju
 * Remove obsolete rightButtonPressed msg.
 *
 * Revision 1.13  2006/09/29 13:41:19  mju
 * Add action_ChangeDispParm.
 *
 * Revision 1.12  2006/09/21 21:31:25  mju
 * Move msg_viewexpose to own header.
 *
 * Revision 1.11  2005/12/13 14:46:24  mju
 * Add action_createChildrenCancel.
 *
 * Revision 1.10  2005/10/19 19:50:09  mju
 * Don't use old gre names.
 *
 * Revision 1.9  2005/10/19 17:32:41  mju
 * Remove unused typedef.
 *
 * Revision 1.8  2005/10/14 17:32:34  mju
 * GRE namespace.
 *
 * Revision 1.7  2004/01/15 15:09:55  mju
 * Add action_seqmask and action_none.
 *
 * Revision 1.6  2004/01/13 17:40:52  mju
 * Add unsequenced version of all action types so can index through table.
 *
 * Revision 1.5  2003/09/15 13:48:59  fileserver!dwilliss
 * Doxygen
 *
 * Revision 1.4  2003/07/30 15:32:09  mju
 * Ignore private sections.
 *
 * Revision 1.3  2003/06/24 20:23:02  mju
 * nc.
 *
 * Revision 1.2  2003/06/24 16:24:28  mju
 * Move VIEWEXPOSE msg here.
 *
 * Revision 1.1  2003/06/16 19:39:52  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_GRE_MSG_H
#define	INC_GRE_MSG_H

#ifndef	INC_GRE_BASE_H
	#include <gre/base.h>
#endif

#ifndef	INC_MI32_ELEMTYPE_H
	#include <mi32/elemtype.h>
#endif

namespace GRE {
//===================================================================================================================

//! Message container.
class MSG {
	public:

		enum ACTION {
			ACTION_SeqNone =				0x0000,	//!< No sequence associated with action
			ACTION_SeqBegin =				0x1000,	//!< Beginning of action, will be followed (eventually) by same action with SeqEnd or SeqCancel
			ACTION_SeqEnd =				0x2000,	//!< Ending of action
			ACTION_SeqContinue =			0x4000,	//!< Continuing action
			ACTION_SeqCancel =			0x8000,	//!< Action cancelled
			ACTION_SeqMask =				0xF000,	//!< Mask for sequence

			ACTION_None =								(0),								//!< No action (never actually sent to msg handler)
			ACTION_Create =							(1|ACTION_SeqNone),			//!< Object has just been created
			ACTION_Destroy =							(2|ACTION_SeqNone),			//!< Object is being destroyed
			ACTION_Selected =							(3|ACTION_SeqNone),			//!< Object has been selected in list
			ACTION_Deselected =						(4|ACTION_SeqNone),			//!< Object has been deselected in list
			ACTION_Change =							(5|ACTION_SeqNone),			//!< Unsequenced change made to object
			ACTION_ChangeBegin =						(5|ACTION_SeqBegin),			//!< Begin non-specific change to object
			ACTION_ChangeEnd =						(5|ACTION_SeqEnd),			//!< End non-specific change to object
			ACTION_ChangeCancel =					(5|ACTION_SeqCancel),		//!< Cancel non-specific change to object
			ACTION_Expose =							(6|ACTION_SeqNone),			//!< Exposure event occurred for view DrawingArea
			ACTION_Draw =								(7|ACTION_SeqNone),			//!< Unsequenced draw operation
			ACTION_DrawBegin =						(7|ACTION_SeqBegin),			//!< Begin draw operation
			ACTION_DrawEnd =							(7|ACTION_SeqEnd),			//!< End draw operation
			ACTION_DrawCancel =						(7|ACTION_SeqCancel),		//!< Cancel draw operation
			ACTION_ChangeOrder =						(8|ACTION_SeqNone),			//!< Change to order of object in list
			ACTION_ChangeOrderBegin =				(8|ACTION_SeqBegin),			//!< Begin change to order of object in list
			ACTION_ChangeOrderEnd =					(8|ACTION_SeqEnd),			//!< End change to order of object in list
			ACTION_CreateChildren =					(9|ACTION_SeqNone),			//!< Creating children of object
			ACTION_CreateChildrenBegin =			(9|ACTION_SeqBegin),			//!< Begin creating children of object
			ACTION_CreateChildrenEnd =				(9|ACTION_SeqEnd),			//!< End creating children of object
			ACTION_CreateChildrenCancel =			(9|ACTION_SeqCancel),		//!< Cancel creating children of object
			ACTION_DestroyChildren =				(10|ACTION_SeqNone),			//!< Destroying children of object
			ACTION_DestroyChildrenBegin =			(10|ACTION_SeqBegin),		//!< Begin destroying children of object
			ACTION_DestroyChildrenEnd =			(10|ACTION_SeqEnd),			//!< End destroying children of object
			ACTION_ChangeObject =					(11|ACTION_SeqNone),			//!< Selected object(s) for layer changed
			ACTION_ChangeName =						(12|ACTION_SeqNone),			//!< Object name changed
			ACTION_Resized =							(13|ACTION_SeqNone),			//!< Object (view) has been resized
			ACTION_ElementSelect =					(14|ACTION_SeqNone),			//!< Change to element highlight set
			ACTION_ElementSelectBegin =			(14|ACTION_SeqBegin),		//!< Begin change to element highlight set
			ACTION_ElementSelectEnd =				(14|ACTION_SeqEnd),			//!< End change to element highlight set
			ACTION_ChangeActiveElem =				(15|ACTION_SeqNone),			//!< Active element changed (for layer)
			ACTION_ChangeVisible =					(17|ACTION_SeqNone),			//!< Unsequenced visibility change for object
			ACTION_ChangeVisibleBegin =			(17|ACTION_SeqBegin),		//!< Begin visibility change for object
			ACTION_ChangeVisibleEnd =				(17|ACTION_SeqEnd),			//!< End visibility change for object
			ACTION_CursorEnter =						(18|ACTION_SeqNone),			//!< Cursor has entered view
			ACTION_CursorLeave =						(19|ACTION_SeqNone),			//!< Cursor has moved out of view
			ACTION_CreateView =						(20|ACTION_SeqNone),			//!< Created view of object
			ACTION_Load =								(21|ACTION_SeqNone),			//!< Load layout/group
			ACTION_LoadBegin =						(21|ACTION_SeqBegin),		//!< Begin load layout/group
			ACTION_LoadEnd =							(21|ACTION_SeqEnd),			//!< End load layout/group
			ACTION_Restore =							(22|ACTION_SeqNone),			//!< Restoring view from memory image
			ACTION_RestoreBegin =					(22|ACTION_SeqBegin),		//!< Begin restoring view from memory image
			ACTION_RestoreContinue =				(22|ACTION_SeqContinue),	//!< Continue restoring view from memory image
			ACTION_RestoreEnd =						(22|ACTION_SeqEnd),			//!< Begin restoring view from memory image
			ACTION_CloseRequest =					(23|ACTION_SeqNone),			//!< User has requested "Close" on object
			ACTION_DataTipShown =					(24|ACTION_SeqNone),			//!< DataTip has been shown in view
			ACTION_DataTipHidden =					(25|ACTION_SeqNone),			//!< DataTip has been hidden in view
			ACTION_DestroyView =						(26|ACTION_SeqNone),			//!< View of object about to be destroyed
			ACTION_DeactivateTool =					(27|ACTION_SeqNone),			//!< Active tool for view about to be deactivated
			ACTION_ActivateTool =					(28|ACTION_SeqNone),			//!< Active tool for view about to be activated
			ACTION_ElementSelectAll =				(29|ACTION_SeqNone),			//!< Selecting/deselecting all elements
			ACTION_ElementSelectAllBegin =		(29|ACTION_SeqBegin),		//!< Begin selecting/deselecting all elements
			ACTION_ElementSelectAllEnd =			(29|ACTION_SeqEnd),			//!< End selecting/deselecting all elements
			ACTION_StyleObjEdited =					(30|ACTION_SeqNone),			//!< Style object for layer has been edited
			ACTION_OverlayUpdate =					(31|ACTION_SeqNone),			//!< Overlay update needed
			ACTION_Edited =							(32|ACTION_SeqNone),			//!< Object (layer) has been "edited" (for TNTedit)
			ACTION_ExtractedFrom =					(33|ACTION_SeqNone),			//!< Object (layer) has been extracted from
			ACTION_SetComponentSelectable =		(34|ACTION_SeqNone),			//!< Component has been made selectable/nonselectable
			ACTION_SetShowInternalTables =		(35|ACTION_SeqNone),			//!< Change made to "show internal tables" setting
			ACTION_SetShowScaleRanges =			(36|ACTION_SeqNone),			//!< Change made to "show scale ranges" setting
			ACTION_DrawMatte =		 				(37|ACTION_SeqNone),			//!< Drawing Matte
			ACTION_DrawMatteBegin = 				(37|ACTION_SeqBegin),		//!< Begin Drawing Matte
			ACTION_DrawMatteEnd = 					(37|ACTION_SeqEnd),			//!< End Drawing Matte
			ACTION_ChangeLegend =					(38|ACTION_SeqNone),			//!< Begin legend change for object
			ACTION_ChangeLegendBegin =				(38|ACTION_SeqBegin),		//!< Begin legend change for object
			ACTION_ChangeLegendEnd =				(38|ACTION_SeqEnd),			//!< End legend change for object
			ACTION_ChangeDispParm =					(39|ACTION_SeqNone),			//!< DispParm has been changed
			ACTION_UserRefresh =						(40|ACTION_SeqNone),			//!< User pressed "Refresh" button on view.
			ACTION_Reclassify =						(41|ACTION_SeqNone),			//!< Reclassify operation was done on layer
			ACTION_ChangeViewPoint3d =				(42|ACTION_SeqNone),			//!< 3D viewpoint change
			ACTION_ChangeViewPoint3dDragging =	(42|ACTION_SeqContinue),	//!< 3D viewpoint changing by user dragging, expect to get ACTION_ChangeViewPoint3d when done
			ACTION_SetDrawStereo =					(43|ACTION_SeqNone),			//!< View setting for whether to draw stereo toggled
			ACTION_MAXTYPE =							43									//!< Maximum action type
			};

		MSG (
			ACTION action,
			GRE::OBJECT *object,
			GRE::VIEW *view = 0,
			ELEMTYPE elemtype = ELEMTYPE_Empty
			) :
			m_action(action),
			m_object(object),
			m_view(view),
			m_elemtype(elemtype)
			{ }

		MSG (
			ACTION action,
			GRE::OBJECT *object,
			ELEMTYPE elemtype
			) :
			m_action(action),
			m_object(object),
			m_view(0),
			m_elemtype(elemtype)
			{ }

		//! Get action which caused this message to be sent.
		//! @return action type
		ACTION GetAction (
			) const { return (m_action); }

		//! Get element type if applicable.
		//! @return element type
		ELEMTYPE GetElemType (
			) const { return (m_elemtype); }

		//! Get object to which message applies.
		//! @return object pointer
		GRE::OBJECT* GetObject (
			) const { return (m_object); }

		//! Get view message originated from if any.
		//! @return view pointer or 0 if not applicable
		GRE::VIEW* GetView (
			) const { return (m_view); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		ACTION m_action;
		GRE::OBJECT *m_object;
		GRE::VIEW *m_view;
		ELEMTYPE m_elemtype;

		#endif // GENERATING_DOXYGEN_OUTPUT
	};


//===================================================================================================================
}	// End namespace GRE

#endif	// INC_GRE_MSG_H
