/**
 * \file mi32/base64.h
 * \brief A Base-64 encoder/decoder class.
 *
 * \if NODOC
 * $Id: base64.h_v 1.3 2011/10/21 20:44:04 mju Exp $
 *
 * $Log: base64.h_v $
 * Revision 1.3  2011/10/21 20:44:04  mju
 * Remove redundant hdr fix dox.
 *
 * Revision 1.1  2003/07/11 16:39:14  dwilliss
 * Initial revision
 * \endif
**/

#ifndef INC_MI32_BASE64_H
#define INC_MI32_BASE64_H

#ifndef INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
class FILEPATH;
class GENERICRW;
#endif // GENERATING_DOXYGEN_OUTPUT

#ifdef MISYSTEMDLL
	#define CLASSLIBEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSLIBEXPORT MI_DLLCLASSIMPORT
#endif


class CLASSLIBEXPORT BASE64_ENCODER {
	public:
		BASE64_ENCODER ();
		virtual ~BASE64_ENCODER ();

		//! Encode a file.
		//! Just opens the file and calles the GENERICRW version of the method.
		ERRVALUE Encode (
			const FILEPATH& filepath
			);

		//! Reads the file in chunks and calls the buffer version of the method
		ERRVALUE Encode (
			GENERICRW& stream
			);

		//! Encode byte stream.  Can be called multiple times if the input to
		//! be encoded is being read in chunks.
		ERRVALUE Encode (
			const UINT8* buffer,
			int NumBytes
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<char> m_WorkBuffer;
		// Copy constructor and assignment operator are private and unimplemented
		// Doesn't make sense to copy one.
		BASE64_ENCODER(const BASE64_ENCODER&);
		BASE64_ENCODER& operator=(const BASE64_ENCODER&);
		#endif // GENERATING_DOXYGEN_OUTPUT

		// Derived class must implement this.
		// Will be called as many times as needed as chunks of data are encoded
		// Do NOT append a newline to the buffer.  It will already have newlines
		// and may contain more than one line.
		// Should return 0 or an error code < 0
		virtual ERRVALUE v_Write (
			const char* buffer
			) = 0;

		// Derived class may override this method if it wants to show status or
		// implement cancel checking.  This method should return EUserCancel to
		// cancel the operation, 0 to continue.  Will not be called by the 
		// version of Encode which just takes a buffer because the total size
		// is not known.  
		virtual ERRVALUE v_ShowStatus (
			double PercentDone
			);

	};


class CLASSLIBEXPORT BASE64_DECODER {
	public:
		BASE64_DECODER();
		virtual ~BASE64_DECODER();

		//! Decode a file
		ERRVALUE Decode (
			const FILEPATH& filepath
			);

		//! Decode character stream
		ERRVALUE Decode (
			const char* buffer
			);

		//! Decode data from a GENERICRW stream.
		//! If size is supplied by the caller, it will only read and decode that
		//! many bytes of input.  This is useful for decoding base-64 data which
		//! is embeded in a larger stream of data.
		ERRVALUE Decode (
			GENERICRW& stream,
			INT64 size = -1
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		// Copy constructor and assignment operator are private and unimplemented
		// Doesn't make sense to copy one.
		BASE64_DECODER(const BASE64_DECODER&);
		BASE64_DECODER& operator=(const BASE64_DECODER&);
		#endif // GENERATING_DOXYGEN_OUTPUT

		// Derived class must implemnt this.
		// Will be called as many times as needed as chunks of data are decoded
		// Should return 0 or an error code < 0
		virtual ERRVALUE v_Write (
			const UINT8* buffer,
			int numbytes
			) = 0;

		// Derived class may override this method if it wants to show status or
		// implement cancel checking.  This method should return EUserCancel to
		// cancel the operation, 0 to continue.
		virtual ERRVALUE v_ShowStatus (
			double PercentDone
			);

	};

#undef CLASSLIBEXPORT

#endif
