/*****************************************************************************
 *
 * \file mi32/htmlelement.h
 * \brief HTML::ELEMENT class
 *
 * $Log: htmlelement.h_v $
 * Revision 1.6  2011/10/25 21:30:02  mju
 * Remove redundant hdrs.
 *
 * Revision 1.5  2011/04/21 22:15:29  bcolombe
 * Add STATE_Link
 *
 * Revision 1.3  2009/03/18 19:00:04  ldyas
 * added errvalue responses to do_layouts.
 *
 * Revision 1.2  2007/08/29 21:05:17  dwilliss
 * Added support for keeping track of counters
 *
 * Revision 1.1  2007/08/29 15:42:35  dwilliss
 * Initial revision
 **/

#ifndef INC_MI32_HTMLELEMENT_H
#define INC_MI32_HTMLELEMENT_H

#ifndef INC_MI32_RECT_H
#include <mi32/rect.h>
#endif

#ifndef INC_MI32_HTMLSTYLE_H
#include <mi32/htmlstyle.h>
#endif

#ifndef INC_MAP
#define INC_MAP
#include <map>
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
class MISTRING;
class STRUTF8;
class XMLNODE;
class REGION2D;
class TRANS2D_AFFINE;
namespace MGD {
	class DEVICE;
	class CONTEXT;
	}
namespace HTML {
	class HTMLDOC;
	class STYLESHEET;
	struct INLINE_POS;
	}
#endif

namespace HTML {

	class ELEMENT {
		public:

			enum STATE {
				STATE_Normal = 0,
				STATE_Hover,
				STATE_Active,
				STATE_Link,
				STATE_NUMSTATES
				};

			ELEMENT (
				HTMLDOC* doc,
				XMLNODE* node
				);

			virtual ~ELEMENT();

			//! Compute the size and position of the element.
			ERRVALUE DoLayout (
				MGD::CONTEXT* gc,
				double MaxWidth,			//!< Maximum width in pixels
				DPOINT2D& pt,				//!< Starting position relative to parent
				REGION2D& Container,		//!< Container to layout into. Child should remove its own area from the container, and can grow it as necessary.
				DRECT2D& Extents			//! Extents of the parent element.  Child can grow if necessary
				);

			//! After figuring out the layout of children, compute the
			//! absolute position of the element
			void DoLayoutPass2 (
				MGD::CONTEXT* gc,
				DPOINT2D& AbsPos
				);

			//! Draw the element and its children
			void Draw (
				MGD::CONTEXT* gc,
				const DPOINT2D& pos,
				const DRECT2D* clip = 0
				);


			void AddToRegion (
				REGION2D& region,
				const DPOINT2D& pos,
				bool bOnlyIfDirty
				);

			//! Find a child element at a given point.
			//! The point is relative to the upper-left corner of this element.
			ELEMENT* FindChild (
				const DPOINT2D& pt
				);

			//! Get the extents of the element.
			//! These are computed when the element is drawn
			const DRECT2D& GetExtents (
				) const { return m_Extents; }

			double GetIntrinsicHeight (
				) const { return v_GetIntrinsicHeight(); }

			double GetIntrinsicWidth (
				) const { return v_GetIntrinsicWidth(); }

			//! Get the height of the element in pixels
			//! Height includes margins, padding and border.
			double GetHeight (
				) const { return m_Height; }

			XMLNODE* GetNode (
				) const { return m_node; }

			//! Get the width of the element in pixels
			//! Width includes margins, padding and border.
			double GetWidth (
				) const { return m_Width; }

			// Get the position of the element relative to the upper-left corner
			// of the parent element (in pixels)
			const DPOINT2D& GetPos (
				) const { return m_Pos; }

			// Return the extents of the element as a region
			void GetRegion (
				REGION2D& region,
				bool bRootRelative = true
				);

			const STYLE* GetStyle (
				STYLE::PSEUDO pseudo = STYLE::PSEUDO_CurrentState
				) const;

			const ELEMENT* GetParent() const;
			ELEMENT* GetParent();
			const ELEMENT* GetFirstChild() const;
			ELEMENT* GetFirstChild();

			const ELEMENT* GetNext() const;
			ELEMENT* GetNext();

			//! Return true if the element has no content
			//! This will ignore white-space.
			bool IsEmpty (
				) const;

			//! Determine if the element should be "hidden"
			bool IsHidden (
				) const { return v_IsHidden(); }

			//! Set the style for the element.  
			//! This will AddRef() the style.
			void SetStyle (
				STYLE* style
				);

			void SetState (
				STATE state
				);

#ifndef GENERATING_DOXYGEN_OUTPUT
			// These methods are used by the layout process and need to be
			// public so derived classes can call them on children, but
			// the outside world should not see them
			
			//! Get the parent document
			const HTMLDOC* GetDoc (
				) const { return m_Doc; }

			//! Set the width of the element in pixel.
			//! Heigut includes margins, padding and border.
			void SetHeight (
				double height
				) { m_Height = height; }

			//! Set the width of the element in pixel.
			//! Width includes margins, padding and border.
			void SetWidth (
				double width
				) { m_Width = width; }

			void SetPos (
				const DPOINT2D& pt
				) { m_Pos = pt; }

			// Used to shrink the width to eliminate unnecessary
			// whitespace on the right of the documnet
			void ShrinkWidth (
				double WidthDelta
				);

			void ShrinkToNaturalWidth (
				double LayoutWidth	// the width we used on the main DoLayout
				);

			// Mark the element as needing to be redrawn
			// All children will be marked as well
			void SetNeedRedraw (
				);

			// Get the value of a specific counter.
			// If no such counter exists, it will add it and initialize it to 1
			INT32 GetCounter (
				const MISTRING& name
				);

			void IncrementCounter (
				const MISTRING& name
				);

			void ResetCounter (
				const MISTRING& name
				);
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
		protected:
#ifdef DEBUG
			// The following two are to simplify debugging so that we don't have 
			// to expand things 3 levels deep just to see which element we're 
			// looking at
			const char* m_tag;			// set by ctor to m_node->_xmlNode->name
			const MIUNICODE* m_content;	// Points to the text of the node if it's a text node
#endif
			double m_Height;
			double m_Width;
			STATE m_State;

			const MGD::DEVICE* LoadImage (
				const MISTRING& src
				);

			// Take the value of a tag attribute and convert it to the CSS
			// equivilant, applying it to the elements's style
			void AttributeToCSS (
				const char* cssname,
				const STRUTF8& value
				);
			
			// should return true if the element contains a given point.
			// The point is relative to the upper-left of the parent element
			bool ContainsPoint (
				const DPOINT2D& pt
				) const { return v_ContainsPoint(pt); }


			// Make sure the container region is at least Height.  Expands it
			// vertically if not.  Note: If it does have to expand the container,
			// it may make it taller than needed to avoid having to expand it again
			// later.
			//
			ERRVALUE ExpandContainer (
				REGION2D& Container,
				MGD::CONTEXT* gc,
				double Width,
				double Height
				);

			//! Fill a rectangle according to the style for the element
			//! The rectangle is in absolute screen coordinates
			void FillBackground (
				MGD::CONTEXT* context,
				const DRECT2D& rect
				);

			//! Intersect the DesiredRect with the Region and return the
			//! resulting available space.  Note that if the result of the 
			//! intersection is non-rectangular, this will return the largest
			//! rectangle that will fit in the space, not just the extents
			//! of the intersection.
			//! If bWidestPossible is true, the returned rectangle will be the
			//! widest possible rectangle of the requested height at the
			//! requested starting position.
			//! Note, if the DesiredRect extends below the Region's current
			//! extents, the Region will be expanded to accomidate it
			ERRVALUE GetAvailableRect (
				REGION2D& Region,
				MGD::CONTEXT* gc,
				double MaxWidth,			// Width of Region if we need to expand it
				const DRECT2D& DesiredRect,
				DRECT2D& AvailableRect,
				bool bWidestPossible = true
				);

			void GetInlinePos (
				SIMPLE_ARRAY<INLINE_POS*>& positions
				) { v_GetInlinePos(positions); }

			double GetRotAngle (
				) const;

			const TRANS2D_AFFINE& GetTransAffine (
				) const;

			void SetState (
				STATE state,
				bool bPropagateToParent,
				bool bPropagateToChildren
				);

			// should return true if the element contains a given point.
			// The point is relative to the upper-left of the parent element
			virtual bool v_ContainsPoint (
				const DPOINT2D& pt
				) const;

			virtual ERRVALUE v_DoLayout (
				MGD::CONTEXT* gc,
				double MaxWidth,			//!< Maximum width in pixels
				DPOINT2D& pt,				//!< Starting position relative to parent, should update to where next sibling should go
				REGION2D& Container,		//!< Container to layout into. Child should remove its own area from the container, and can grow it as necessary.
				DRECT2D& Extents			//! Extents of the parent element.  Child can grow if necessary
				);

			// Should draw just this element.  Child elements will be drawn
			// by the caller
			virtual void v_Draw (
				MGD::CONTEXT* gc,
				const DPOINT2D& pos		// Absolute position to draw at
				);

			//! Determine if the element should be "hidden"
			//! This is mainly for tables with the empty-cells: hidden style.  
			//! Rather than coding special cases in the Draw code, Draw will
			//! call this virtual method.
			virtual bool v_IsHidden (
				) const;


		private:
			friend class HTMLDOC;

			XMLNODE* m_node;
			HTMLDOC* m_Doc;	// Pointer to parent document
			mutable STYLE* m_Styles[STATE_NUMSTATES][STYLE::PSEUDO_NUMPSEUDO];
			DPOINT2D m_Pos;
			DPOINT2D m_ChildOffset;	// Precomputed distance from Pos to where our children start
			DRECT2D m_Extents;	// Extents of this element in absolute coordinates
			bool m_bNeedRedraw;
			bool m_bHasBackground;	// Set by DoLayoutPass2
			INT32 m_ElemNum;			// For sorting top-down

			typedef std::map<MISTRING, INT32> COUNTERMAP;
			COUNTERMAP m_Counters;

			static int CompareZOrder (ELEMENT**, ELEMENT**, void*);
#endif

			virtual void v_GetInlinePos (
				SIMPLE_ARRAY<INLINE_POS*>& positions
				);

			virtual double v_GetIntrinsicHeight (
				) const;

			virtual double v_GetIntrinsicWidth (
				) const;

		};
	
	} // end of HTML namespace


#endif
