/******************************************************************************
 * \file mi32/mg2rastt.h
 * \brief MGD::DEVICE_RASTER_TRANS_MASK definitions.
 *
 * \if NODOC
 * $Log: mg2rastt.h_v $
 * Revision 1.15  2012/02/20 18:40:55  mju
 * Nest context from base.
 *
 * Revision 1.14  2011/11/02 14:03:42  mju
 * Remove redundant hdr.
 *
 * Revision 1.13  2011/10/26 17:58:22  mju
 * Frame buffer support.
 *
 * Revision 1.12  2011/10/25 17:16:24  mju
 * Derive from device_mem32rgba so properly handles transparency blending.
 *
 * Revision 1.11  2009/08/31 15:14:32  mju
 * Use getObjItem.
 *
 * Revision 1.10  2006/04/14 21:16:10  mju
 * Redesign to use rvc classes and mem32 base.
 *
 * Revision 1.9  2003/12/15 23:37:29  dwilliss
 * Diverging for 7.0
 * Made changes necessary for graphical data tips
 *
 * Revision 1.1  2001/07/11 17:33:50  dwilliss
 * Initial revision
 * \endif
 */

#ifndef INC_MI32_MG2RASTT_H
#define INC_MI32_MG2RASTT_H

#ifndef INC_MI32_MG2MEM32RGBA_H
#include <mi32/mg2mem32rgba.h>
#endif

#ifndef INC_RVC_OPACITYMASK_H
#include <rvc/opacitymask.h>
#endif

namespace MGD {

//=====================================================================================================================

//! Device for rendering to 24-bit RVC raster with corresponding opacity mask object.
//! Raster must have CELLTYPE_RedGreenBlue_24 or BlueGreenRed_24.
//! Device will have COLORSPACE_R8_G8_B8_A8 or B8_G8_R8_A8 respectively.
//! If the size is <= 1024x1024, both will be held in memory for speed
class DEVICE_RASTER_TRANS_MASK: public DEVICE_MEM32RGBA {

	public:

		//! Default constructor.
		//! After constructing, you must call Create() or Open().
		DEVICE_RASTER_TRANS_MASK ();

		//! Destructor
		virtual ~DEVICE_RASTER_TRANS_MASK ();

		// (Note: copy constructor private and unimplemented.  Can't copy one.)

		//! Create raster and corresponding opacity mask to draw into.
		ERRVALUE Create (
			RVC::OBJECT& ParentObject,
			const RVC::DESCRIPTOR& Descriptor,
			RVC::RASTER::MAKEPARMS& MakeParms,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Create raster and corresponding opacity mask to draw into.
		ERRVALUE Create (
			RVC::OBJITEM& ObjItem,
			RVC::RASTER::MAKEPARMS& MakeParms,
			const RVC::CREDENTIALS* Credentials = 0
			);

		//! Get OBJITEM for raster.
		ERRVALUE GetObjItemRaster (
			RVC::OBJITEM& ObjItem
			) const { return (m_ObjRaster.GetObjItem(ObjItem)); }

		//! Get OBJITEM for mask.
		ERRVALUE GetObjItemMask (
			RVC::OBJITEM& ObjItem
			) const { return (m_ObjMask.GetObjItem(ObjItem)); }

		//! Open existing raster to draw into.
		//! Will create opacity mask subobject if none exists.
		ERRVALUE Open (
			const RVC::OBJITEM& ObjItem,
			const RVC::CREDENTIALS* Credentials = 0
			);

	protected:

		class CONTEXT : public DEVICE_MEM32RGBA::CONTEXT {
			public:

				virtual ~CONTEXT ();

			protected:

				CONTEXT (
					DEVICE_RASTER_TRANS_MASK *pDevice
					);

			private:
		   	#ifndef GENERATING_DOXYGEN_OUTPUT
				CONTEXT(const CONTEXT&);

				friend class MGD::DEVICE_RASTER_TRANS_MASK;
				#endif // GENERATING_DOXYGEN_OUTPUT
			
			};

		//! In the Raster case, this will actually read the raster
		virtual UINT8* v_GetImagePointer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);

		//! If GetImagePointer returns a pointer directly into the frame buffer,
		//!   this function does not need to be overridden.
		//! If GetImagePointer has to actually read something (say via an X
		//!   call to the display or reading out of a raster, then this version
		//!   should just return a temporary buffer big enough to hold a single
		//!   line.
		virtual UINT8* v_GetImagePointerNoRead (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT16 FrameBuffer
			);
		
		//! Write an image buffer to the display
		//! If GetImagePointer reutrns a pointer directly into the frame buffer,
		//! this method can just return 0.  If GetImagePointer has to actually
		//! read something, this method is used to write the result back out.
		//! Typically used when drawing to a raster
		virtual ERRVALUE v_WriteImageBuffer (
			INT32 x,
			INT32 y,
			INT32 len,
			UINT8 *buf,
			UINT16 FrameBuffer
			);

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT
		friend class CONTEXT;

		//! Copy constructor private and unimplemented
		DEVICE_RASTER_TRANS_MASK(const DEVICE_RASTER_TRANS_MASK&);

		RVC::RASTER m_ObjRaster;
		RVC::OPACITYMASK m_ObjMask;
		int m_TileSizeX;
		int m_TileSizeY;
		bool m_bUseMemory;
		bool m_bDirty;

		SIMPLE_ARRAY<UINT8> m_pixbuf;
		SIMPLE_ARRAY<UINT8> m_rastbuf;
		SIMPLE_ARRAY<UINT8> m_maskbuf;

		ERRVALUE ReadPixArray ();
		ERRVALUE ReadSpan (UINT8 *pixbuf, INT32 row, INT32 col, INT32 ncols);
		ERRVALUE WritePixArray ();
		ERRVALUE WriteSpan (const UINT8 *pixbuf, INT32 row, INT32 col, INT32 ncols);

		virtual MGD::CONTEXT* v_CreateContext ();

		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

} // End of namespace MGD


#endif	// INC_MI32_MG2RASTT_H
