/**
 * \file mi32/micon.h
 * \brief Definitions for portable "icon" functions.
 *
 * \if NODOC
 * $Id: micon.h_v 1.36 2011/11/02 14:15:28 mju Exp $
 *
 * $Log: micon.h_v $
 * Revision 1.36  2011/11/02 14:15:28  mju
 * Remove redundant hdr.
 *
 * Revision 1.35  2011/09/06 21:17:42  mju
 * Exclude macro from dox.
 *
 * Revision 1.34  2008/09/02 17:41:52  mju
 * Fix doc warning.
 *
 * Revision 1.33  2008/02/04 20:40:36  mju
 * Add miconGenerateID.
 *
 * Revision 1.32  2008/01/10 15:53:55  mju
 * Add iconid member to miconstruct.
 *
 * Revision 1.31  2007/10/10 21:53:10  dwilliss
 * Changed read function to take a stylepattern_bitmap instead of just a
 * simple_array of bytes
 *
 * Revision 1.30  2007/03/29 20:29:09  dwilliss
 * Added version of miconread to read directly to a stylepattern_bitmap
 *
 * Revision 1.29  2007/02/12 18:27:33  dwilliss
 * deleted prototypes for legacy functions no longer used
 *
 * Revision 1.28  2007/02/12 14:48:38  dwilliss
 * Decided not to use descriptor for icon name.  They're all hardcoded anyway
 *
 * Revision 1.27  2007/02/09 23:58:40  dwilliss
 * Renamed the new methods to be able to distinguish them from the old ones
 * for the deprecated function scanner
 *
 * Revision 1.26  2007/02/09 23:42:35  dwilliss
 * Added prototypes for new methods that don't use numeric object handles
 *
 * Revision 1.25  2006/10/18 16:45:00  mju
 * Use miconstruct* instead of micon typedef.
 *
 * Revision 1.24  2006/08/22 22:14:22  dwilliss
 * Added a version to read an icon in the format needed by X for a window
 * property
 *
 * Revision 1.21  2005/01/14 18:25:40  mju
 * Add miconCreate using ICONID.
 *
 * Revision 1.11  2001/12/14 18:52:51  mju
 * Add MiconRead which uses ICONID.
 *
 * Revision 1.9  2001/12/04 14:48:33  mju
 * Use enum for MICONFLAGS.
 *
 * Revision 1.1  1999/01/07  23:10:27  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MI32_MICON_H
#define	INC_MI32_MICON_H

#ifndef	INC_MI32_ICONID_H
	#include <mi32/iconid.h>
#endif

#ifndef	INC_MI32_SIMPLEAR_H
	#include <mi32/simplear.h>
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
namespace RVC {
	class STYLE;
	class STYLEPATTERN_BITMAP;
	class DESCRIPTOR;
	}
class FILEPATH;
#endif

//----------------------------------------------------------------------------
//		Types and constants																		
//----------------------------------------------------------------------------

//! Enumeration for flags used in various Micon functions.
enum MICONFLAGS {
	MICONFLAG_None =					0x0000,	//!< No flags specified
	MICONFLAG_Stippled =				0x0001,	//!< Create a "dimmed" pixmap as well as a normal pixmap
	MICONFLAG_DefaultBackground =	0x0002,	//!< Use MxTopLevel()'s XmNbackground resource
	MICONFLAG_StippledSelected =	0x0008,	//!< Create "dimmed" pixmap for armed/selected state
	MICONFLAG_SizeMask =				0x0FF0,	//!< Bit mask for sizes from 1 to 255
	MICONFLAG_Size16 =				0x0100,	//!< Resize icon to 16 by 16 if smaller
	MICONFLAG_Size32 =				0x0200,	//!< Resize icon to 32 by 32 if smaller
	MICONFLAG_NoDarkenSelected =	0x1000,	//!< Don't darken background for "selected" state
	MICONFLAG_DisplayError =		0x2000,	//!< Display icon reading error even in non-programmer mode
	};
#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(MICONFLAGS)
#endif


struct MICONSTRUCT {
	MIUNICODE group[16];
	MIUNICODE name[16];
	UINT8 height;
	UINT8 width;
	UINT8 resized;
	UINT8 spare;								//!< For alignment
	UINT32 refcount;
	MPIXEL bgpixel;							//!< Background pixel
	ICONID id;
	void *privptr;
	};

DEPRECATED_MSG("Use MICONSTRUCT*") typedef MICONSTRUCT *MICON;


//----------------------------------------------------------------------------
//		Function prototypes																		
//----------------------------------------------------------------------------

// Note, these functions use C++ linkage.

//!\addtogroup Micon Micon functions
//!@{

//! Create new icon.
//!
//! Flags:
//!		MICONFLAG_DefaultBackground	Use MxTopLevel()'s XmNbackground resource
//!		MICONFLAG_SizeMask				Bit mask for sizes from 1 to 255	
//!		MICONFLAG_Size16					16 by 16
//!		MICONFLAG_Size32					32 by 32
//!		MICONFLAG_Stippled 				Create a "dimmed" pixmap as well as a normal pixmap
int MiconCreate (
	MICONSTRUCT **icon, 								//!< Icon returned
	const char *group, 						//!< Icon group, NULL for "standard"
	const char *name, 						//!< Icon name
	MPIXEL bgpixel, 							//!< Background pixel
	UINT8 width, 								//!< Icon width (0 to use value from file)
	UINT8 height, 								//!< Icon height (0 to use value from file)
	UINT32 flags								//!< Flags
	);

//! Create new icon using ICONID
//!
//! Flags:
//!		MICONFLAG_DefaultBackground	Use MxTopLevel()'s XmNbackground resource
//!		MICONFLAG_SizeMask				Bit mask for sizes from 1 to 255	
//!		MICONFLAG_Size16					16 by 16
//!		MICONFLAG_Size32					32 by 32
//!		MICONFLAG_Stippled 				Create a "dimmed" pixmap as well as a normal pixmap
int MiconCreate (
	MICONSTRUCT **icon, 								//!< Icon returned
	ICONID iconid,								//!< Icon ID
	MPIXEL bgpixel, 							//!< Background pixel
	UINT8 width, 								//!< Icon width (0 to use value from file)
	UINT8 height, 								//!< Icon height (0 to use value from file)
	UINT32 flags								//!< Flags
	);

//! Destroy icon.
void MiconDestroy (
	MICONSTRUCT *icon									//!< Icon to destroy
	);

//! Get icon height in pixels.
inline UINT8 MiconGetHeight (
	const MICONSTRUCT *icon							//!< Icon to use
	) { return (icon->height); }

//! Get icon width in pixels.
inline UINT8 MiconGetWidth (
	const MICONSTRUCT *icon							//!< Icon to use
	) { return (icon->width); }

//! Generate ICONID from icon name/group.
ICONID MiconGenerateID (
	const char *name,	 						//!< Icon name
	const char *group = 0,					//!< Icon group, NULL for "standard"
	int size = 16								//!< Icon size
	);

//! Open icon group in user-specified icon file.
ERRVALUE MiconOpen (
	RVC::STYLE& Object,							//!< Style object to open
	const FILEPATH& filepath,					//!< Icon file to open
	const MIUNICODE *group						//!< Icon group name
	);

//! Open icon group in icons.ref
ERRVALUE MiconOpen (
	RVC::STYLE& Object,							//!< Style object to open
	const MIUNICODE *group						//!< Icon group name
	);

int MiconReadPatternExt (
	RVC::STYLE& StyleObj,
	INT32 pattnum, 							//!< Pattern number if known
	const MIUNICODE* name, 					//!< Icon name
	int width, 									//!< Width, 0 for value from file
	int height, 								//!< Height, 0 for value from file
	SIMPLE_ARRAY<UINT8>& pattdata			//!< Pattern buffer allocated and returned
	);

int MiconReadPatternExt (
	RVC::STYLE& StyleObj,
	INT32 pattnum, 							//!< Pattern number if known
	const MIUNICODE* name, 					//!< Icon name
	int width, 									//!< Width, 0 for value from file
	int height, 								//!< Height, 0 for value from file
	RVC::STYLEPATTERN_BITMAP& Pattern
	);

//! Read icon(s) in standard format for use with widget.
//!
//! Flags:
//!		MICONFLAG_DefaultBackground	Use MxTopLevel()'s XmNbackground resource
//!		MICONFLAG_NoDarkenSelected		Don't darken background for "armed/selected" state
//!		MICONFLAG_SizeMask				Bit mask for sizes from 1 to 255	
//!		MICONFLAG_Size16					16 by 16
//!		MICONFLAG_Size32					32 by 32
//!		MICONFLAG_Stippled 				Create a "dimmed" pixmap as well as a normal pixmap
int MiconRead (
	const char *group, 						//!< Icon group, NULL for "standard"
	const char *name, 						//!< Icon name
	MICONSTRUCT **normalicon, 						//!< Icon with "normal" colors returned, NULL if don't need
	MICONSTRUCT **armselicon, 						//!< Icon with "armed/selected" colors returned, NULL if don't need
	MPIXEL normalbg, 							//!< Normal background pixel if default not used
	MPIXEL armselbg, 							//!<	Armed/selected background pixel if default not used
	UINT32 flags,								//!< Flags
	RVC::STYLE* StyleObj = 0
	);

//! Read icon(s) in standard format for use with widget.
//!
//! Flags:
//!		MICONFLAG_DefaultBackground	Use MxTopLevel()'s XmNbackground resource
//!		MICONFLAG_NoDarkenSelected		Don't darken background for "armed/selected" state
//!		MICONFLAG_SizeMask				Bit mask for sizes from 1 to 255	
//!		MICONFLAG_Size16					16 by 16
//!		MICONFLAG_Size32					32 by 32
//!		MICONFLAG_Stippled 				Create a "dimmed" pixmap as well as a normal pixmap 
int MiconRead (
	const char *group, 						//!< Icon group, NULL for "standard"
	const char *name, 						//!< Icon name
	MICONSTRUCT **normalicon, 						//!< Icon with "normal" colors returned, NULL if don't need
	MICONSTRUCT **armeselicon, 						//!< Icon with "armed/selected" colors returned, NULL if don't need
	UINT32 flags								//!< Flags
	);

//! Read icon for use with widget.
ERRVALUE MiconRead (
	ICONID iconid,								//!< Icon ID
	MICONSTRUCT **normalicon, 						//!< Icon with "normal" colors returned, NULL if don't need
	MICONSTRUCT **armselicon = 0,					//!< Icon with "armed/selected" colors returned, NULL if don't need
	MICONFLAGS flags = MICONFLAG_None	//!< Flags
	);

//! Read icon for use with widget with background pixel override.
ERRVALUE MiconRead (
	ICONID iconid,								//!< Icon ID
	MICONSTRUCT **normalicon, 						//!< Icon with "normal" colors returned, NULL if don't need
	MICONSTRUCT **armselicon, 						//!< Icon with "armed/selected" colors returned, NULL if don't need
	MPIXEL normalbg, 							//!< Normal background pixel if default not used
	MPIXEL armselbg, 							//!< Armed/selected background pixel if default not used
	MICONFLAGS flags = MICONFLAG_None	//!< Flags
	);

//! Read icon for use as a _NET_WM_ICON X window property
//! The data returned is an array of 32-bit values in the local byte-order
//! (the X server worries about byte swapping if necessary).
//! Format is height followed by width, followed by height*width 32-bit
//! color values in ARGB order.  
ERRVALUE MiconRead (
	ICONID iconid,								//!< Icon ID
	SIMPLE_ARRAY<UINT32>& data				//!< Icon data returned
	);

//! Read icon as bitmap pattern.
ERRVALUE MiconRead (
	ICONID iconid,
	RVC::STYLEPATTERN_BITMAP& pattern
	);
	
//! Read icon as bitmap pattern.
ERRVALUE MiconRead (
	const char *group, 						//!< Icon group, NULL for "standard"
	const char *name, 						//!< Icon name
	RVC::STYLEPATTERN_BITMAP& pattern
	);

//!@}
	
#endif	// INC_MI32_MICON_H
