/**
 *	\file mi32/mihash.h
 *	\brief MIHASH template class definition and implementation.
 *
 * \if NODOC
 * $Id: mihash.h_v 1.11 2011/11/02 14:26:08 mju Exp $
 *
 * $Log: mihash.h_v $
 * Revision 1.11  2011/11/02 14:26:08  mju
 * Remove redundant hdr.
 *
 * Revision 1.10  2006/02/09 20:47:48  dwilliss
 * fixed to compile with g++ 4.0
 *
 * Revision 1.6  2000/11/20 16:26:47  mju
 * Make begin() skip over empty buckets.
 * Make iterator increment skip empty buckets.
 *
 * Revision 1.1  2000/08/24 21:45:33  scowan
 * Initial revision
 * \endif
**/

#ifndef INC_MI32_MIHASH_H
#define INC_MI32_MIHASH_H

#ifndef	INC_MI32_MILIST_H
	#include <mi32/milist.h>
#endif

//! Items to be stored in the hash MUST define a GetHashNumber() method and 
//! define an operator==() for the hash entry for each type of object that will be 
//! used to locate the hash entry.  Eg. For RVC object searches, the hash stores the
//! inode number and file handle, the object defined to search with contains the name
//! parentinode, and object type.  Therefore an operator==() must be defined to 
//! compare the hash entry with the search object.

template <typename _ITEMTYPE> class MIHASH {
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef MILIST<_ITEMTYPE> HASHBUCKET;
		#endif // GENERATING_DOXYGEN_OUTPUT

	public:

		class ITERATOR;					//!< Forward declarations
		class CONST_ITERATOR;
		friend class CONST_ITERATOR;
		friend class ITERATOR;			//!< Allow access to LISTITEM

		class CONST_ITERATOR {
		public:
			//! Default constructor.
			CONST_ITERATOR (
				) :
				m_Hash(0),
				m_BucketIndex(0)
				{ }

			//! Copy constructor
			CONST_ITERATOR (
				const ITERATOR& rhs
				) : 
				m_Hash(rhs.m_Hash),
				m_BucketIndex(rhs.m_BucketIndex),
				m_Item(rhs.m_Item)
				{}

			//! Dereference operator, returns reference to templated item.
			const _ITEMTYPE& operator* (
				) const {
				return (*m_Item);
				}

			//! Dereference operator, returns pointer to templated item.
			const _ITEMTYPE* operator-> (
				) const {
				return (&**this);
				}

			//! Pre-increment operator
			CONST_ITERATOR& operator++ (
				) {
				if (m_BucketIndex < m_Hash->m_NumBuckets) {
					++m_Item;
					while (m_Item == m_Hash->m_BucketList[m_BucketIndex].End()) {
						++m_BucketIndex;
						if (m_BucketIndex == m_Hash->m_NumBuckets) {
							m_Item = m_Hash->m_BucketList[m_BucketIndex-1].End();
							}
						else {
							m_Item = m_Hash->m_BucketList[m_BucketIndex].Begin();
							}
						}
					}
				return (*this);
				}

			//! const_iterator operator++(int) not defined to avoid expensive copy operation
			//! const_iterator operator--(int) not defined to avoid expensive copy operation

			bool operator== (				//! Equality operator
				const CONST_ITERATOR& rhs
				) const {
				return (m_Item == rhs.m_Item && m_BucketIndex == rhs.m_BucketIndex);
				}

			bool operator!= (				//! Inequality operator
				const CONST_ITERATOR& rhs
				) const {
				return (!(*this == rhs));
				}

		protected:
			const MIHASH<_ITEMTYPE> *const m_Hash;
			UINT32 m_BucketIndex;
			typename HASHBUCKET::ITERATOR m_Item;
			
			//! Construct and iterate to first item.
			CONST_ITERATOR (
				const MIHASH<_ITEMTYPE> *const hash
				) :
				m_Hash(hash),
				m_BucketIndex(0),
				m_Item(m_Hash->m_BucketList[0].Begin())
				{
				if (m_Item == m_Hash->m_BucketList[0].End()) {
					//! First bucket is empty, iterate to non-empty bucket.
					operator++();
					}
				}

			//! Internal constructor to initialize to end.
			CONST_ITERATOR (
				const MIHASH<_ITEMTYPE> *const hash,
				int
				) :
				m_Hash(hash),
				m_BucketIndex(m_Hash->m_NumBuckets),
				m_Item(m_Hash->m_BucketList[0].Begin())
				{}

			//! Internal constructor to initialize to specific item.
			CONST_ITERATOR (
				const MIHASH<_ITEMTYPE> *const hash,
				UINT32 BucketIndex,
				typename HASHBUCKET::ITERATOR Item
				) :
				m_Hash(hash),
				m_BucketIndex(BucketIndex),
				m_Item(Item)
				{ }

			friend class MIHASH<_ITEMTYPE>;
			};

			//! CLASS iterator
		class ITERATOR : public CONST_ITERATOR {
			public:

				//! Default constructor.
				ITERATOR (
					) :
					CONST_ITERATOR()
					{}

				//! Dereference operator, returns reference to templated item
				_ITEMTYPE& operator* (
					) const {
					return (*CONST_ITERATOR::m_Item);
					}

				//! Dereference operator, returns pointer to templated item
				_ITEMTYPE* operator-> (
					) const {
					return (&**this);
					}

				//! Pre-increment operator
				ITERATOR& operator++ (
					) {
					CONST_ITERATOR::operator++();
					return (*this);
					}

				//! iterator operator++(int) not defined to avoid expensive copy operation
				//! iterator operator--(int) not defined to avoid expensive copy operation

				bool operator== (				//! Equality operator
					const ITERATOR& rhs
					) const {
					return (CONST_ITERATOR::m_Item == rhs.m_Item && CONST_ITERATOR::m_BucketIndex == rhs.m_BucketIndex);
					}

				bool operator!= (				//! Inequality operator
					const ITERATOR& rhs
					) const {
					return (!(*this == rhs));
					}
			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				//! Construct and iterate to first item.
				ITERATOR (
					const MIHASH<_ITEMTYPE> *const hash
					):
					CONST_ITERATOR(hash)
					{ }

				//! Internal constructor to initialize to end.
				ITERATOR (
					const MIHASH<_ITEMTYPE> *const hash,
					int
					):
					CONST_ITERATOR(hash,0)
					{ }

				//! Internal constructor to initialize to specific item.
				ITERATOR (
					const MIHASH<_ITEMTYPE> *const hash,
					UINT32 BucketIndex,
					typename HASHBUCKET::ITERATOR Item
					) :
					CONST_ITERATOR(hash, BucketIndex, Item)
					{ }

				friend class MIHASH<_ITEMTYPE>;
			#endif // GENERATING_DOXYGEN_OUTPUT
			};


		// CONSTRUCTORS / DESTRUCTOR
		
		//! Constructor
		explicit MIHASH (
			UINT32 NumBuckets = 32			//!< Number of buckets to hash into
			) :
			m_NumBuckets(NumBuckets),
			m_BucketList(new HASHBUCKET[NumBuckets])
			{}
			
		//! Copy constructor
		MIHASH (
			const MIHASH& rhs
			) :
			m_NumBuckets(rhs.m_NumBuckets),
			m_BucketList(new HASHBUCKET[rhs.m_NumBuckets])
			{
			for (UINT32 i = 0;(i < m_NumBuckets);++i) {
				m_BucketList[i] = rhs.m_BucketList[i];
				}
			}

		//! Destructor		
		~MIHASH (
			) {
			delete [] m_BucketList;
			}
			
		// METHODS
		
		//! Add an item to the hash
		void Add (
			const _ITEMTYPE& item
			) {
			UINT32 HashNum = item.GetHashNumber() % m_NumBuckets;
			typename HASHBUCKET::CONST_ITERATOR it;
			for (it = m_BucketList[HashNum].Begin();(it != m_BucketList[HashNum].End());++it) {
				if (*it == item) break;
				}
			if (it == m_BucketList[HashNum].End()) m_BucketList[HashNum].PushBack(item);
			return;
			}
			
		//! Create const iterator to the beginning of the hash table
		//!
		//! @return Constant iterator to hash table
		CONST_ITERATOR Begin (
			) const {
			return (CONST_ITERATOR(this));
			}

		//! Create iterator to the beginning of the hash table
		//!
		//! @return Iterator to hash table
		ITERATOR Begin (
			) {
			return (ITERATOR(this));
			}
			
		//! Clear the hash of all entries
		void Clear (
			) {
			for (UINT32 i = 0;(i < m_NumBuckets);++i) {
				m_BucketList[i].Clear();
				}
			return;
			}		
			
		//! Create const iterator to the end of the hash table
		//!
		//! @return Constant iterator to hash table
		CONST_ITERATOR End (
			) const {
			return (CONST_ITERATOR(this,0));
			}

		//! Create iterator to the end of the hash table
		//!
		//! @return Iterator to hash table
		ITERATOR End (
			) {
			return (ITERATOR(this,0));
			}

		//! Determine if an item is in the hash table
		//!
		//! @return True if item is in the hash table
		template <typename _CMP> bool Exists (
			const _CMP& item
			) const {
			return (Find(item) != End());
			}
			
		//! Locate an item in the hash table
		//!
		//! @return Iterator if item exists, End() if not
		template <typename _CMP> ITERATOR Find (
			const _CMP& item
			) {
			UINT32 HashNum = item.GetHashNumber() % m_NumBuckets;
			typename HASHBUCKET::ITERATOR it;
			for (it = m_BucketList[HashNum].Begin();(it != m_BucketList[HashNum].End());++it) {
				if (item == *it) break;		//!< The comparison MUST remain in this order, changing it will break much code.
				}
			ITERATOR retval = (it != m_BucketList[HashNum].End()) ? ITERATOR(this, HashNum, it) : End();
			return (retval);
			}
			
		//! Locate an item in the hash table
		//!
		//! @return Constant iterator if item exists, End() if not
		template <typename _CMP> CONST_ITERATOR Find (
			const _CMP& item
			) const {
			UINT32 HashNum = item.GetHashNumber() % m_NumBuckets;
			typename HASHBUCKET::ITERATOR it;
			for (it = m_BucketList[HashNum].Begin();(it != m_BucketList[HashNum].End());++it) {
				if (item == *it) break;		//!< The comparison MUST remain in this order, changing it will break much code.
				}
			CONST_ITERATOR retval = (it != m_BucketList[HashNum].End()) ? CONST_ITERATOR(this, HashNum, it) : End();
			return (retval);
			}
			
		//! Remove item from hash table
		void Remove (
			ITERATOR item
			) {
			if (item != End()) m_BucketList[item.m_BucketIndex].Remove(item.m_Item);
			return;
			}
			
		//! Resize the number of buckets in the hash table, clears the table first
		void Resize (
			INT32 NumBuckets
			) {
			Clear();
			delete [] m_BucketList;
			m_BucketList = new HASHBUCKET[NumBuckets];
			m_NumBuckets = NumBuckets;
			return;
			}

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		MIHASH& operator= (
			const MIHASH& rhs
			);

		UINT32 m_NumBuckets;
		HASHBUCKET *m_BucketList;
		#endif // GENERATING_DOXYGEN_OUTPUT
		
	};

#endif		// INC_MI32_MIHASH_H
