/**
 * \file mi32/mimatrix.h
 * \brief MIMATRIX class definitions.
 *
 * \if NODOC
 * $Id: mimatrix.h_v 1.9 2012/07/30 20:04:19 mju Exp $
 *
 * $Log: mimatrix.h_v $
 * Revision 1.9  2012/07/30 20:04:19  mju
 * Add () operators for element access.
 *
 * Revision 1.8  2007/09/11 16:21:02  mju
 * Class moved to geom dll.
 *
 * Revision 1.7  2006/07/27 15:07:18  dwilliss
 * Added a RotateClockwise method
 *
 * Revision 1.6  2005/08/12 15:18:04  mju
 * Make operators inline.
 *
 * Revision 1.5  2005/08/10 14:23:02  mju
 * Disallow access as 2D array as implies a particular internal storage format which may not be maintained.
 * Remove op+= and -= using scalar as inappropriate for matrices.
 *
 * Revision 1.4  2005/08/10 14:02:37  mju
 * Add binary operators.
 *
 * Revision 1.3  2005/08/10 13:18:11  mju
 * Add ability to access as 'vector'.
 *
 * Revision 1.2  2005/08/09 16:03:12  mju
 * Add operators, clarifications.
 *
 * Revision 1.1  2005/08/09 14:47:59  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_MIMATRIX_H
#define	INC_MI32_MIMATRIX_H

#ifndef INC_MI32_DOUBLEAR_H
#include <mi32/doublear.h>
#endif

#ifdef GEOMDLL
	#define CLASSEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSEXPORT MI_DLLCLASSIMPORT
#endif

//! Matrix class.
class CLASSEXPORT MIMATRIX {
	public:

		//! Default constructor.
		MIMATRIX ();

		//! Copy constructor.
		MIMATRIX (const MIMATRIX& rhs);

		//! Destructor.
		~MIMATRIX ();

		//! Assignment operator.
		MIMATRIX& operator= (const MIMATRIX& rhs);

		//! Multiply by scalar.
		MIMATRIX& operator*= (
			double value
			) { MultiplyScalar(value); return (*this); }

		//! Divide by scalar.
		MIMATRIX& operator/= (
			double value
			) { MultiplyScalar(1.0/value); return (*this); }

		//! Negation.
		const MIMATRIX operator- (
			) { return (MIMATRIX(OP_Negate,*this)); }

		//! 1-D matrix element accessor.
		double operator () (
			UINT32 num							//! Vector element number (0-based)
			) const { return (m_values[static_cast<int>(num)]); }

		//! 1-D matrix element accessor.
		double& operator () (
			UINT32 num							//! Vector element number (0-based)
			) { return (m_values[static_cast<int>(num)]); }

		//! 2-D matrix element accessor.
		double operator () (
			UINT32 row,							//!< Row (0-based)
			UINT32 col							//!< Column (0-based)
			) const { return (m_values[static_cast<int>(row*m_Cols+col)]); }

		//! 2-D matrix element accessor.
		double& operator () (
			UINT32 row,							//!< Row (0-based)
			UINT32 col							//!< Column (0-based)
			) { return (m_values[static_cast<int>(row*m_Cols+col)]); }

		//! Add value to specified element.
		void Add (
			UINT32 row,
			UINT32 col,
			double value
			);

		//! Add matrix.
		//! Matrix to be added must be same size as this matrix.
		ERRVALUE Add (
			const MIMATRIX& matrix
			);

		//! Add scalar value to all elements in matrix.
		void AddScalar (
			double value
			);

		//! Compute "condition number" of matrix.
		//!
		//! @return 0.0 instead of INFINITY to simplify handling of the result.
		//!
		//!	The condition number of a matrix is defined as the ratio of the largest
		//!	(in magnitude) of the elements of matrix W to the smallest one.
		//!	Where W is one of the results of SVD defined as:
		//!					    	 T
		//!			A = U * W * V	(see ComputeSVD).					
		//!	A matrix is singular if it's condition number is large.
		double ComputeConditionNumber () const;

		//! Compute determinant.
		double ComputeDeterminant () const;

		//! Compute Eigenvectors and Eigenvalues from symmetric matrix.
		//! Matrix must be symmetric, error will be returned if not.
		ERRVALUE ComputeEigen (
			MIMATRIX& EigenVectors,			//!< Eigenvectors returned as columns
			MIMATRIX& EigenValues			//!< Eigenvalues returned
			) const;

		//! Compute inverse of square matrix.
		ERRVALUE ComputeInverse (
			MIMATRIX& result,					//!< Result matrix
			UINT32 MaxIterations = 0,		//!< Maximum number of iterations
			double MaxError = 0
			) const;

		//! Compute maximum absolute element value.
		double ComputeMaxAbsolute (
			) const;

		//! Compute pseudoinverse of matrix using SVD  
		//!
		//! If Input matrix is not singular pseudoinverse is equal to
		//!	normal inverse of the matrix.	
		//!	Pseudoinverse is such matrix X that satisfy the following condition:
		//!		A * X * A = A, if matrix A is singular X still exists.
		ERRVALUE ComputePseudoInverse (
			MIMATRIX& result,					//!< Result matrix
			double tolerance = 0				//!< Tolerance to use, 0 for default
			) const;

		//! Compute Singular Value Decomposition (SVD).
		//!
		//! This function computes SVD as A(MxN) = U(MxN) * W(N) * V(NxN)[transpose] ;
		//!	Output matrix U has the same dimension as input matrix A and the columns
		//!	will be the orthonormal basis vectors for A.
		//!	The diagonal matrix of singular values W is output as a vector.
		//!	The matrix V (not transpose of V) is output as a matrix.
		ERRVALUE ComputeSVD (
			MIMATRIX& U,						//! Returned
			MIMATRIX& W,						//! Returned
			MIMATRIX& V							//! Returned
			) const;

		//! Compute trace (sum of diagonal elements).
		double ComputeTrace () const;

		//! Create matrix of specified size.
		ERRVALUE Create (
			UINT32 Rows,						//!< Number of rows
			UINT32 Cols,						//!< Number of columns
			const double *values = 0		//!< Initial values as Rows * Cols array, if not specified all values will be zero
			);

		//! Create diagonal matrix of specified size.
		ERRVALUE CreateDiagonal (
			UINT32 Size,						//!< Size of matrix to create
			double value						//!< Value to replicate on diagonal
			);

		//! Create diagonal matrix from vector.
		//! Will create a square matrix of order N + ABS(K) with elements of input on the K-th diagonal.
		//!	K = 0 is the main diagonal, K > 0 is above main diagonal, K < 0 below.
		ERRVALUE CreateDiagonal (
			const MIMATRIX& vector			//!< One-dimensional matrix containing diagonal values
			);

		//! Create matrix as product of two matrices.
		ERRVALUE CreateProduct (
			const MIMATRIX& left,			//!< Left side of multiplication
			const MIMATRIX& right			//!< Right side of multiplication
			);

		//! Get diagonal elements as column vector.
		ERRVALUE GetDiagonal (
			MIMATRIX& result					//!< Column vector returned
			) const;

		//! Get number of columns in matrix.
		UINT32 GetNumColumns (
			) const { return (m_Cols); }

		//! Get number of matrix elements.
		UINT32 GetNumElements (
			) const { return (m_values.GetNumItems()); }

		//! Get number of rows in matrix.
		UINT32 GetNumRows (
			) const { return (m_Rows); }

		//! Get value at location with no validation.
		double GetValueFast (
			UINT32 row,							//!< Row (0-based)
			UINT32 col							//!< Column (0-based)
			) const { return (m_values[static_cast<int>(row*m_Cols+col)]); }

		//! Get value at location with validation of element location.
		double GetValueValidated (
			UINT32 row,							//!< Row (0-based)
			UINT32 col							//!< Column (0-based)
			) const;

		//! Get vector value at location with no validation.
		//! Intended for use on 'vector' matrix (1 x N or N x 1) only.
		double GetValueFast (
			UINT32 num							//! Vector element number
			) const { return (m_values[static_cast<int>(num)]); }

		//! Get vector value at location with validation of element location.
		//! Intended for use on 'vector' matrix (1 x N or N x 1) only.
		double GetValueValidated (
			UINT32 num							//! Vector element number
			) const;

		//! Get values as array.
		void GetValues (
			double *values						//!< Values returned, must be large enough to hold all elements
			) const;

		//! Determine if matrices are equal within specified tolerance.
		//! Will return false if matrices are not the same size.
		bool IsEqual (
			const MIMATRIX& matrix,
			double tolerance = 0
			) const;

		//! Determine if matrix has at one non-zero element on main diagonal and zero everywhere else.
		bool IsDiagonal (
			double tolerance = 0
			) const;

		//! Determine if matrices are the same size.
		bool IsSameSize (
			const MIMATRIX& matrix
			) const { return (m_Rows == matrix.m_Rows && m_Cols == matrix.m_Cols); }

		//! Determine if matrix is singular.
		//! Uses ComputeConditionNumber with default limits.  This is not a fast operation.
		bool IsSingular (
			) const;

		//! Determine if matrix is symmetric.
		bool IsSymmetric (
			double tolerance = 0
			) const;

		//! Determine if matrix is a 'vector' (1 x N or N x 1 matrix).
		bool IsVector (
			) const { return (m_Rows == 1 || m_Cols == 1); }

		//! Left multiply (this = left * this).
		ERRVALUE MultiplyLeft (
			const MIMATRIX& left
			);

		//! Left multiply (this = this * right).
		ERRVALUE MultiplyRight (
			const MIMATRIX& right
			);

		//! Multiply all matrix elements by specified value.
		void MultiplyScalar (
			double value
			);

		//! Replace elements with submatrix of elements from another matrix.
		ERRVALUE ReplaceSubMatrix (
			UINT32 TgtStartRow,				//! Target matrix starting row
			UINT32 TgtStartCol,				//! Target matrix starting column
			const MIMATRIX& SrcMatrix,		//! Source matrix
			UINT32 SrcStartRow = 0,			//! Source matrix starting row
			UINT32 SrcStartCol = 0,			//! Source matrix starting column
			UINT32 NumRows = 0,				//! Number of rows to replace, 0 for maximum possible
			UINT32 NumCols = 0				//! Number of columns to replace, 0 for maximum possible
			);

		//! Resize matrix.
		//! Newly created elements will be initialized to 0.
		ERRVALUE Resize (
			UINT32 Rows,
			UINT32 Cols
			);

		//! Rotate the matrix 90 degrees clockwise
		//! If tha matrix is not square, it will be resized (i.e., an NxM matrix
		//! will become an MxN matrix
		void RotateClockwise (
			);

		//! Set all elements in matrix to a specified scalar value.
		void SetScalar (
			double value
			);

		//! Set matrix to identity.
		void SetIdentity (
			);

		//! Set value at location with no validation.
		void SetValueFast (
			UINT32 row,
			UINT32 col,
			double value
			) { m_values[static_cast<int>(row*m_Cols+col)] = value; }

		//! Set value at location with validation of element location.
		void SetValueValidated (
			UINT32 row,
			UINT32 col,
			double value
			);

		//! Set vector value at location with no validation.
		//! Intended for use on 'vector' matrix (1 x N or N x 1) only.
		void SetValueFast (
			UINT32 num,							//! Vector element number
			double value
			) { m_values[static_cast<int>(num)] = value; }

		//! Set vector value at location with validation of element location.
		//! Intended for use on 'vector' matrix (1 x N or N x 1) only.
		void SetValueValidated (
			UINT32 num,							//! Vector element number
			double value
			);

		//! Set all matrix elements to zero.
		void SetZero (
			) { m_values.ClearItems(); }

		//! Solve system of linear equations and put result in this matrix.
		//!	A * X = B where A(M x N) matrix and X (N) vector. B(M) is also a vector.
		//! This is done using ComputeSVD, if already have performed SVD then can
		//! use SolveLinearSVD for improved performance.
		ERRVALUE SolveLinear (
			const MIMATRIX& A,				//!< M x N matrix of linear equations
			const MIMATRIX& B,				//!< Vector of linear equation values
			double tolerance = 0				//!< Tolerance, 0 to use default based on maximum value in SVD W matrix
			);

		//! Solve system of linear equations using SVD components.
		ERRVALUE SolveLinearSVD (
			const MIMATRIX& U,				//!< From ComputeSVD
			const MIMATRIX& W,				//!< From ComputeSVD
			const MIMATRIX& V,				//!< From ComputeSVD
			const MIMATRIX& B,
			double tolerance = 0				//!< Tolerance, 0 to use default based on maximum value in SVD W matrix
			);

		//! Subtract matrix.
		//! Matrix to be subtracted must be same size as this matrix.
		ERRVALUE Subtract (
			const MIMATRIX& matrix
			);

		//! Swap this matrix with another.
		//! This is done by swapping memory pointers so is very fast.
		void Swap (
			MIMATRIX& other					//!< Matrix to swap with
			);

		//! Transpose matrix 
		void Transpose (
			);

		#ifndef GENERATING_DOXYGEN_OUTPUT

		enum OP {
			OP_Add,
			OP_Subtract,
			OP_Multiply,
			OP_Negate
			};

		// Algebraic constructor used by binary operators to facilitate Return Value Optimization.
		MIMATRIX (
			OP op,
			const MIMATRIX& lhs,
			const MIMATRIX& rhs
			);

		// Algebraic constructor used by binary operators to facilitate Return Value Optimization.
		MIMATRIX (
			OP op,
			const MIMATRIX& lhs,
			double rhs = 0
			);

	private:

		DOUBLE_ARRAY<double> m_values;
		UINT32 m_Rows;
		UINT32 m_Cols;

		// Access vector element (const).
		const double& element (
			UINT32 num
			) const { return (m_values[static_cast<int>(num)]); }

		// Access vector element (non-const).
		double& element (
			UINT32 num
			) { return (m_values[static_cast<int>(num)]); }

		// Access element (const).
		const double& element (
			UINT32 row,
			UINT32 col
			) const { return (m_values[static_cast<int>(row*m_Cols+col)]); }

		// Access element (non-const).
		double& element (
			UINT32 row,
			UINT32 col
			) { return (m_values[static_cast<int>(row*m_Cols+col)]); }

		ERRVALUE HouseholderReduction (MIMATRIX& E, MIMATRIX& D);
		ERRVALUE QLImplicit (MIMATRIX& E, MIMATRIX& D);

		#endif	// GENERATING_DOXYGEN_OUTPUT
	};


//! Matrix addition.
inline const MIMATRIX operator+ (
	const MIMATRIX& lhs,
	const MIMATRIX& rhs
	) { return (MIMATRIX(MIMATRIX::OP_Add,lhs,rhs)); }

//! Matrix subtraction.
inline const MIMATRIX operator- (
	const MIMATRIX& lhs,
	const MIMATRIX& rhs
	) { return (MIMATRIX(MIMATRIX::OP_Subtract,lhs,rhs)); }

//! Matrix multiplication.
inline const MIMATRIX operator* (
	const MIMATRIX& lhs,
	const MIMATRIX& rhs
	) { return (MIMATRIX(MIMATRIX::OP_Multiply,lhs,rhs)); }

//! Multiply by scalar.
inline const MIMATRIX operator* (
	const MIMATRIX& lhs,
	double rhs
	) { return (MIMATRIX(MIMATRIX::OP_Multiply,lhs,rhs)); }

//! Divide by scalar.
inline const MIMATRIX operator/ (
	const MIMATRIX& lhs,
	double rhs
	) { return (MIMATRIX(MIMATRIX::OP_Multiply,lhs,1.0/rhs)); }

//! Multiply scalar by matrix.
inline const MIMATRIX operator* (
	double lhs,
	const MIMATRIX& rhs
	) { return (MIMATRIX(MIMATRIX::OP_Multiply,rhs,lhs)); }


#undef CLASSEXPORT

#endif	// INC_MI32_MIMATRIX_H 

