/**
 * \file <mi32/recordarray.h>
 * \brief RECORDARRAY definitions
 *
 * \if NODOC
 * $Id: recordarray.h_v 1.4 2011/07/15 17:34:11 mju Exp $
 *
 * $Log: recordarray.h_v $
 * Revision 1.4  2011/07/15 17:34:11  mju
 * Add record::setArrayIdx.
 *
 * Revision 1.3  2011/07/08 13:23:09  mju
 * Add record::getArrayIdx.
 *
 * Revision 1.2  2011/06/24 19:57:47  mju
 * Fix inclusion guards.
 *
 * Revision 1.1  2011/06/24 19:54:47  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_RECORDARRAY_H
#define	INC_MI32_RECORDARRAY_H

#ifndef	INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

#ifdef MISYSTEMDLL
	#define CLASSLIBEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSLIBEXPORT MI_DLLCLASSIMPORT
#endif


//=====================================================================================================================

//! General array of records with simple iteration support.
class CLASSLIBEXPORT RECORDARRAY {
	public:

		//! Record in array with iteration capability.
		class CLASSLIBEXPORT RECORD {
			public:

				//! Default constructor, creates NULL record not associated with array.
				RECORD ();

				//! Copy constructor.
				RECORD (const RECORD& rhs);

				//! Construct and attach to array.
				RECORD (RECORDARRAY& Array);

				//! Construct and attach to array with index.
				RECORD (RECORDARRAY& Array, UINT32 ArrayIdx);

				//! Destructor.
				~RECORD ()
					{ }
											
				// Assignment.
				RECORD& operator= (
					const RECORD& rhs
					);

				//! Pre-increment to next record in array.
				RECORD& operator++ (
					) {
					++m_ArrayIdx;
					m_Data += m_RecordSize;
					return (*this);
					}

				//! Pre-decrement to next record in array.
				RECORD& operator-- (
					) {
					if (m_ArrayIdx == 0) {
						m_ArrayIdx = m_ArraySize;
						m_Data = m_pArray->GetPtrBuffer() + m_ArrayIdx * m_RecordSize;
						}
					else {
						--m_ArrayIdx;
						m_Data -= m_RecordSize;
						}
					return (*this);
					}

				//! Determine if point record is at (past actually) end of array.
				bool AtEnd () const
					{ return (m_ArrayIdx >= m_ArraySize); }

				//! Clear contents of record.
				void Clear ();

				//! Get reference to array containing record.
				const RECORDARRAY& GetArray () const
					{ return (*m_pArray); }

				//! Get reference to array containing record.
				RECORDARRAY& GetArray ()
					{ return (*m_pArray); }

				//! Get array index.
				UINT32 GetArrayIdx () const
					{ return (m_ArrayIdx); }

				//! Get (const) data pointer.
				const UINT8* GetData () const
					{ return (m_Data); }

				//! Get (non-const) data pointer.
				UINT8* GetData ()
					{ return (m_Data); }

				//! Get record size.
				UINT32 GetRecordSize () const
					{ return (m_RecordSize); }

				//! Set array index.
				void SetArrayIdx (
					UINT32 ArrayIdx
					) {
					m_ArrayIdx = ArrayIdx;
					m_Data = m_pArray->GetPtrBuffer() + m_ArrayIdx * m_RecordSize;
					}

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				RECORDARRAY *m_pArray;
				UINT8 *m_Data;
				UINT32 m_ArrayIdx;
				UINT32 m_ArraySize;
				UINT32 m_RecordSize;
				#endif
			};

		//! Construct with optional record size.
		explicit RECORDARRAY (
			UINT32 RecordSize = 1			//!< Record size, cannot be 0
			);

		//! Destructor.
		~RECORDARRAY ();

		//! Get number of records.
		UINT32 GetNumRecords () const
			{ return (m_Buffer.GetNumItems() / m_RecordSize); }

		//! Get (const) pointer to start of buffer.
		const UINT8* GetPtrBuffer () const
			{ return (m_Buffer); }

		//! Get (non-const) pointer to start of buffer.
		UINT8* GetPtrBuffer ()
			{ return (m_Buffer); }

		//! Get (const) pointer to record at specified index.
		const UINT8* GetPtrRecord (
			UINT32 idx
			) const { return (static_cast<const UINT8*>(m_Buffer) + idx*m_RecordSize); }

		//! Get (non-const) pointer to start of buffer.
		UINT8* GetPtrRecord (
			UINT32 idx
			) { return (static_cast<UINT8*>(m_Buffer) + idx*m_RecordSize); }

		//! Get size of record.
		UINT32 GetRecordSize () const
			{ return (m_RecordSize); }

		//! Resize array.
		//! All RECORD instances referencing this array will become invalid.
		//! Product of number of records and record size must be less than 2^32.
		ERRVALUE Resize (
			UINT32 NumRecords				//!< Number of records
			);

		//! Set record size.
		//! Number of records will be reset to 0.
		void SetRecordSize (
			UINT32 RecordSize				//!< Record size, cannot be 0
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<UINT8> m_Buffer;
		UINT32 m_RecordSize;

		// Unimplemented.
		RECORDARRAY (const RECORDARRAY&);
		RECORDARRAY& operator= (const RECORDARRAY&);
		#endif
	};

//=====================================================================================================================

#undef	CLASSLIBEXPORT

#endif	// INC_MI32_RECORDARRAY_H
