/**
 * \file spatrefext.h <mi32/spatrefext.h>
 * \brief Spatial Reference (SPATREF) external system converters
 *
 * $Id: spatrefext.h_v 1.29 2010/05/17 19:44:42 mju Exp $
 *
 * $Log: spatrefext.h_v $
 * Revision 1.29  2010/05/17 19:44:42  mju
 * Add FGDC.
 *
 * Revision 1.28  2009/03/11 17:04:11  mju
 * Add geotiff::writer_memory.
 *
 * Revision 1.27  2008/06/18 21:39:22  mju
 * For geotiff::convertFromCRS return whether wrote projected as geographic.
 *
 * Revision 1.26  2008/06/18 21:25:21  mju
 * Add geotiff::convertFromCRS and writer interface class.
 *
 * Revision 1.25  2008/06/18 20:19:01  mju
 * Add codespace parm to geotiff::convertToCRS.
 *
 * Revision 1.24  2008/06/18 16:12:38  mju
 * Export geotiff reader classes.
 *
 * Revision 1.23  2008/06/18 13:24:18  mju
 * Add geotiff converter.
 *
 * Revision 1.22  2007/05/02 14:44:55  mju
 * Fwd declare identifier/list.
 *
 * Revision 1.21  2007/03/19 17:00:07  dfriberg
 * Forward declare GDALDataset
 *
 * Revision 1.20  2007/03/19 16:46:51  dfriberg
 * Make new class for GDAL
 *
 * Revision 1.19  2006/10/18 16:21:51  mju
 * Add esri::convertZoneToLegacyMI.
 *
 * Revision 1.18  2006/10/06 15:43:28  mju
 * Fwd declare filepath.
 *
 * Revision 1.17  2006/07/21 16:24:50  dfriberg
 * Set default parameter value for GetWKT
 *
 * Revision 1.16  2006/07/21 15:51:35  dfriberg
 * Comments
 *
 * Revision 1.15  2006/07/21 15:43:42  dfriberg
 * Change ESRI_OLD to ESRI
 *
 * Revision 1.14  2005/12/19 21:39:33  mju
 * Export readImageGeoref from dll.
 *
 * Revision 1.13  2005/12/19 20:44:29  mju
 * Add ReadImageGeoreference.
 *
 * Revision 1.12  2005/11/22 14:04:50  mju
 * Add tydac converter class.
 *
 * Revision 1.11  2005/09/07 16:46:11  mju
 * Add mapinfo.readTAB method.
 *
 * Revision 1.10  2005/05/27 17:00:16  mju
 * Change URN to generalize extaction from idlist and conversion to id.
 *
 * Revision 1.9  2005/05/27 14:18:36  mju
 * Add URN conversions to/from COORDOPDEF and COORDOPMETHOD.
 *
 * Revision 1.8  2005/05/13 16:22:21  mju
 * Add ogc urn converter.
 *
 * Revision 1.7  2005/03/15 14:59:46  mju
 * Add converter for ERMapper.
 *
 * Revision 1.6  2005/03/02 21:45:02  mju
 * Add mapinfo class.
 *
 * Revision 1.5  2005/01/31 17:56:41  dwilliss
 * Would help if ESRI_OLD was actually exported from the dll :-)
 *
 * Revision 1.4  2005/01/28 15:14:54  mju
 * Add ESRI_OLD converter.
 *
 * Revision 1.3  2004/08/23 15:07:51  mju
 * Add ogc_wkt.getName.
 *
 * Revision 1.2  2004/08/23 14:26:42  mju
 * Require codespace for WKT conversion.
 *
 * Revision 1.1  2004/07/20 21:56:03  mju
 * Initial revision
**/

#ifndef	INC_MI32_SPATREFEXT_H
#define	INC_MI32_SPATREFEXT_H

#ifndef	INC_MI32_SPATREF_H
#include <mi32/spatref.h>
#endif

#ifndef	INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

#ifndef	INC_MI32_CTRLPT_H
#include <mi32/ctrlpt.h>
#endif

#ifndef	INC_MI32_CODESPACE_H
#include <mi32/codespace.h>
#endif

class FILEPATH;
class TRANS2D_AFFINE;
class GDALDataset;
class IDENTIFIER;
class IDENTIFIERLIST;

#ifdef GEOMDLL
	#define LIBEXPORT MI_DLLEXPORT
	#define CLASSLIBEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSLIBEXPORT MI_DLLCLASSIMPORT
	#define LIBEXPORT MI_DLLIMPORT
#endif

namespace SPATREF {

//=====================================================================================================================

//! Converter for Open Geospatial Consortium Well-Known Text forms.
//! Supports multiple codespaces including OGC_WKT, ESRI_WKT and Oracle_WKT.
class CLASSLIBEXPORT OGC_WKT {
	public:
		
		//! Convert COORDREFSYS to WKT string.
		static ERRVALUE ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			MISTRING& wktstr,								//!< Well-Known Text string returned
			CODESPACE codespace							//!< WKT Codespace (eg, CODESPACE_OGC_WKT, _ESRI_WKT, _Oracle_WKT)
			);

		//! Convert WKT string to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const MISTRING& wktstr,						//!< Well-Known Text string to convert
			COORDREFSYS& CoordRefSys,					//!< Coordinate Reference System returned
			CODESPACE codespace							//!< WKT Codespace (eg, CODESPACE_OGC_WKT, _ESRI_WKT, _Oracle_WKT)
			);

		//! Get name for operation method.
		//! @return TRUE if found in specified codespace, FALSE if found in alternative (OGC_WKT) codespace, < 0 if not found
		static int GetName (
			const COORDOPMETHOD& method,				//!< Method to determine name of
			CODESPACE codespace,							//!< WKT Codespace (eg, CODESPACE_OGC_WKT, _ESRI_WKT, _Oracle_WKT)
			MISTRING& name									//!< Name returned
			);

		//! Construct with specified codespace.
		explicit OGC_WKT (
			CODESPACE codespace							//!< WKT Codespace (eg, CODESPACE_OGC_WKT, _ESRI_WKT, _Oracle_WKT)
			);

		//! Copy constructor.
		OGC_WKT (
			const OGC_WKT& rhs
			);

		//! Construct from WKT string.
		OGC_WKT (
			const MISTRING& wktstr,
			CODESPACE codespace							//!< WKT Codespace (eg, CODESPACE_OGC_WKT, _ESRI_WKT, _Oracle_WKT)
			);

		//! Construct from COORDREFSYS.
		OGC_WKT (
			const COORDREFSYS& CoordRefSys,
			CODESPACE codespace							//!< WKT Codespace (eg, CODESPACE_OGC_WKT, _ESRI_WKT, _Oracle_WKT)
			);

		// Destructor.
		~OGC_WKT ();

		//! Assignment from OGC_WKT.
		OGC_WKT& operator= (
			const OGC_WKT& rhs
			);

		//! Assignment from WKT string.
		OGC_WKT& operator= (
			const MISTRING& wktstr
			);

		//! Assignment from COORDREFSYS.
		OGC_WKT& operator= (
			const COORDREFSYS& CoordRefSys
			);

		//! Get coordinate reference system.
		ERRVALUE GetCoordRefSys (
			COORDREFSYS& CoordRefSys					//!< Coordinate Reference System returned
			);

		//! Get Well-Known Text string.
		ERRVALUE GetString (
			MISTRING& wktstr								//!< Well-Known Text string returned
			);

		//! Set codespace to use for identifiers.
		void SetCodeSpace (
			CODESPACE codespace							//!< WKT Codespace (eg, CODESPACE_OGC_WKT, _ESRI_WKT, _Oracle_WKT)
			);

	private:
		COORDREFSYS m_CoordRefSys;
		MISTRING m_wktstr;
		CODESPACE m_codespace;
	};

//=====================================================================================================================

//! Converter for Open Geospatial Consortium (OGC) Universal Resource Name (URN) forms.
class CLASSLIBEXPORT OGC_URN {
	public:
		
		//! Convert COORDOPDEF to URN string.
		static ERRVALUE ConvertFromCoordOpDef (
			const COORDOPDEF& CoordOpDef,				//!< Coordinate Operation Definition to convert
			MISTRING& urnstr								//!< Universal Resource Name string returned
			);

		//! Convert COORDREFSYS to URN string.
		static ERRVALUE ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			MISTRING& urnstr								//!< Universal Resource Name string returned
			);

		//! Convert URN string to COORDOPDEF.
		static ERRVALUE ConvertToCoordOpDef (
			const MISTRING& urnstr,						//!< Universal Resource Name string to convert
			COORDOPDEF& CoordOpDef						//!< Coordinate Operation Definition returned
			);

		//! Convert URN string to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const MISTRING& urnstr,						//!< Universal Resource Name string to convert
			COORDREFSYS& CoordRefSys					//!< Coordinate Reference System returned
			);

		//! Convert URN string to IDENTIFIER.
		static ERRVALUE ConvertToId (
			const MISTRING& urnstr,						//!< Universal Resource Name string to convert
			const char *objectType,						//!< OGC object type string
			IDENTIFIER& id									//!< Identifier returned
			);

		//! Extract from IDENTIFIERLIST to URN string.
		static ERRVALUE ExtractFromIdList (
			const IDENTIFIERLIST& IdList,				//!< Identifier list
			const char *objectType,						//!< OGC object type string
			const char *ogcversion,						//!< OGC version, will default to 1.0 if NULL
			MISTRING& urnstr								//!< Universal Resource Name string returned
			);

	};

//=====================================================================================================================

//! Converter for ESRI projection parameters.  Handles both "old" (< version 8) and WKT strings.
class CLASSLIBEXPORT ESRI {
	public:
		
		//! Convert COORDREFSYS to string list.
		//! @return TRUE if successful, FALSE if can't convert, < 0 if error
		static ERRVALUE ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			MISTRINGLIST& StrList,						//!< Formatted string list returned
			bool GetWKT = true							//!< Return WKT string
			);
	
		//! Convert COORDREFSYS to PRJ file.
		//! @return TRUE if successful, FALSE if can't convert, < 0 if error
		static ERRVALUE ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			const FILEPATH& fp,							//!< PRJ file
			bool GetWKT	= true							//!< Return WKT string
			);

		//! Convert string list to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const MISTRINGLIST& StrList,				//!< Formatted string list
			COORDREFSYS& CoordRefSys,					//!< Coordinate Reference System returned
			double& xshift,								//!< X shift value returned
			double& yshift,								//!< Y shift value returned
			double& scale									//!< Scale value returned
			);

		//! Convert PRJ file to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const FILEPATH& fp,							//!< PRJ file
			COORDREFSYS& CoordRefSys,					//!< Coordinate Reference System returned
			double& xshift,								//!< X shift value returned
			double& yshift,								//!< Y shift value returned
			double& scale									//!< Scale value returned
			);

		//! Convert ESRI "zone" to legacy MicroImages zone.
		//! Only valid for State Plane CRS.
		//! @return Legacy MicroImages zone or error < 0 if can't convert.
		static int ConvertZoneToLegacyMI (
			int esrizone									//!< ESRI zone to convert
			);
	};

//=====================================================================================================================

//! Converter for Federal Geographic Data Committee (FDGC) metadata projection parameters.
class CLASSLIBEXPORT FGDC {
	public:

		//! Convert "spref" node from metadata to CRS.
		static ERRVALUE ConvertToCRS (
			const XMLNODE *pNodeSpRef,						//!< Pointer to "spref" node in metadata
			COORDREFSYS& CoordRefSys						//!< CRS returned
			);
	};
	
//=====================================================================================================================

//! Converter for GDAL projection parameters.
class CLASSLIBEXPORT GDAL {
	public:
		static ERRVALUE ConvertToCRS (
			const GDALDataset& dataset,
			COORDREFSYS& CoordRefSys,
			TRANS2D_AFFINE& Trans,
			SIMPLE_ARRAY<CTRLPOINT3>& ControlPoints
			);
	};
	
//=====================================================================================================================

//! Converter for MapInfo CoordSys strings.
class CLASSLIBEXPORT MAPINFO {
	public:
		//! Convert COORDREFSYS to MapInfo CoordSys string.
		//! @return TRUE if successful, FALSE if can't convert, < 0 if error
		static int ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			SIMPLE_ARRAY<char>& string					//!< MapInfo CorodSys string returned
			);

		//! Convert MapInfo CoordSys string to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const char *string,							//!< MapInfo CoordSys string
			COORDREFSYS& CoordRefSys					//!< Coordinate Reference System returned
			);

		//! Read COORDREFSYS and control points from TAB file.
		static ERRVALUE ReadTAB (
			const FILEPATH& filepath,
			COORDREFSYS& CoordRefSys,
			SIMPLE_ARRAY<CTRLPOINT3>& points
			);

	};


//=====================================================================================================================

//! Converter for Tydac Projection strings.
class CLASSLIBEXPORT TYDAC {
	public:
		//! Convert COORDREFSYS to Tydac Projection string.
		//! @return TRUE if successful, FALSE if can't convert, < 0 if error
		static int ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			SIMPLE_ARRAY<char>& string					//!< Tydac CorodSys string returned
			);

		//! Convert Tydac Projection string to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const char *string,							//!< Tydac Projection string
			COORDREFSYS& CoordRefSys					//!< Coordinate Reference System returned
			);

	};

//=====================================================================================================================

//! Converter for ERMapper coordinate reference definition.
class CLASSLIBEXPORT ERMAPPER {
	public:
		//! Convert COORDREFSYS to ERMapper strings.
		//! @return TRUE if successful, FALSE if can't convert, < 0 if error
		static int ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			SIMPLE_ARRAY<char>& DatumStr,				//!< ERMapper Datum string returned
			SIMPLE_ARRAY<char>& ProjectionStr,		//!< ERMapper Projection string returned
			SIMPLE_ARRAY<char>& CoordTypeStr			//!< ERMapper CoordinateType string returned
			);

		//! Convert ERMapper strings to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const char *DatumStr,						//!< ERMapper Datum string
			const char *ProjectionStr,					//!< ERMapper Projection string
			const char *CoordTypeStr,					//!< ERMapper CoordinateType string
			COORDREFSYS& CoordRefSys					//!< Coordinate Reference System returned
			);
	};

//=====================================================================================================================

//! Converter for GeoTIFF coordinate reference definition.
class CLASSLIBEXPORT GEOTIFF {
	public:

		//! Reader for accessing GeoTIFF key values.
		class CLASSLIBEXPORT READER {
			public:
				//! Get floating-point value for key.
				//! @return true if key found, false if not.
				virtual bool v_GetKeyValue (UINT16 key, double& value) const = 0;

				//! Get integer value for key.
				//! @return true if key found, false if not.
				virtual bool v_GetKeyValue (UINT16 key, UINT16& value) const = 0;
			};

		class CLASSLIBEXPORT READER_MEMORY : public READER {
			public:
				READER_MEMORY ();
				virtual ~READER_MEMORY ();
				//! Initialize reader from memory records.
				ERRVALUE Initialize (
					const UINT16 *GeoKeyDirectoryBuffer,	//!< Buffer holding GeoKeyDirectory record
					int GeoKeyDirectoryNumItems,				//!< Number of values in buffer
					const double *GeoDoubleParmsBuffer,		//!< Buffer holding GeoDoubleParms record, 0 if unused
					int GeoDoubleParmsNumItems					//!< Number of values in buffer, 0 if none
					);
			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				struct GEOKEY {
					UINT16 m_key;
					UINT16 m_location;
					UINT16 m_count;
					UINT16 m_ValueOffset;
					};
				SIMPLE_ARRAY<GEOKEY> m_GeoKeys;
				SIMPLE_ARRAY<double> m_DoubleParms;
				virtual bool v_GetKeyValue (UINT16 key, double& value) const;
				virtual bool v_GetKeyValue (UINT16 key, UINT16& value) const;
				#endif
			};

		//! Writer for storing GeoTIFF key values.
		class WRITER {
			public:
				//! Set floating-point value for key.
				virtual ERRVALUE v_SetKeyValue (UINT16 key, double value) = 0;

				//! Set integer value for key.
				virtual ERRVALUE v_SetKeyValue (UINT16 key, UINT16 value) = 0;

				//! Set string value for key.
				virtual ERRVALUE v_SetKeyValue (UINT16 key, const char *string) = 0;
			};

		class CLASSLIBEXPORT WRITER_MEMORY : public WRITER {
			public:
				WRITER_MEMORY ();
				virtual ~WRITER_MEMORY ();
				const double* GetDoubleParmsBuffer (
					) const { return (m_DoubleParms); }
				const UINT16* GetGeoKeyDirectoryBuffer (
					) const { return (reinterpret_cast<const UINT16*>(&m_GeoKeys[0])); }
				UINT16 GetNumDoubleParms (
					) const { return (static_cast<UINT16>(m_DoubleParms.GetNumItems())); }
				UINT16 GetNumGeoKeys (
					) const { return (static_cast<UINT16>(m_GeoKeys.GetNumItems())); }
			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				struct GEOKEY {
					UINT16 m_key;
					UINT16 m_location;
					UINT16 m_count;
					UINT16 m_ValueOffset;
					};
				SIMPLE_ARRAY<GEOKEY> m_GeoKeys;
				SIMPLE_ARRAY<double> m_DoubleParms;

				virtual ERRVALUE v_SetKeyValue (UINT16 key, double value);
				virtual ERRVALUE v_SetKeyValue (UINT16 key, UINT16 value);
				virtual ERRVALUE v_SetKeyValue (UINT16 key, const char *string);
				#endif
			};

		//! Convert COORDREFSYS to GeoTIFF.
		//! @return TRUE if successful, FALSE if can't convert, < 0 if error
		static int ConvertFromCRS (
			const COORDREFSYS& CoordRefSys,			//!< Coordinate Reference System to convert
			WRITER& writer,								//!< Writer to store GeoTIFF key values
			bool WriteUserDefinedCRS,					//!< Whether to write CRS if user-defined
			bool& WroteProjAsGeographic				//!< Wrote unsupported projection as geographic
			);

		//! Convert GeoTIFF to COORDREFSYS.
		static ERRVALUE ConvertToCRS (
			const READER& reader,						//!< Reader to access GeoTIFF key values
			COORDREFSYS& CoordRefSys,					//!< Coordinate Reference System returned
			CODESPACE codespace = CODESPACE_EPSG	//!< Codespace override
			);
	};


//=====================================================================================================================

//! Attempt to read georeference for image.
//! @return TRUE if successful, FALSE if no georeference found, < 0 if error
int LIBEXPORT ReadImageGeoreference (
	const FILEPATH& imgfilepath,						//!< Path to image file
	SPATREF::COORDREFSYS& CoordRefSys,				//!< Coordinate reference system returned
	TRANS2D_AFFINE& TransAffine,						//!< Affine transformation returned if used
	SIMPLE_ARRAY<CTRLPOINT3>& CtrlPoints			//!< Array of control points returned if uses control points
	);
	
//=====================================================================================================================

};	// End namespace SPATREF

#undef	CLASSLIBEXPORT
#undef	LIBEXPORT

#endif	// INC_MI32_SPATREFEXT_H
