 /**
 *	\file mi32/watermdl.h
 *	\brief Definitions of WATERSHED class
 *
 * \if NODOC
 * $Id: watermdl.h_v 1.65 2011/11/08 14:13:27 mju Exp $
 *
 * $Log: watermdl.h_v $
 * Revision 1.65  2011/11/08 14:13:27  mju
 * Remove redundant hdrs.
 *
 * Revision 1.64  2011/09/06 21:22:42  mju
 * Exclude macros from dox.
 *
 * Revision 1.63  2007/11/23 21:26:32  vdronov
 * added elevation segmentation
 *
 * Revision 1.62  2007/11/06 16:59:23  vdronov
 * make db tables for segmented flowpath
 *
 * Revision 1.61  2007/10/18 22:32:35  vdronov
 * added exponential segmentation
 *
 * Revision 1.60  2007/10/17 19:24:21  vdronov
 * added segmentation
 *
 * Revision 1.59  2007/08/24 17:44:36  scowan
 * Updated changes to micache class.
 *
 * Revision 1.56  2007/08/02 18:56:24  vdronov
 * added use of micache
 *
 * Revision 1.54  2007/05/04 20:02:12  vdronov
 * added geomorphic attributes
 *
 * Revision 1.53  2007/04/20 14:56:12  vdronov
 * combine mask with nullmask at begining of processing
 *
 * Revision 1.51  2007/02/14 23:04:22  vdronov
 * added spare space in POLYGON::RECORD
 *
 * Revision 1.33  2006/06/20 22:17:06  vdronov
 * added some optimization
 *
 * Revision 1.32  2006/06/15 19:42:41  vdronov
 * redone ComputeExtrema
 *
 * Revision 1.31  2006/02/02 23:39:41  vdronov
 * added ComputeBoundaries
 *
 * Revision 1.30  2006/02/01 21:34:40  vdronov
 * added temporary binary raster
 *
 * Revision 1.23  2005/12/23 23:32:30  vdronov
 * use database instead of memory
 *
 * Revision 1.17  2000/09/22 17:57:49  vdronov
 * stream order added
 *
 * Revision 1.11  2000/08/15 21:18:08  vdronov
 * added filling depressions by different criteria
 *
 * Revision 1.10  2000/08/09 17:22:15  vdronov
 * added mask capabilities
 *
 * Revision 1.9  2000/08/02 21:12:02  vdronov
 * added databases attributes
 *
 * Revision 1.1  2000/06/08 15:44:07  vdronov
 * Initial revision
 * \endif
**/

#ifndef	INC_MI32_WATERMDL_H
#define	INC_MI32_WATERMDL_H

#ifndef INC_RVC_DBTABLE_H
#include <rvc/dbtable.h>
#endif

#ifndef INC_RVC_RASTER_H
#include <rvc/raster.h>
#endif

#ifndef INC_RVC_NULLMASK_H
#include <rvc/nullmask.h>
#endif

#ifndef INC_RVC_ARRAY_H
#include <rvc/array.h>
#endif

#ifndef INC_RVC_VECTOR_H
#include <rvc/vector.h>
#endif

#ifndef INC_MI32_MICACHE_H
#include <mi32/micache.h>
#endif

#include <map>


#ifndef GENERATING_DOXYGEN_OUTPUT
template <typename _TT> class ARRAYCACHE : public MICACHEIO {
public:
	ARRAYCACHE () {}
	virtual ~ARRAYCACHE () {}

	ERRVALUE Init (
		const INT32 NumCache, 
		RVC::ARRAY* Array
		) {
		ERRVALUE err = m_Cache.Create(NumCache, sizeof(_TT), *this);
		if (err < 0) return err;
		m_Array = Array;
		return 0;
		}

	ERRVALUE Select (
		const INT32 itemid,
		MICACHE::ITEMINFO& iteminfo
		) { return (m_Cache.Select(itemid, iteminfo, MICACHE::SELECTFLAG_None, m_Array)); }

	ERRVALUE Read (
		const INT32 itemid,
		_TT& item
		) { return (m_Cache.Read(itemid, 0, sizeof(_TT), &item, m_Array)); }


	void Clear (
		) {
		m_Cache.Flush();
		m_Cache.Free();
		m_Array = 0;
		return;
		}

private:

	SIMPLE_ARRAY<_TT> m_Buffer;

	RVC::ARRAY* m_Array;
	MICACHE m_Cache;

	ERRVALUE CacheLoad (
		MICACHE::ITEMINFO &iteminfo
		) { 
		RVC::ARRAY* Array = (RVC::ARRAY*)iteminfo.GetUserDataPtr();
		return Array->Read(iteminfo.GetItemID(), iteminfo.GetMemPtr()); 
		}

	ERRVALUE CacheSave (
		MICACHE::ITEMINFO& iteminfo
		) { 
		RVC::ARRAY* Array = (RVC::ARRAY*)iteminfo.GetUserDataPtr();
		return Array->Write(iteminfo.GetItemID(), iteminfo.GetMemPtr()); 
		}

	};

#endif // GENERATING_DOXYGEN_OUTPUT



//! Watershed analysys based on DEM.
//!
//! This class is designed to provide computation of different hydrological features.
//! It is based on well-known D8 algorithm. When you are working with
//! this class you have to expect 3 states in calculation:
//!	-	Initial
//!	-	Depressions exist and calculated
//!	-	All depressions were filled out or didn't exist.
//! Status of current state you can get with method GetStatus.
//!
//!	1. Initial state
//!		Object has this status immediately after initialisation or after calling any method
//!		that working with mask. It is preferably use masking methods after initialisation.
//!		Masking methods are important for selection area of interest in DEM. Value 0 in mask
//!		has the same meaning as null value in elevation raster.
//!		By calling Compute method with different combination of flags process jumps into other
//!		state depends on flags.
//!
//!	2.	State with depressions
//!		In this state it is available to research watershed system: how many watersheds with 
//!		depressions, how those watersheds with depressions connected to each others, 
//!		where water pours out and in for every watershed with or without depressions. 
//!		It is possible to fill out some selected set of watershed with depressions and 
//!		see combining watershed as result of this filling.
//!		It can be done by calling FillDepressions method with different combination of flags 
//!		and/or selected set of watershed ID. When all depressions will be filling out process
//!		is transferred in next state.
//!
//!	3.	State without depressions
//!		When all depressions are filled out or haven't existed it is possible to calculated those 
//!		hydrological features as flow paths, basins, ridges. It can be done by calling 
//!		ComputeElements with different combinations of flags and/or selected set of seed points.
//!		If selected set is present user-defined flow paths and basins will be computed, otherwise
//!		process will calculate flow paths, basins and ridges depends on automaticaly-calculated 
//!		seeds points defined by parameters BASIN, BRANCH, INLET, OUTLET, VALLEYSEPARATION.
 
class WATERSHED {

public:

	// ENUMERATIONS

	enum STATUS {										//! Status of watershed process 
		STATUS_NotInitialized = -1,			  	//!<	Not initialized state
		STATUS_Initial = 0,							//!<	Initial state
		STATUS_HasDepressions = 1,					//!<	State with depressions
		STATUS_DepressionsFilled = 2				//!<	State without depressions 
		};

	enum FLAGS {											//! Flags that can change flow of watershed process
		FLAGS_None							= 0x0000,	//!< No action flag
		FLAGS_FillDepressions		   = 0x0001,	//!< Fill depressions (not for state without depressions)
		FLAGS_FillUpperDepressions		= 0x0002,	//!<	Fill upper depressions that pour to selected set 
																//!		of depressions (only for state with depressions)
		FLAGS_FillLowerDepressions		= 0x0004,	//!< Fill lower depressions that pour from selected set
																//!	of depressions (only for state with depressions)
		FLAGS_FillDoubleDepressions	= 0x0008,	//!< Fill only depressions that pour only to each other.
																//!		(only for state with depressions)
		FLAGS_FillDepressionsMask		= 0x000F,	//!< Mask for depressions actions
		FLAGS_FlowPath						= 0x0010,	//!< Calculate flow paths in state without depressions
		FLAGS_Basin							= 0x0020,	//!< Calculate basins in state without depressions
		FLAGS_Ridge							= 0x0040,	//!< Calculate ridges in state without depressions
		FLAGS_CatchmentArea				= 0x0100,	//!< Calculate specific catchment area
		FLAGS_TopographicIndex			= 0x0200,	//!< Calculate compound topographic index
		FLAGS_MaximumUpFlow				= 0x0400,	//!< Calculate maximum up stream flow distance
		FLAGS_DownFlow						= 0x0800,	//!< Calculate down stream flow distance
		FLAGS_BasinAttribute				= 0x1000,	//!< Calculate geomorphic attributes for basins
		FLAGS_WatershedAttribute		= 0x2000,	//!< Calculate geomorphic attributes for watersheds
		FLAGS_ObjectMask					= 0x3F70		//!< Mask for calculations actions
		};

	enum DIRECTION {									//! Meaning of flow direction raster cell value 
		DIRECTION_Null = 0,							//!< No directions defined
		DIRECTION_TopRight = 1,
		DIRECTION_Right = 2,
		DIRECTION_BottomRight = 3,
		DIRECTION_Bottom = 4,
		DIRECTION_BottomLeft = 5,
		DIRECTION_Left = 6,
		DIRECTION_TopLeft = 7,
		DIRECTION_Top = 8,
		DIRECTION_DepressionFlat = 12,			//!Cell on depressions minimum flat
		DIRECTION_Minimum = 13,						//!< Cell on local minimum
		DIRECTION_Multiple = 14,					//!< Cell with multiple possible direction of flow 
		DIRECTION_Flat = 15							//!< Cell on flat area
		};

	enum EXTREMA {										//! Meaning of flow direction raster cell value 
		EXTREMA_Null = 0,								//!< Cell with null value 
		EXTREMA_Normal = 1,							//!<	Normal cell without extrema
		EXTREMA_Maximum = 2,							//!< Cell with local maximum
		EXTREMA_Minimum = 3,							//!< Cell with local minimum
		EXTREMA_Flat = 4,								//!< Cell on flat area
		EXTREMA_FlatExit = 5,						//!< Cell on flat area is gate to exit from flat area
		EXTREMA_FlatEntry = 6,						//!< Cell on flat area is gate to entry to flat area
		EXTREMA_Saddle = 7							//!< Saddle cell
		};

	enum OUTPUT {										//! Output objects produced by watershed process
		OUTPUT_None = -1,
		OUTPUT_First = 0,
		OUTPUT_RasterAdjusted = 0,					//!< Object that represents modified version of original DEM
															//!	due by depressions filling
		OUTPUT_RasterExtrema = 1,					//!< 4-bits raster represents extrema values on original DEM
		OUTPUT_RasterFlowDirection = 2,			//!< 4-bits raster represents flow directionon adjusted DEM
		OUTPUT_RasterAccumulation = 3,			//!< 32-bits signed raster represents flow accumulation in
															//!	number of cells in adjusted DEM
		OUTPUT_RasterWatershed = 4,				//!< 32-bits signed raster represents watershed area. 
															//!	Cells	belong to the same watershed if they have 
															//!	the same cell value 
		OUTPUT_RasterCatchmentArea = 5,			//!< 32-bits floating point raster represents specific catchment area values
		OUTPUT_RasterTopographicIndex = 6,		//!< 32-bits floating point raster represents compound flow directionon adjusted DEM
		OUTPUT_RasterMaximumUpFlow = 7,			//!< 32-bits floating point raster represents maximum length up stream flow
		OUTPUT_RasterDownFlow = 8,				   //!< 32-bits floating point raster represents length down stream flow
		OUTPUT_VectorWatershed = 9,				//!< 3D vector object is vectorization form of watershed raster
															//!   and 3D vector points as pour points if any 
		OUTPUT_VectorFlowPath = 10,				//!< 3D vector object represents flow paths on DEM for
															//!   automatically calaculated seed points
		OUTPUT_VectorBasin = 11,					//!< 3D vector object represents basins on DEM for
															//!   automatically calaculated seed points
		OUTPUT_VectorRidge = 12,					//!< 3D vector object represents ridges on DEM
		OUTPUT_VectorUserFlowPath = 13,			//!< 3D vector object represents flow paths on DEM for
															//!	user-defined seed points
		OUTPUT_VectorUserBasin = 14,				//!< 3D vector object represents basins on DEM for
															//!	user-defined seed points
		OUTPUT_VectorSegmentedFlowPath = 15,
		OUTPUT_Count = 16
		};

	enum SEGMENTATIONPARAMETER {
		SEGMENTATIONPARAMETER_None = 0,
		SEGMENTATIONPARAMETER_FlowAccumulation = 1,
		SEGMENTATIONPARAMETER_Elevation = 2
		};

	enum SEGMENTATIONINTERVAL {
		SEGMENTATIONINTERVAL_Equal = 0,
		SEGMENTATIONINTERVAL_Unequal = 1,
		SEGMENTATIONINTERVAL_Exponential = 2
		};

	class ATTRIBUTE;

	// NESTED CLASSES
	//! This is temporary class 
	class POURCANDIDATE {

	public:

		// MEMBERS

		// CONSTRUCTION / DESTRUCTION
		POURCANDIDATE (								//! Default constructor
			) {
			}

		POURCANDIDATE (								//! Constructor by location and neighbors watershed 
			const DPOINT3D& Point,					//!< Location of pour point                         
			const INT32 LeftWatershed,				//!< Left watershed                                 
			const INT32 RightWatershed				//!< Right watershed                                
			) :
			m_Record(Point, LeftWatershed, RightWatershed)
			{
			}

		POURCANDIDATE (											//! Copy constructor 		
			const POURCANDIDATE &rhs
			) :
			m_Record(rhs.m_Record)
			{
			}

		~POURCANDIDATE (											//! Destructor
			) {
			}
 
		// OPERATORS
 
		POURCANDIDATE& operator= (								//! Assignment operator 
			const POURCANDIDATE& rhs
			) {
			if (this != &rhs) {
				m_Record = rhs.m_Record;
				}
			return (*this);
			}

		void Init (										//!  
			const DPOINT3D& Point,					//!< Location of pour point                         
			const INT32 LeftWatershed,				//!< Left watershed                                 
			const INT32 RightWatershed				//!< Right watershed                                
			) {
			m_Record.m_Point = Point;
			m_Record.m_LeftWatershed = LeftWatershed;
			m_Record.m_RightWatershed = RightWatershed;
			return;
			}

		// METHODS

		//! Get left watershed ID. It is usally watershed where water pours from.
		//!
		//! @return Watershed ID
		const INT32 GetLeftWatershed (
			) const {
			return m_Record.m_LeftWatershed;
			}

		//! Get left watershed ID. It is usally watershed where water pours to.
		//!
		//! @return Watershed ID or -1 if on boundary of DEM
		const INT32 GetRightWatershed (
			) const {
			return m_Record.m_RightWatershed;
			}

		//! Get location of pour point as DPOINT3D in Original DEM coordinates.
		//!
		//! @return DPOINT3D
		const DPOINT3D& GetPoint (
			) const {
			return m_Record.m_Point;
			}

	private:

		friend class WATERSHED::ATTRIBUTE;

		struct RECORD {

			DPOINT3D m_Point;								//!< Pour point coordinate
			INT32 m_LeftWatershed;						//!< Left watershed
			INT32 m_RightWatershed;						//!< Right watershed

			RECORD (
				) :
				m_Point(),
				m_LeftWatershed(-1),
				m_RightWatershed(-1)
				{
				};

			RECORD (
				const DPOINT3D& Point,		
				const INT32 LeftWatershed,
				const INT32 RightWatershed
				) :
				m_Point(Point),
				m_LeftWatershed(LeftWatershed),
				m_RightWatershed(RightWatershed)
				{
				};

			};
		// MEMBERS

			RECORD m_Record;

		};

	//! This class is designed to provide pouring hydrological feature. It defines location
	//! on boundary between two watersheds where water pours only from one to another or 
	//!	in both directions. It is base class and provides get methods only. 
	class POUR {

	public:

		// ENUMERATIONS

		enum TYPE {										//! Type of pour point
			TYPE_None = 0,								//!<	Point is not pour point
			TYPE_Single = 1,							//!<	Pour point pours water only in specified direction
			TYPE_Double = 2							//!< Pour point pours water in both directions
			};


		// MEMBERS

		// CONSTRUCTION / DESTRUCTION

		POUR (											//! Default constructor
			) {
			}

		POUR (										//! Constructor by location and neighbors watershed 
			const TYPE type,						//!< type of pour point                         
			const POURCANDIDATE& pourcandidate,		//!< pour candidate                                 
			const INT32 id							//!< vector point id                                
			) :
			m_Record(type, pourcandidate, id)
			{
			}

		POUR (											//! Copy constructor 		
			const POUR &rhs
			) :
			m_Record(rhs.m_Record)
			{
			}

		~POUR (								//! Destructor
			) {
			}
 
		// OPERATORS
 
		POUR& operator= (								//! Assignment operator 
			const POUR& rhs
			) {
			if (this != &rhs) {
				m_Record = rhs.m_Record;
				}
			return (*this);
			}

		void Init (											//!  
			const TYPE type,								//!< type of pour point                         
			const POURCANDIDATE& pourcandidate,		//!< pour candidate                                 
			const INT32 id									//!< vector point id                                
			) {
			m_Record.m_Type = type;
			m_Record.m_Point = pourcandidate.GetPoint();
			m_Record.m_LeftWatershed = pourcandidate.GetLeftWatershed();
			m_Record.m_RightWatershed = pourcandidate.GetRightWatershed();
			m_Record.m_ElementIDOfVectorPoint = id;
			}

		// METHODS

		//! Get left watershed ID. It is usally watershed where water pours from.
		//!
		//! @return Watershed ID
		const INT32 GetLeftWatershed (
			) const {
			return m_Record.m_LeftWatershed;
			}

		//! Get left watershed ID. It is usally watershed where water pours to.
		//!
		//! @return Watershed ID or -1 if on boundary of DEM
		const INT32 GetRightWatershed (
			) const {
			return m_Record.m_RightWatershed;
			}

		//! Get location of pour point as DPOINT3D in Original DEM coordinates.
		//!
		//! @return DPOINT3D
		const DPOINT3D& GetPoint (
			) const {
			return m_Record.m_Point;
			}

		//! Get type of pour point.
		//!
		//! @return TYPE
		const TYPE GetType (
			) const {
			return m_Record.m_Type;
			}

		//! Get element ID of vector point in VectorWatershed object.
		//!
		//! @return Element ID
		const INT32 GetElementIDOfVectorPoint (
			) const {
			return m_Record.m_ElementIDOfVectorPoint;
			}

	private:

		friend class WATERSHED::ATTRIBUTE;

		struct RECORD {

			TYPE m_Type;									//!< Pour point type
			DPOINT3D m_Point;								//!< Pour point coordinate
			INT32 m_LeftWatershed;						//!< Left watershed
			INT32 m_RightWatershed;						//!< Right watershed
			INT32 m_ElementIDOfVectorPoint;			//!< Element ID of pour point as vector point or -1

			RECORD (
				) :
				m_Type(TYPE_None),
				m_Point(),
				m_LeftWatershed(-1),
				m_RightWatershed(-1),
				m_ElementIDOfVectorPoint(-1)
				{
				};

			RECORD (
				const TYPE type,								//!< type of pour point                         
				const POURCANDIDATE& pourcandidate,		//!< pour candidate                                 
				const INT32 id									//!< vector point id                                
				) :
				m_Type(type),
				m_Point(pourcandidate.GetPoint()),
				m_LeftWatershed(pourcandidate.GetLeftWatershed()),
				m_RightWatershed(pourcandidate.GetRightWatershed()),
				m_ElementIDOfVectorPoint(id)
				{
				};

			};
		// MEMBERS

		RECORD m_Record;

		};


	//! This class is designed to provide 2dimensional hydrological feature. It defines set of neighbors
	//! cell on DEM called watershed. Watershed might have a depression where water is collected from all
	//! cells in watershed or might pour water out of DEM.	Watershed in vector form is collection of 1
	//! or more vector polygons as side effect of D8 algorithm. If watershed has depression it has also
	//! at least one lower pour point where water will flow out when watershed is overflown. Watershed 
	//! might also have some upper pour points where water from neighbors wateshed will flow in.
	//! It is base class and provides get methods only.
	class POLYGON {

	public:

		// MEMBERS

		// CONSTRUCTION / DESTRUCTION

		POLYGON (										//! Constructor 		
			) {
			}

		POLYGON (										//! Copy constructor 		
			const POLYGON &rhs
			) :
			m_Record(rhs.m_Record),
			m_VectorPolygons(rhs.m_VectorPolygons),
			m_UpperPours(rhs.m_UpperPours),
  			m_LowerPours(rhs.m_LowerPours)
			{
			}

		~POLYGON (							//! Destructor
			) {
			m_VectorPolygons.Clear();
			m_UpperPours.Clear();
			m_LowerPours.Clear();
			}

		// OPERATORS

		POLYGON& operator= (							//! Assignment operator 
			const POLYGON& rhs
			) {
			if (this != &rhs) {
				m_Record = rhs.m_Record;
				m_VectorPolygons = rhs.m_VectorPolygons;
				m_UpperPours = rhs.m_UpperPours;
	  			m_LowerPours = rhs.m_LowerPours;
				}
			return (*this);
			}


		// METHODS

		//! Defines if watershed has a depression
		//!
		//! @return true if has or false - otherwise
		const bool HasDepression (
			) const {
			return (m_Record.m_Z > m_Record.m_MinZ);
			}

		//! Get area of watershed 
		//!
		//! @return Area of watershed 
		const double GetArea (
			) const {
			return m_Record.m_Area;
			}

		//! Get perimeter of watershed boundary
		//!
		//! @return Perimeter of watershed boundary
		const double GetPerimeter (
			) const {
			return m_Record.m_Perimeter;
			}

		//! Get depression area of watershed with depression 
		//!
		//! @return Depression area of watershed with depression
		const double GetDepressionArea (
			) const {
			return m_Record.m_DepressionArea;
			}

		//! Get depression volume of watershed with depression
		//!
		//! @return Depression volume of watershed with depression
		const double GetDepressionVolume (
			) const {
			return m_Record.m_DepressionVolume;
			}

		//! Get minimum z value on watershed boundary in meters
		//!
		//! @return Minimum z value on watershed boundary in meters
		const double GetZ (
			) const {
			return m_Record.m_Z;
			}

		//! Get minimum raw z value on watershed boundary in original units
		//!
		//! @return Minimum raw z value on watershed boundary in original units
		const double GetRawZ (
			) const {
			return m_Record.m_RawZ;
			}

		//! Get minimum z value inside watershed
		//!
		//! @return Minimum z value inside watershed
		const double GetMinZ (
			) const {
			return m_Record.m_MinZ;
			}

		//! Get maximum z value inside watershed
		//!
		//! @return Maximum z value inside watershed
		const double GetMaxZ (
			) const {
			return m_Record.m_MaxZ;
			}

		//! Get number of vector polygons in watershed
		//!
		//! @return Number of vector polygons
		const INT32 GetNumberOfVectorPolygons (
			) const {
			return m_VectorPolygons.GetNumItems();
			}

		//! Get element ID of vector polygon in watershed by index
		//!
		//! @return Element ID of vector polygon or -1 if index is invalid
		const INT32 GetIDOfVectorPolygon (
			const UINT32 index						//!<	Index of element ID of vector polygon
			) const {
			if (index < (UINT32)m_VectorPolygons.GetNumItems()) return m_VectorPolygons[(int)index];
			return -1;
			}

		//! Get number of lower pour points in watershed
		//!
		//! @return Number of lower pour points
		const UINT32 GetNumberOfLowerPours (
			) const {
			return (m_LowerPours.GetNumItems());
			}

		//! Get ID of lower pour point in watershed by index
		//!
		//! @return ID of lower pour point or -1 if index is invalid
		const INT32 GetIDOfLowerPour (
			const UINT32 index						//!<	Index of pour point ID
			) const {
			if (index < (UINT32)m_LowerPours.GetNumItems()) return m_LowerPours[(int)index];
			return -1;
			}

		//! Get number of upper pour points in watershed
		//!
		//! @return Number of upper pour points
		const INT32 GetNumberOfUpperPours (
			) const {
			return (m_UpperPours.GetNumItems());
			}

		//! Get ID of upper pour point in watershed by index
		//!
		//! @return ID of upper pour point or -1 if index is invalid
		const INT32 GetIDOfUpperPour (
			const UINT32 index						//!<	Index of pour point ID
			) const {
			if (index < (UINT32)m_UpperPours.GetNumItems()) return m_UpperPours[(int)index];
			return -1;
			}

	private:

		friend class WATERSHED::ATTRIBUTE;

		struct RECORD {
			double m_DepressionArea;					//!< Area of depression if exist
			double m_DepressionVolume;					//!< Volume of depressions if exist
			double m_Area;									//!< Area of watershed in squared meters
 			double m_Perimeter;							//!< Perimeter of watershed in meters
			double m_MinZ;									//!< Minimum z value inside watershed in meters
			double m_Z;										//!< Minimum z value on boundary of watershed in meters
			double m_RawZ;									//!< Minimum z value on boundary of watershed in original
			double m_MaxZ;									//!< Maximum z value inside watershed in meters
			INT32 m_NextVector;
			INT32 m_NextUpper;
			INT32 m_NextLower;
			INT32 m_Spare;

			RECORD (
				) :
				m_DepressionArea(0.0),		
				m_DepressionVolume(0.0),	
				m_Area(0.0),
 				m_Perimeter(0.0),			
				m_MinZ(DBL_MAX),
				m_Z(DBL_MAX),
				m_RawZ(DBL_MAX),
				m_MaxZ(-DBL_MAX),
				m_NextVector(-1),
				m_NextUpper(-1),
				m_NextLower(-1),
				m_Spare(0)
				{
				};

			};

		struct NEXTRECORD {
			INT32 m_ID;
			INT32 m_Next;

			NEXTRECORD (
				) :
				m_ID(-1),
				m_Next(-1)
				{
				};
			};

		// MEMBERS

		RECORD m_Record;

		SIMPLE_ARRAY<INT32> m_VectorPolygons;	//!< Container of vector polygons elements ID
		SIMPLE_ARRAY<INT32> m_UpperPours;		//!< Container of upper pour points ID
		SIMPLE_ARRAY<INT32> m_LowerPours;		//!< Container of lower pour points ID

		};

	//! This class is designed to provide hydrological features. It computes and 
	//! defines watersheds, depressions and pour points.
	class ATTRIBUTE {

	public:

		// CONSTRUCTION / DESTRUCTION

		ATTRIBUTE (										//! Constructor by number of polygons in Watershed raster
			const UINT32 NumberRasterPolygons	//!<	Number of polygons in Watershed raster
			);

		~ATTRIBUTE (									//! Destructor
			);

		//! Check minimum and maximum z value inside watershed for given elevation
		ERRVALUE CheckInternalElevation (
			const INT32 watershed,					//!< Watershed ID
			const double elevation					//!< Elevation for checking
			);

		//! Check minimum and maximum z value inside watershed for given elevation and add area
		ERRVALUE CheckElevationAndAddArea (
			const INT32 watershed,					//!< Watershed ID
			const double elevation,					//!< Elevation for checking
			const double area							//!< area
			);

		//! Check minimum z value on boundary of watershed for given elevation
		ERRVALUE CheckBoundaryElevation (
			const INT32 watershed,					//!< Watershed ID
			const double z,							//!< Elevation for checking
			const double rawZ						//!< Raw Elevation for checking
			);

		//! Initialization of attributes
		ERRVALUE Init (										
			const INT32 NumPolys,					//!< Number of vector polygons
			const INT32 NumLines,					//!< Number of vector lines
			const INT32 NumNodes					//!< Number of vector nodes
			);

		//! Add link between watershed and vector polygon 
		ERRVALUE AddLink (
			const INT32 watershed,					//!< Watershed ID
			const INT32 VectorPolygon				//!< Vector polygon element ID
			);

		//! Add area to watershed 
		ERRVALUE AddArea (
			const INT32 watershed,					//!< Watershed ID
			const double area						//!< area
			);

		//! Add depression area to watershed 
		ERRVALUE AddDepressionArea (
			const INT32 watershed,					//!< Watershed ID
			const double area						//!< area
			);

		//! Add depression volume to watershed 
		ERRVALUE AddDepressionVolume (
			const INT32 watershed,					//!< Watershed ID
			const double volume						//!< volume
			);

		//! Add depression area and volume to watershed 
		ERRVALUE AddDepressionData (
			const INT32 watershed,					//!< Watershed ID
			const double area,						//!< area
			const double volume						//!< volume
			);

		//! Add perimeter to watershed 
		ERRVALUE AddPerimeter (
			const INT32 watershed,					//!< Watershed ID
			const double perimeter					//!< perimeter
			);

		//! Add given vector polygon element ID to collection of vector polygons for watershed
		ERRVALUE AddVectorPolygon (
			const INT32 watershed,					//!< Watershed ID
			const INT32 VectorPolygon				//!< Vector polygon element ID
			);

		//! Add given pour point ID to collection of lower pour points for watershed
		ERRVALUE AddLowerPour (
			const INT32 watershed,					//!< Watershed ID
			const INT32 LowerPour					//!< Pour point ID
			);

		//! Add given pour point ID to collection of upper pour points for watershed
		ERRVALUE AddUpperPour (
			const INT32 watershed,					//!< Watershed ID
			const INT32 UpperPour					//!< Pour point ID
			);

		//! Add pour to collection of pours 
		ERRVALUE AddPour (
			const POUR& pour							//!<	Pour object
			);

		//! Add pour candidate to collection of pour candidates 
		ERRVALUE AddPourCandidate (
			const POURCANDIDATE& pourcandidate							//!<	Pour candidate object
			);

		//! Define real pour points in collection of pour candidates and save them in given vector object
		//!
		//! @return ERRVALUE
		ERRVALUE DefinePours (
			RVC::VECTOR& vector
			);

		//! Rise type 
		void RisePourType (
			POUR::TYPE& type
			) {
			switch (type) {
				case POUR::TYPE_None: type = POUR::TYPE_Single; break;
				case POUR::TYPE_Single: type = POUR::TYPE_Double; break;
				}
			return;
			}

		//! Save pour point as vector point
		//!
		//! @return ERRVALUE
		ERRVALUE SaveAsVectorPoint (
			RVC::VECTOR& vector,
			const DPOINT3D& point,
			const INT32 id							//!< Element Id of vector point to be saved
			);

		//! Defines if watershed system has a depressions
		//!
		//! @return true if has or false - otherwise
		const bool HasDepressions (
			) const {
			return (m_NumberOfDepressions > 0);
			};

		//! Defines if watershed polygon has a depressions
		//!
		//! @return true if has or false - otherwise
		const bool HasDepression (
			const INT32 watershed					//!< Watershed ID
			);

		//! Returns watershed minimum z on boundary
		//!
		//! @return watershed z
		const double GetWatershedZ (
			const INT32 watershed					//!< Watershed ID
			);

		//! Get number of watershed polygons with depression in watershed system
		//!
		//! @return Number of watershed polygons with depression
		const INT32 GetNumberOfWatershedPolygonsWithDepression (
			) const {
			return m_NumberOfDepressions;
			}

		//! Get number of watershed polygons in watershed system
		//!
		//! @return Number of watershed polygons
		const INT32 GetNumberOfWatershedPolygons (
			) const {
			return m_PolygonArray.GetNumItems();
			};

		//! Get watershed polygon object in watershed system by index
		//!
		//! @return true if valid index
		bool GetWatershedPolygon (
			const INT32 index,							//!< Watershed polygon ID
			POLYGON& polygon
			);

		//! Get number of vector polygons in watershed system
		//!
		//! @return Number of vector polygons
		const INT32 GetNumberOfVectorPolygons (
			) const {
			return m_VectorArray.GetNumItems();
			};

		//! Get watershed polygon ID in watershed system for given vector polygon element ID 
		//! that belongs to this watershed
		//!
		//! @return Watershed polygon ID
		const INT32 GetIDOfWatershedPolygonForVectorPolygon (
			const INT32 index							//!< Vector polygon element ID
			);

		//! Get number of pour points in watershed system
		//!
		//! @return Number of pour points
		const INT32 GetNumberOfPours (
			) const {
			return  (m_PourArray.GetNumItems());
			};

		//! Get pour object in watershed system by index
		//!
		//! @return true if valid index
		bool GetPour (
			const INT32 index,							//!<Pour point ID
			POUR& pour
			);

		//! Get number of pour candidate points in watershed system
		//!
		//! @return Number of pour candidate points
		const INT32 GetNumberOfPourCandidates (
			) const {
			return  (m_PourCandidateArray.GetNumItems());
			};

		//! Get pour candidate object in watershed system by index
		//!
		//! @return true if valid index
		bool GetPourCandidate (
			const INT32 index,
			POURCANDIDATE& pourcandidate
			);

		//! Get number of double pour points in watershed system
		//!
		//! @return Number of double pour points
		const INT32 GetNumberOfDoublePours (
			) const {
			return m_NumberOfDoublePours;
			}


	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		struct LINKRECORD {
			INT32 m_Watershed;

			LINKRECORD (
				) :
				m_Watershed(-1)
				{
				};

			};


		// MEMBERS

		INT32 m_NumberOfDepressions;					//!< Number of depressions in watershed system
		INT32 m_NumberOfDoublePours;					//!< Number of double pours in watershed system

		RVC::OBJECT m_File;

		RVC::ARRAY m_PourCandidateArray;
		RVC::ARRAY m_PourArray;
		RVC::ARRAY m_PolygonArray;
		RVC::ARRAY m_VectorArray;
		RVC::ARRAY m_UpperPourArray;
		RVC::ARRAY m_LowerPourArray;
		RVC::ARRAY m_LinkArray;

		ARRAYCACHE<POURCANDIDATE::RECORD> m_PourCandidateCache;  // MICACHE
		ARRAYCACHE<POUR::RECORD> m_PourCache;
		ARRAYCACHE<POLYGON::RECORD> m_PolygonCache;
		ARRAYCACHE<POLYGON::NEXTRECORD> m_VectorCache;
		ARRAYCACHE<POLYGON::NEXTRECORD> m_UpperPourCache;
		ARRAYCACHE<POLYGON::NEXTRECORD> m_LowerPourCache;
		ARRAYCACHE<LINKRECORD> m_LinkCache;

		POURCANDIDATE::RECORD m_PourCandidateRecord;
		POUR::RECORD m_PourRecord;
		POLYGON::RECORD m_PolygonRecord;
		POLYGON::NEXTRECORD m_NextRecord;
		LINKRECORD m_LinkRecord;

		// CONSTRUCTION / DESTRUCTION

		ATTRIBUTE (										//! Default constructor (not implemented)
			);

		ATTRIBUTE (										//! Copy constructor (not implemented)
		const ATTRIBUTE &rhs
		);

		// OPERATORS

		ATTRIBUTE& operator= (						//! Assignment operator (not implemented)
			const ATTRIBUTE& rhs
			);

		ERRVALUE Make ();
		ERRVALUE Delete();
		ERRVALUE FillPolygonsTable(const UINT32 NumberOfRasterPolygons);
		ERRVALUE FillLinksTable(const UINT32 NumberOfVectorPolygons);

		#endif // GENERATING_DOXYGEN_OUTPUT

		};

	// TYPEDEFS

	// MEMBERS

	static const UINT32 DefaultOutlet;			//!< Default outlet value in pixels. Outlet value defines 
															//!  flow path that pours water out of boundary
															//!  by collecting it from area more than outlet value
	static const UINT32 DefaultInlet;			//!< Default inlet value in pixels. Inlet value defines
															//!  flow path than begins by collecting water from area
															//!  more than inlet value.
	static const UINT32 DefaultBranch;			//!< Default branch value in pixels. Branch value defines
															//!  branch flow path that collects water from area more
															//!  than branch value.
	static const UINT32 DefaultBasin;			//!< Default basin value in pixels. Basin value defines
															//!  basin with total area more than basin value.
	static const bool	DefaultValleySeparation;//!< Valley separation if true separates main flow stream
															//!  valley in areas from one branch merging to other.

	static const UINT32 MinimumOutlet;			//!< Minimum outlet value in pixels.
	static const UINT32 MinimumInlet;			//!< Minimum inlet value in pixels.
	static const UINT32 MinimumBranch;			//!< Minimum branch value in pixels.
	static const UINT32 MinimumBasin;			//!< Minimum basin value in pixels.


	static const UINT32 s_DefaultDepressionMaximumArea;									
	static const UINT32 s_DefaultDepressionMaximumNullArea;									
	static const double s_DefaultDepressionMaximumDepth;									
	static const UINT32 s_DefaultDepressionMinimumDepthArea;									
	static const bool s_DefaultPutNullCellAtUnfilledDepression;						

	static const UINT32 s_MinimumDepressionMaximumArea;									
	static const UINT32 s_MinimumDepressionMaximumNullArea;									
	static const double s_MinimumDepressionMaximumDepth;									
	static const UINT32 s_MinimumDepressionMinimumDepthArea;									

	static const UINT32 s_MaximumDepressionMaximumArea;									
	static const UINT32 s_MaximumDepressionMaximumNullArea;									
	static const double s_MaximumDepressionMaximumDepth;									
	static const UINT32 s_MaximumDepressionMinimumDepthArea;									
								 
	static UINT32 GetDefaultInlet (
		WATERSHED *watershed
		) {
		if (watershed == 0) return DefaultInlet;
		return watershed->GetDefaultInlet();
		}

	static UINT32 GetDefaultOutlet (
		WATERSHED *watershed
		) {
		if (watershed == 0) return DefaultOutlet;
		return watershed->GetDefaultOutlet();
		}

	static UINT32 GetDefaultBasin (
		WATERSHED *watershed
		) {
		if (watershed == 0) return DefaultBasin;
		return watershed->GetDefaultBasin();
		}

	static UINT32 GetDefaultBranch (
		WATERSHED *watershed
		) {
		if (watershed == 0) return DefaultBranch;
		return watershed->GetDefaultBranch();
		}

	// CONSTRUCTION / DESTRUCTION

	WATERSHED (											//! Constructor by RVC::OBJITEM
		const RVC::OBJITEM& objitem					
		);

	~WATERSHED (										//! Destructor
		);

	// OPERATORS

	// METHODS
			
	//!	Computes different hydrological features depends on flags
	//!
	//! @return ERRVALUE
	ERRVALUE Compute (
		const FLAGS flags								//!<	Flags (see FLAGS enum)
		);

	//!	Fills depressions in different ways depends on 
	//! selected set of watersheds with depressions and/or flags
	//!
	//! @return ERRVALUE
	ERRVALUE FillDepressions (
		const SIMPLE_ARRAY<INT32> &depressions,		//!<	Selected set of watersheds ID with depressions
		const FLAGS flags										//!< Flags (see FLAGS enum)
		);

	//!	Computes different hydrological elements as flow paths, basins, ridges
	//! depends on selected set of seed points and/or flags
	//!
	//! @return ERRVALUE
	ERRVALUE ComputeElements (
		const SIMPLE_ARRAY<LPOINT2D> &seeds,	//!<	Seed points
		const FLAGS flags								//!< Flags (see FLAGS enum)
		);

	//!	Computes flow path segmentation depending on parameter and interval values.
	//!
	//! @return ERRVALUE
	ERRVALUE ComputeFlowpathSegmentation (
		const SEGMENTATIONPARAMETER parameter,
		const SEGMENTATIONINTERVAL interval,
		const DOUBLE_ARRAY<double>& values
		);

	//! Computes hydrological elements as SCA, CTI, Maximum Up Stream and Down Stream flow distances.
	//! depends on flags
	//!
	//! @return ERRVALUE
	ERRVALUE ComputeExtraRasters (
		const FLAGS flags								//!< Flags (see FLAGS enum)
		);

	//!	Get output objitem that was calculated
	//!
	//! @return true if it was calculated
	bool GetObjItem (
		const OUTPUT id,								//!<	ID of output object (see OUTPUT enum)
		RVC::OBJITEM& objitem
		) const;

	//!	Get number of output objects that were calculated
	//!
	//! @return Number of output objects
	INT32 GetNumberOfObjects (
		) const;

	//! Get status of watershed process
	//!
	//! @return Status (see STATUS enum)
	STATUS GetStatus (
		) const {
		return m_Status;
		}


	//! Set use mask from external <= 8bit raster by RVC::OBJITEM
	//!
	//! @return ERRVALUE
	ERRVALUE SetMask (								//! Set mask to watershed model
		const RVC::OBJITEM& objitem,
		const bool use
		);


	//!	Get current basin value
	//!
	//! @return Current basin value
	UINT32 GetBasin (
		) const {
		return m_Basin;
		}

	//!	Set value as current basin value
	void SetBasin (
		const UINT32 basin							//!< Value to be set as current basin value
		) {
		if	(basin < MinimumBasin) m_Basin = MinimumBasin;
		else m_Basin = basin;
		return;
		}

	//!	Get current branch value
	//!
	//! @return Current branch value
	UINT32 GetBranch (
		) const {
		return m_Branch;
		}

	//!	Set value as current branch value		
	void SetBranch (
		const UINT32 branch							//!< Value to be set as current branch value
		) {
		if	(branch < MinimumBranch) m_Branch = MinimumBranch;
		else m_Branch = branch;
		return;
		}

	//!	Get current outlet value
	//!
	//! @return Current outlet value
	UINT32 GetOutlet (
		) const {
		return m_Outlet;
		}

	//!	Set value as current outlet value		
	void SetOutlet (
		const UINT32 outlet							//!< Value to be set as current outlet value
		) {
		if	(outlet < MinimumOutlet) m_Outlet = MinimumOutlet;
		else m_Outlet = outlet;
		return;
		}

	//!	Get current inlet value
	//!
	//! @return Current inlet value
	UINT32 GetInlet (
		) const {
		return m_Inlet;
		}

	//!	Set value as current inlet value		
	void SetInlet (
		const UINT32 inlet							//!< Value to be set as current inlet value
		) {
		if	(inlet < MinimumInlet) m_Inlet = MinimumInlet;
		else m_Inlet = inlet;
		return;
		}

	//!	Is valley set separated
	//!
	//! @return true if set to be separated otherwise - false
	bool IsValleySeparated (
		) const {
		return m_ValleySeparation;
		}

	//!	Set valley separation by given value
	void SetValleySeparation (
		const bool separation						//!< Separation value to be set as current
		) {
		m_ValleySeparation = separation;
		return;
		}

	//!	Get current depression maximum area to fill
	//!
	//! @return Current depression maximum area to fill
	UINT32 DepressionMaximumArea (
		) const {
		return m_DepressionMaximumArea;
		}

	//!	Set value as current depression maximum area to fill		
	void SetDepressionMaximumArea (
		const UINT32 area							
		) {
		m_DepressionMaximumArea = bound(area, s_MinimumDepressionMaximumArea, s_MaximumDepressionMaximumArea);
		return;
		}

	//!	Get current depression maximum null area to fill
	//!
	//! @return Current depression maximum null area to fill
	UINT32 DepressionMaximumNullArea (
		) const {
		return m_DepressionMaximumNullArea;
		}

	//!	Set value as current depression maximum null area to fill		
	void SetDepressionMaximumNullArea (
		const UINT32 area							
		) {
		m_DepressionMaximumNullArea = bound(area, s_MinimumDepressionMaximumNullArea, s_MaximumDepressionMaximumNullArea);
		return;
		}

	//!	Get current depression maximum depth to fill
	//!
	//! @return Current depression maximum depth to fill
	double DepressionMaximumDepth (
		) const {
		return m_DepressionMaximumDepth;
		}

	//!	Set value as current depression maximum depth to fill		
	void SetDepressionMaximumDepth (
		const double depth							
		) {
		m_DepressionMaximumDepth = bound(depth, s_MinimumDepressionMaximumDepth, s_MaximumDepressionMaximumDepth);
		return;
		}

	//!	Get current depression minimum depth area to fill
	//!
	//! @return Current depression minimum depth area to fill
	UINT32 DepressionMinimumDepthArea (
		) const {
		return m_DepressionMinimumDepthArea;
		}

	//!	Set value as current depression maximum area to fill		
	void SetDepressionMinimumDepthArea (
		const UINT32 area							
		) {
		m_DepressionMinimumDepthArea = bound(area, s_MinimumDepressionMinimumDepthArea, s_MaximumDepressionMinimumDepthArea);
		return;
		}

	//!	Get put null cell at unfilled depression
	//!
	//! @return true if set to be put otherwise - false
	bool GetPutNullCellAtUnfilledDepression (
		) const {
		return m_PutNullCellAtUnfilledDepression;
		}

	//!	Set put null cell at unfilled depression
	void SetPutNullCellAtUnfilledDepression (
		const bool value						
		) {
		m_PutNullCellAtUnfilledDepression = value;
		return;
		}


	//! Get number of watershed polygons in watershed system
	//!
	//! @return Number of watershed polygons
	const INT32 GetNumberOfWatershedPolygons (
		) const {
		if (m_Attribute == 0) return 0;
		return (m_Attribute->GetNumberOfWatershedPolygons() - 1);
		};

	//! Get watershed polygon object in watershed system by index
	//!
	//! @return true if exist, otherwise - false
	bool GetWatershedPolygon (
		POLYGON &polygon,								//!< Watershed polygon object returned
		const INT32 index								//!< Watershed polygon ID
		) {
		if (m_Attribute == 0) return false;
		return m_Attribute->GetWatershedPolygon(index, polygon);
		}

	//! Get number of watershed polygons with depression in watershed system
	//!
	//! @return Number of watershed polygons with depression
	const INT32 GetNumberOfWatershedPolygonsWithDepression (
		) const {
		if (m_Attribute == 0) return 0;
		return m_Attribute->GetNumberOfWatershedPolygonsWithDepression();
		}

	//! Get number of vector polygons in watershed system
	//!
	//! @return Number of vector polygons
	const INT32 GetNumberOfVectorPolygons (
		) const {
		if (m_Attribute == 0) return 0;
		return (m_Attribute->GetNumberOfVectorPolygons());
		};

	//! Get watershed polygon ID in watershed system at given point 
	//!
	//! @return Watershed polygon ID or -1
	const INT32 GetIDOfWatershedPolygonAtPoint (
		const DPOINT2D &point								//!< Given point
		);

	//! Get watershed polygon ID in watershed system for given vector polygon element ID 
	//! that belongs to this watershed
	//!
	//! @return Watershed polygon ID or -1
	const INT32 GetIDOfWatershedPolygonForVectorPolygon (
		const INT32 index								//!< Vector polygon element ID
		) const {
		if (m_Attribute == 0) return -1;
		return (m_Attribute->GetIDOfWatershedPolygonForVectorPolygon(index));
		};

	//! Get number of pour points in watershed system
	//!
	//! @return Number of pour points
	const INT32 GetNumberOfPours (
		) const {
		if (m_Attribute == 0) return 0;
		return (m_Attribute->GetNumberOfPours());
		};

	//! Get pour object in watershed system by index
	//!
	//! @return true if exist, otherwise - false
	bool GetPour (
		POUR &pour,										//!< POUR object returned 
		const INT32 index								//!< Pour point ID
		) {
		if (m_Attribute == 0) return false;
		return m_Attribute->GetPour(index, pour);
		}

	//! Get number of double pour points in watershed system
	//!
	//! @return Number of double pour points
	const INT32 GetNumberOfDoublePours (
		) const {
		if (m_Attribute == 0) return 0;
		return m_Attribute->GetNumberOfDoublePours();
		}

	UINT32 GetDefaultInlet (
		) {
		const double z = (double)MAX(m_NumColumns, m_NumLines) / 1024;
		return FAST_ROUND_POSITIVE(MinimumInlet * MAX(1.0, z));
		}

	UINT32 GetDefaultOutlet (
		) {
		const double z = (double)MAX(m_NumColumns, m_NumLines) / 1024;
		return FAST_ROUND_POSITIVE(MinimumOutlet * MAX(1.0, z));
		}

	UINT32 GetDefaultBasin (
		) {
		const double x = (double)m_NumColumns / 1024;
		const double y = (double)m_NumLines / 1024;
		return FAST_ROUND_POSITIVE(MinimumBasin * MAX(1.0, x * y));
		}

	UINT32 GetDefaultBranch (
		) {
		const double x = (double)m_NumColumns / 1024;
		const double y = (double)m_NumLines / 1024;
		return FAST_ROUND_POSITIVE(MinimumBranch * MAX(1.0, x * y));
		}

private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

	// ENUMERATIONS

	enum RASTERS {
		RASTER_First				= 0,					
		RASTER_Adjusted			= 0,				//!< As OUTPUT_RasterAdjusted
		RASTER_Extrema				= 1,				//!< As OUTPUT_RasterExtrema
		RASTER_FlowDirection		= 2,				//!< As OUTPUT_RasterFlowDirection
		RASTER_Accumulation		= 3,				//!< As OUTPUT_RasterAccumulation
		RASTER_Watershed			= 4,				//!< As OUTPUT_RasterWatershed
		RASTER_CatchmentArea		= 5,				//!< As OUTPUT_RasterCatchmentArea
		RASTER_TopographicIndex = 6,				//!< As OUTPUT_RasterTopographicIndex
		RASTER_MaximumUpFlow		= 7,				//!< As OUTPUT_RasterMaximumUpFlow	
		RASTER_DownFlow			= 8,				//!< As OUTPUT_RasterDownFlow				
		RASTER_Temp					= 9,				//!< Temporary 4bit raster
		RASTER_Bin					= 10,				//!< Temporary binary raster
		RASTER_Count				= 11
		};

	enum VECTORS {
		VECTOR_First			= 0,					
		VECTOR_Watershed		= 0,					//!< As OUTPUT_VectorWatershed
		VECTOR_FlowPath		= 1,					//!< As OUTPUT_VectorFlowPath
		VECTOR_Basin			= 2,					//!< As OUTPUT_VectorBasin
		VECTOR_Ridge			= 3,					//!< As OUTPUT_VectorRidge
		VECTOR_UserFlowPath	= 4,					//!< As OUTPUT_VectorUserFlowPath
		VECTOR_UserBasin		= 5,					//!< As OUTPUT_VectorUserBasin
		VECTOR_SegmentedFlowPath = 6,				//!< As OUTPUT_VectorSegmentedFlowPath
		VECTOR_Count			= 7
		};

	struct STREAMORDER { 
		INT32 m_Line;
		INT32 m_Parent;
		UINT32 m_Accumulation;
		INT32 m_Child;
		UINT32 m_ChildAccumulation;
		UINT8 m_NumChildren;
		INT32 m_ID;
		UINT16 m_Horton;
		UINT16 m_Strahler;
		UINT16 m_Shreve;
		UINT16 m_Scheidegger;

		STREAMORDER (
			) :
			m_Line(-1),
			m_Parent(-1),
			m_Accumulation(0),
			m_Child(-1),
			m_ChildAccumulation(0),
			m_NumChildren(0),
			m_ID(0),
			m_Horton(0),
			m_Strahler(0),
			m_Shreve(0),
			m_Scheidegger(0)
			{
			};

		};

	struct ORDERSTAT {
		double length;
		double area;
		INT32 number;
		};

	// TYPEDEFS

	typedef std::map<INT32, ORDERSTAT> ORDERSTATMAP;

	// MEMBERS

	const RVC::OBJITEM m_DEMObjItem;
	INT32 m_NumLines;
	INT32 m_NumColumns;

	RVC::OBJECT m_RasterTempFiles[RASTER_Count];
	RVC::RASTER m_Rasters[RASTER_Count];
	RVC::OBJECT m_VectorTempFiles[VECTOR_Count];
	RVC::VECTOR m_Vectors[VECTOR_Count];
	RVC::OBJITEM m_ObjItems[OUTPUT_Count];
	ATTRIBUTE *m_Attribute;							//!< Watershed attributes
	STATUS m_Status;									//!< Status of watershed object 
	UINT32 m_Outlet;									//!< Current outlet value
	UINT32 m_Inlet;									//!< Current inlet value
	UINT32 m_Branch;									//!< Current branch value
	UINT32 m_Basin;									//!< Current basin value
	bool m_ValleySeparation;						//!< Current valley separation status
	UINT32 m_DepressionMaximumArea;									
	UINT32 m_DepressionMaximumNullArea;									
	double m_DepressionMaximumDepth;									
	UINT32 m_DepressionMinimumDepthArea;									
	bool m_PutNullCellAtUnfilledDepression;						

	bool m_UseMask;
	RVC::OBJITEM m_MaskObjItem;					//!< Mask objitem
	SIMPLE_ARRAY<UINT8> m_NullBuffer;

	FLAGS m_Flags;


	// CONSTRUCTION / DESTRUCTION

	WATERSHED (										//! Default constructor (not implemented)
		);

	WATERSHED (										//! Copy constructor (not implemented)
		const WATERSHED &rhs
		);

	// OPERATORS

	WATERSHED& operator= (						//! Assignment operator (not implemented)
		const WATERSHED& rhs
		);


	// METHODS

	ERRVALUE CheckNeededSpace ();
	ERRVALUE CopyMaskAsNull ();

	ERRVALUE MakeAdjustedElevation (RVC::RASTER& dem);
	ERRVALUE MakeFlowDirection (RVC::RASTER& dem);
	ERRVALUE MakeAccumulation (RVC::RASTER& dem);
	ERRVALUE MakeExtrema (RVC::RASTER& dem);
	ERRVALUE MakeWatershedRaster (RVC::RASTER& dem);
	ERRVALUE MakeTemp (RVC::RASTER& dem);
	ERRVALUE MakeBin (RVC::RASTER& dem);
	ERRVALUE MakeCatchmentArea (RVC::RASTER& dem);
	ERRVALUE MakeTopographicIndex (RVC::RASTER& dem);
	ERRVALUE MakeMaximumUpFlow (RVC::RASTER& dem);
	ERRVALUE MakeDownFlow (RVC::RASTER& dem);

	ERRVALUE MakeVector (
		RVC::RASTER& dem,
		const RVC::DESCRIPTOR& descriptor,
		const WATERSHED::VECTORS id,
		const RVC::VECTOR::POINTTYPE PointType,
		const RVC::VECTOR::TOPOLOGYTYPE TopologyType
		);

	ERRVALUE ComputeAccumulation ();
	ERRVALUE ComputeAttribute ();
	ERRVALUE ComputeBasin (const SIMPLE_ARRAY<LPOINT3D> &seeds);
	ERRVALUE ComputeBasin (const SIMPLE_ARRAY<LPOINT2D> &seeds);
	ERRVALUE ComputeDatabases ();
	ERRVALUE ComputeExtrema ();
	ERRVALUE ComputeFlowDirection ();
	template<typename _TT> ERRVALUE ComputeFlowPath (SIMPLE_ARRAY<LPOINT3D> &seeds);
	template<typename _TT> ERRVALUE ComputeFlowPath (const SIMPLE_ARRAY<LPOINT2D> &seeds);
	template<typename _TT> ERRVALUE ComputeRidge ();		  
	ERRVALUE ComputeStandard ();
	ERRVALUE ComputeBasinTables (const INT32 numBasins);
	ERRVALUE ComputeBasinAttributes (const INT32 numBasins);
	ERRVALUE ComputeWatershedAttributes ();
	ERRVALUE ComputeFlowPathTables ();
	ERRVALUE ComputeVectorTables (const INT32 numBasins);
	ERRVALUE ComputeTableFlowFeatures ();
	ERRVALUE ComputeTableWatershedVectorPolygons ();
	ERRVALUE ComputeTableWatershed ();
	ERRVALUE ComputeTablePour ();
	ERRVALUE ComputeTableLower ();
	ERRVALUE ComputeTableStreamOrder ();
	ERRVALUE ComputeTableStreamProperties ();
	ERRVALUE ComputeTableFlowDistance (const VECTORS vid);
	ERRVALUE ComputeTableFlowDown (const VECTORS vid);
	ERRVALUE ComputeTableUpper ();
	ERRVALUE ComputeWatershedRaster ();
	ERRVALUE ComputeWatershedVector ();
	ERRVALUE ComputeCatchmentArea ();
	ERRVALUE ComputeTopographicIndex ();
	template<typename _TT> ERRVALUE CalculateTopographicIndex ();
	ERRVALUE ComputeMaximumUpFlow ();
	ERRVALUE InitMaximumUpFlow ();
	ERRVALUE CalculateMaximumUpFlow ();
	ERRVALUE ComputeDownFlow ();
	ERRVALUE ComputeBoundaries (const VECTORS id, const INT32 limit);
	ERRVALUE FillDepressions ();
	ERRVALUE CalculateLowerDepressions (const INT32 watershed, SIMPLE_ARRAY<bool> &filling);
	ERRVALUE CalculateUpperDepressions (const INT32 watershed, SIMPLE_ARRAY<bool> &filling);
	ERRVALUE CalculateDepressionsToBeFilled (SIMPLE_ARRAY<bool> &filling, const SIMPLE_ARRAY<INT32> &depressions);
	void DestroyInternalObjects ();
	ERRVALUE MakeRasterObject (const RASTERS id);
	ERRVALUE MakeVectorObject (const VECTORS id);
	RVC::RASTER& GetRasterObject (const RASTERS id, const bool checkifopen = true);
	RVC::OBJECT& GetRasterTempFile (const RASTERS id);
	RVC::VECTOR& GetVectorObject (const VECTORS id, const bool checkifopen = true);
	RVC::OBJECT& GetVectorTempFile (const VECTORS id);
	ERRVALUE  MakeDefaultRasterCalculation ();
	ERRVALUE  MakeDefaultRasterCalculation (const RASTERS id);
	void SetObjItem (const OUTPUT id);
	ERRVALUE ClearVectorObject (const VECTORS id);
	void ClearObjItem (const OUTPUT id);
	void ClearObjItems ();

	ERRVALUE CopyObjectTo (RVC::OBJECT& object, const char *filename);

	template<typename _TT> ERRVALUE CalculateExtrema (const _TT tmin, const _TT tmax);
	template<typename _TT> ERRVALUE ReadAdjLine (RVC::NULLMASK& NullMask, _TT* adjbuf, UINT8* vldbuf, const INT32 line);
	template<typename _TT> ERRVALUE ReadAdjLine (_TT* adjbuf, const INT32 line);
	template<typename _TT> void MinMaxAdjLine(const _TT *const adjbuf, const UINT8 *const vldbuf, _TT*  minbuf, _TT*  maxbuf, const _TT tmin, const _TT tmax, const INT32 numColumns);
	template<typename _TT> ERRVALUE CalculateFlowDirection (const _TT tmax);
	ERRVALUE InitFlowDirectionOnFlat ();
	template<typename _TT> ERRVALUE ResolveFlowDirectionOnFlat ();
	ERRVALUE InitAccumulation ();
	ERRVALUE CalculateAccumulation ();
	ERRVALUE ReadDirLine(UINT8* dirbuf, const INT32 line);
	ERRVALUE ReadAccLine(UINT32* accbuf, const INT32 line);
	ERRVALUE InitWatershedRaster ();
	ERRVALUE CalculateWatershedRaster ();
	ERRVALUE ComputeBoundaryInitTemp (const INT32 limit);
	ERRVALUE ReadWtrLine(INT32* wtrbuf, const INT32 line);
	ERRVALUE ComputeBoundaryInitBinary ();
	template<typename _TT> ERRVALUE CalculateBoundaries (const VECTORS id);
	template<typename _TT> ERRVALUE ComputeAttributeWatershed ();
	template<typename _TT> ERRVALUE ComputeAttributeLines (RVC::VECTOR& vector);
	template<typename _TT> ERRVALUE ComputeAttributeNodes (RVC::VECTOR& vector);
	template<typename _TT> ERRVALUE ComputeAttributeDepression ();
	template<typename _TT> ERRVALUE FillSelectedDepressions (const SIMPLE_ARRAY<bool> &filling);
	ERRVALUE ComputeBasinMarkPolygons (const SIMPLE_ARRAY<LPOINT3D> &seeds);
	ERRVALUE ComputeBasinMarkPolygons (const SIMPLE_ARRAY<LPOINT2D> &seeds, SIMPLE_ARRAY<INT32>& polygons);
	ERRVALUE ComputeBasinRestorePolygons (const SIMPLE_ARRAY<LPOINT3D> &seeds);
	ERRVALUE ComputeBasinRestorePolygons (const SIMPLE_ARRAY<INT32>& polygons);
	
	#endif // GENERATING_DOXYGEN_OUTPUT
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(WATERSHED::FLAGS)
DEFINE_ENUM_OP_BITWISE(WATERSHED::OUTPUT)
#endif

#endif

