/**
 * \file rvc/dbelem.h
 * \brief interface for the RVC::DBASE_ELEMENT class.
 *
 * \if NODOC
 * $Id: dbelem.h_v 1.41 2011/12/07 22:26:55 mju Exp $
 *
 * $Log: dbelem.h_v $
 * Revision 1.41  2011/12/07 22:26:55  mju
 * Remove redundant hdr.
 *
 * Revision 1.40  2010/05/12 21:07:04  mju
 * Require parent objtype in ctor.
 *
 * Revision 1.38  2006/10/05 14:49:31  bcolombe
 * Added optional threshold value to CompareElementRecords
 *
 * Revision 1.35  2006/06/02 22:23:32  dwilliss
 * Changes to Join
 *
 * Revision 1.34  2005/12/20 15:49:21  scowan
 * Added static method to get element type based on objectr type.
 *
 * Revision 1.30  2005/04/05 14:28:50  dwilliss
 * Added a new get object filter
 *
 * Revision 1.29  2004/05/27 19:48:22  scowan
 * Added method to transfer style assignment tables.
 *
 * Revision 1.28  2003/12/04 21:18:31  scowan
 * Added element record compare methodd.
 *
 * Revision 1.5  2001/06/26 16:53:55  scowan
 * Specify issmultiobject() to turn off base class setting.
 *
 * Revision 1.1  2001/05/24 15:13:17  scowan
 * Initial revision
 *
 * \endif
**/

#ifndef INC_RVC_DBELEM_H
#define INC_RVC_DBELEM_H

#ifndef INC_RVC_DBTABLE_H
	#include <rvc/dbtable.h>
#endif

#ifndef INC_RVC_ELEMENT_H
	#include <rvc/element.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations
struct DBLISTELMT;
namespace RVC {
	class GEOSPATIAL;
	}
#endif // GENERATING_DOXYGEN_OUTPUT

namespace RVC {

//! Database object associated with specific geospatial element type.
//! Provides methods to retrieve and manage list of records attached to each element.
class DBASE_ELEMENT : public DBASE {
	public:
	
		class ATTACHMENTITEM;
		class GETOBJFILTER_ISSELECTABLE;
		typedef SIMPLE_ARRAY<ATTACHMENTITEM> ATTACHMENTARRAY;
		
		typedef SIMPLE_ARRAY<DBLISTELMT> DBLIST;
		
		//! Construct using parent object type and element type.
		DBASE_ELEMENT (
			OBJTYPE ObjTypeParent,
			ELEMTYPE ElemType
			);

		//! Copy constructor
		DBASE_ELEMENT (
			const RVC::DBASE_ELEMENT& rhs
			);

		//! Destructor
		virtual ~DBASE_ELEMENT (
			);

		//! Assignment
		RVC::DBASE_ELEMENT& operator= (
			const RVC::DBASE_ELEMENT& rhs
			);
			
		//! Attach the record to the element
		ERRVALUE AddAttachment (
			const RVC::ELEMENT& elem,
			const RVC::DBASE_ELEMENT::ATTACHMENTITEM& item
			);

		//! Attach the array of records to the element
		ERRVALUE AddAttachments (
			const RVC::ELEMENT& elem,
			const RVC::DBASE_ELEMENT::ATTACHMENTARRAY& items
			);

		//! Append list entries to this element.
		//! This method will remove duplicates.
		INT32 AppendDBList (
			INT32 ElemNum,
			const DBLIST& DBList
			);

		//! Append one list entry to this element.
		//! This method will remove duplicates.
		INT32 AppendDBList (
			INT32 ElemNum,
			const DBLISTELMT& item
			);

		//! Remove invalid entries from the element database attachment list
		ERRVALUE CleanDBList (
			);

		//! Compare attached element records from two databases to determine if the element attachmets are to the same records
		//! @return 'True' if they are, false if not.
		bool CompareElementRecords (
			INT32 LhsElemNum,							//!< Left side element number
			const DBASE_ELEMENT& rhs,					//!< Other database to compare to
			INT32 RhsElemNum,							//!< Right side element number
			const SIMPLE_ARRAY<INT32>& TableMatch,		//!< Table translation array from Lhs database to Rhs database, See DBASE::FindTableMatches()
			double threshhold = 0.0						//bcolombe: Added optional threshold value (10/4/06)	
			) const;									

		//! Convert style assignment table from 'this' to 'DestDBaseObj' based on the 
		//! database object types
		ERRVALUE ConvertStyleAssignTableTo (
			DBASE_ELEMENT& DestDBaseObj					//!< Must be created outside this function and be open
			);

		//! Copy a database from the source to under the destination object using an element translation list
		ERRVALUE CopyDatabaseTo (
			RVC::GEOSPATIAL& DestParentObj,
			const SIMPLE_ARRAY<INT32>& TransList,
			COPYFLAG flags
			);

		//! Copy a database from the source to under the destination object using an element translation list
		ERRVALUE CopyDatabaseTo (
			RVC::GEOSPATIAL& DestParentObj,
			RVC::DBASE_ELEMENT& DestDBaseObj,		//!< To be filled in with the new database object
			const DBASE::ELEMTRANSLIST& TransList,	//!< Element translation list
			COPYFLAG flags
			);

		//! Get the list of attached elements to a specific record			
		int GetAttachedElements (
			const RVC::DBTABLE::RECORD& record,			//!< Record to find attached elements of
			SIMPLE_ARRAY<INT32>& ElementList,
			ELEMTYPE ElemType = ELEMTYPE_Empty			//!< Defaults to DBASE_ELEMENT ElemType
			) const;

		//! Get the list of attached elements to a specific record			
		int GetAttachedElements (
			const RVC::DBTABLE::RECORD& record,			//!< Record to find attached elements of
			SIMPLE_ARRAY<RVC::ELEMENT>& ElementList,
			ELEMTYPE ElemType = ELEMTYPE_Empty			//!< Defaults to DBASE_ELEMENT ElemType
			) const;

		//! Get records attached to the specified element for all tables
		ERRVALUE GetAttachedRecordList (
			INT32 ElemNum,
			DBLIST& DBList,							//!< List of attachments allocated and returned
			FINDFLAG FindFlag = FINDFLAG_AllAttached,
			INT32 BaseTableNum = -1
			) const;
			
		//! Get the record numbers attached to the element for this table.
		//! @return Number of records attached or <0 if error 
		INT32 GetAttachedRecordNumbers (
			const RVC::ELEMENT& element,					//!< Element
			const RVC::DBTABLE& table,						//!< Table to get attached records for
			SIMPLE_ARRAY<INT32>& recordnums,				//!< Array of record numbers returned
			FINDFLAG findflags = FINDFLAG_AllAttached
			) const;

		//! Get array of attachments to this element
		INT32 GetAttachments (
			const RVC::ELEMENT& elem,
			RVC::DBASE_ELEMENT::ATTACHMENTARRAY& items
			) const;

		//! Get the size of an individual entries for an element database attachment list
		//! @return Number of items for an individual DBList entry
		INT32 GetDBListItemSize (
			INT32 ItemNum,
			bool ExcludeInternal = false
			) const;

		//! Get the maximum number of entries for an element database attachment list
		//! @return Maximum number of DBList entries
		INT32 GetDBListMaxItemSize (
			) const;

		//! Get the number of element database attachment list entries
		//! @return Number of DBList entries
		INT32 GetDBListNumElements (
			) const;

		//! Get the default element type based on the database object type
		//! @return The default element type for the object type, or ELEMTYPE_Empty if cannot match
		static ELEMTYPE GetDftElemType (
			RVC::OBJTYPE ObjType
			);

		//! Get element type database is associated with.
		ELEMTYPE GetElemType (
			) const { return (m_ElemType); }

		//! Determine if the given record has any elements attached
		//! @return 'True' if elements are attached, 'false' if not
		bool HasAttachedElements (
			const RVC::DBTABLE::RECORD& record,
			ELEMTYPE ElemType = ELEMTYPE_All
			) const;
			

		//! Join (copy) tables from this database to another database
		ERRVALUE JoinDatabaseTo (
			RVC::DBASE_ELEMENT& DestDatabase,	
			const SIMPLE_ARRAY<INT32>& TransList,
			RVC::DBASE::COPYFLAG CopyFlags = RVC::DBASE::COPYFLAG_None,
			RVC::DBASE::JOINFLAGS JoinFlags = RVC::DBASE::JOINFLAG_None,
			INT32 *pNextElemID = 0,		//! Pointer to INT32 return next ElemID in
			ELEMTYPE SrcElemType = ELEMTYPE_Empty,  //! Will default to the default element type of the source database
			ELEMTYPE DestElemType = ELEMTYPE_Empty  //! Will default to the default element type of the destination database
			);

		//! Join (copy) tables from this database to another database
		ERRVALUE JoinDatabaseTo (
			RVC::DBASE_ELEMENT& DestDatabase,	
			const DBASE::ELEMTRANSLIST& TransList,
			RVC::DBASE::COPYFLAG CopyFlags = RVC::DBASE::COPYFLAG_None,
			RVC::DBASE::JOINFLAGS JoinFlags = RVC::DBASE::JOINFLAG_None,
			INT32 *pNextElemID = 0,		//! Pointer to INT32 return next ElemID in
			ELEMTYPE SrcElemType = ELEMTYPE_Empty,  //! Will default to the default element type of the source database
			ELEMTYPE DestElemType = ELEMTYPE_Empty  //! Will default to the default element type of the destination database
			);

		//! Open the database.  If a database object does not exist, make one
		ERRVALUE OpenAsSubobject (
			const RVC::OBJITEM& ParentObjItem,				//!< Parent object to find and/or create database under
#ifdef RVC_CREDENTIALS_MANDATORY
			OPENMODE mode,
			MDLGPARENT dlgparent,					//!< Parent dialog handle, 0 for default handle
			const CREDENTIALS *pCredentials //!< User credentials passed in, NULL for default behavior
#else
			OPENMODE mode = OPENMODE_Read,
			MDLGPARENT dlgparent = 0,					//!< Parent dialog handle, 0 for default handle
			const CREDENTIALS *pCredentials = 0				//!< User credentials passed in, NULL for default behavior
#endif
			);

		//! Open the database.  If a database object does not exist, make one
		ERRVALUE OpenAsSubobject (
			const RVC::GEOSPATIAL& parent,				//!< Parent object to find and/or create database under
#ifdef RVC_CREDENTIALS_MANDATORY
			OPENMODE mode,
			MDLGPARENT dlgparent,					//!< Parent dialog handle, 0 for default handle
			const CREDENTIALS *pCredentials //!< User credentials passed in, NULL for default behavior
#else
			OPENMODE mode = OPENMODE_Read,
			MDLGPARENT dlgparent = 0,					//!< Parent dialog handle, 0 for default handle
			const CREDENTIALS *pCredentials = 0				//!< User credentials passed in, NULL for default behavior
#endif
			);

		//! Read an individual entry from the element database attachment list
		//! @return Number of entries read
		INT32 ReadDBList (
			INT32 ElemNum,
			DBLIST& DBList
			) const;

		//! Remove the record attachment to the element
		ERRVALUE RemoveAttachment (
			const RVC::ELEMENT& elem,
			const RVC::DBASE_ELEMENT::ATTACHMENTITEM& item
			);

		//! Remove the array of record attachments to the element
		ERRVALUE RemoveAttachments (
			const RVC::ELEMENT& elem,
			const RVC::DBASE_ELEMENT::ATTACHMENTARRAY& items
			);

		//! Resize the number of entries in the element database attachment list
		ERRVALUE ResizeDBList (
			INT32 NumItems
			);

		//! Resize an individual entry from the element database attachment list
		ERRVALUE ResizeDBListItem (
			INT32 ElemNum,
			INT32 NumItems
			);

		//! Swap individual entries from the element database attachment list
		ERRVALUE SwapDBList (
			INT32 Item1,
			INT32 Item2
			);

		//! Write an individual entry from the element database attachment list
		//! @return Number of entries written
		INT32 WriteDBList (
			INT32 ElemNum,
			const DBLIST& DBList
			);

		//! Unattach items from an individual entry from the element database attachment list
		//! @return Number of entries unattached
		INT32 UnattachDBList (
			INT32 ElemNum,
			const DBLIST& DBList
			);

		//! Inform the database code that the number of elements has changed
		//! May not be needed
		ERRVALUE UpdateDBListNumElements (
			INT32 NumElements
			);

	protected:

		#ifndef GENERATING_DOXYGEN_OUTPUT
		//! Join tables from another database into this one.
		ERRVALUE JoinPointDatabaseToLow (
			RVC::DBASE_ELEMENT& DestDatabase,	
			const SIMPLE_ARRAY<INT32>& PointTransList,
			const SIMPLE_ARRAY<INT32>& NodeTransList,
			RVC::DBASE::COPYFLAG CopyFlags = RVC::DBASE::COPYFLAG_None,
			RVC::DBASE::JOINFLAGS JoinFlags = RVC::DBASE::JOINFLAG_None,
			INT32 *pNextElemID = 0,		//! Pointer to INT32 return next ElemID in
			ELEMTYPE SrcElemType = ELEMTYPE_Empty,  //! Will default to the default element type of the source database
			ELEMTYPE DestElemType = ELEMTYPE_Empty  //! Will default to the default element type of the destination database
			);
		#endif // GENERATING_DOXYGEN_OUTPUT

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		OBJTYPE m_ObjType;
		ELEMTYPE m_ElemType;

		// Implementation of RVC::OBJECT virtual methods
		virtual OBJTYPE v_GetDftObjectType () const;
		virtual bool v_IsMultiObject () const;

		#endif // GENERATING_DOXYGEN_OUTPUT

	};
	
	
class DBASE_ELEMENT::ATTACHMENTITEM {
	public:

		//! Default constructor.		
		ATTACHMENTITEM (
			) : m_TableNum(TABLE_Invalid), m_RecordNum(RECORD_New) {}
		
		//! Construct from DBTABLE and record number.
		ATTACHMENTITEM (
			const RVC::DBTABLE& table,
			INT32 recordnum
			);

		//! Get the table number
		//! @return Table Number
		INT32 GetTableNum (
			) const {return (m_TableNum);}

		//! Get the record number
		//! @return Record number
		INT32 GetRecordNum (
			) const {return (m_RecordNum);}

		//! Set the table number
		void SetTableNum (
			INT32 TableNum
			) {m_TableNum = TableNum;}

		//! Set the record number
		void SetRecordNum (
			INT32 RecordNum
			) {m_RecordNum = RecordNum;}

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		
		INT32 m_TableNum;
		INT32 m_RecordNum;
		
		operator const DBLISTELMT* (
			) const {return (reinterpret_cast<const DBLISTELMT*>(this));}
		
		operator DBLISTELMT* (
			) {return (reinterpret_cast<DBLISTELMT*>(this));}
		
		friend class DBASE_ELEMENT;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};
	
//! Database object selection filter to select either a database object, which could be a subobject of a geospatial object.
class DBASE_ELEMENT::GETOBJFILTER_ISSELECTABLE : public RVC::GETOBJFILTER {
	public:
		// Default constructor.
		// If bCreateIfNotExist is true, navigating under a geospatial object
		// with this filter will force it to create the database subobjects
		// if they don't already exist.  You would normally only use this if
		// your filter is for the purposes of selecting an destination database.
		GETOBJFILTER_ISSELECTABLE (
			bool bCreateIfNotExist = false
			);

		// Constructor for a specific database subobject type.
		// If bCreateIfNotExist is true, navigating under a geospatial object
		// with this filter will force it to create the database subobjects
		// if they don't already exist.  You would normally only use this if
		// your filter is for the purposes of selecting an destination database.
		GETOBJFILTER_ISSELECTABLE (
			RVC::OBJTYPESET& type, 
			bool bCreateIfNotExist = false
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		virtual bool v_HasSelectableSubObjects (const RVC::OBJITEM& objitem);
		virtual bool v_IsSelectable (const RVC::OBJITEM& objitem, int position);
		virtual bool v_IsValidParent (const RVC::OBJITEM& objitem);
		RVC::OBJTYPESET m_Types;
		bool m_bCreateIfNotExist;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};	
	
}		// End of RVC namespace

#endif 	// INC_RVC_DBELEM_H
