/**
 * \file rvc/optriangulation.h
 * \brief Definitions for RVC::OP::TRIANGULATION class.
 *
 * \if NODOC
 * $Id: optriangulation.h_v 1.10 2009/06/26 13:55:30 mju Exp $
 *
 * $Log: optriangulation.h_v $
 * Revision 1.10  2009/06/26 13:55:30  mju
 * Make some accessors inline.
 *
 * Revision 1.9  2009/06/26 12:50:57  mju
 * Use component class.
 *
 * Revision 1.8  2009/03/24 16:26:43  mju
 * Add needed headers.
 *
 * Revision 1.7  2008/06/16 19:17:20  vdronov
 * added support for shape
 *
 * Revision 1.6  2008/06/13 16:15:16  vdronov
 * added maximum memory
 *
 * Revision 1.1  2007/01/11 22:59:43  vdronov
 * Initial revision
 *
 * \endif
**/

#ifndef  INC_RVC_OPTRIANGULATION_H
#define  INC_RVC_OPTRIANGULATION_H

#ifndef	INC_RVC_OPCOMPONENT_H
	#include <rvc/opcomponent.h>
#endif

#ifndef	INC_MI32_SPATREF_H
#include <mi32/spatref.h>
#endif

#ifndef	INC_RVC_OBJITEM_H
#include <rvc/objitem.h>
#endif

#include <float.h>


namespace RVC {
namespace OP {

//! Class for creating triangulation.
class TRIANGULATION {
	public:

	   //! enum for available sources for triangulation.
		enum SOURCE {
			SOURCE_Invalid = -1,
			SOURCE_First = 0,
			SOURCE_Raster = 0,
			SOURCE_Table = 1,
			SOURCE_Vector = 2,
			SOURCE_Shape = 3,
			SOURCE_Last = 3,
			SOURCE_Count = 4
			};

	   //! enum for available methods of triangulations
		enum METHOD {
			METHOD_Invalid = -1,
			METHOD_First = 0,
			METHOD_Delaunay = 0,     
			METHOD_AdaptiveDensification = 1,  
			METHOD_Last = 1,
			METHOD_Count = 2
			};

		//! Class describes parameters of source object.
		class SOURCE_PARAMETERS {
			public:

				SOURCE_PARAMETERS (
					const SOURCE source
					);

				virtual ~SOURCE_PARAMETERS (
					);

				SOURCE GetSource (
					) const { return m_Source; }

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				SOURCE_PARAMETERS (
					);

				const SOURCE m_Source;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Class describes parameters of Raster source object.
		class RASTER_PARAMETERS : public SOURCE_PARAMETERS {
			public:

				RASTER_PARAMETERS (
					);

				~RASTER_PARAMETERS (
					);

			private:
			};

		//! Class describes parameters of table source object constructing 3D points from each record in table 
		class TABLE_PARAMETERS : public SOURCE_PARAMETERS {
			public:

				//! enum describes id's of field names constructing 3D points.
				enum NAME {
					NAME_XField = 0,
					NAME_YField = 1,
					NAME_ZField = 2,
					NAME_Count = 3
					};

				TABLE_PARAMETERS (
					);

				~TABLE_PARAMETERS (
					);

				//! Returns COORDREFSYS for constructed 3D points.
				const SPATREF::COORDREFSYS& GetCoordRefSys (
					) const { return (m_CoordRefSys); }

				const COMPONENT& GetComponent (
					) const { return (m_Pin); }

				//! Returns field name for each coordinate of constructed 3D point.
				void GetName (
					const NAME id,
					RVC::OBJECTNAME& name
					) const;

				//! Returns generic query for constructing 3D point.
				const MISTRING& GetQuery (
					) const { return (m_Query); }

				//! Returns unit code for x coordinate of constructed 3D point.
				UINT8 GetXUnitCode (
					) const { return (m_XUnitCode); }

				//! Returns unit code for y coordinate of constructed 3D point.
				UINT8 GetYUnitCode (
					) const { return (m_YUnitCode); }

				//! Returns unit code for z coordinate of constructed 3D point.
				UINT8 GetZUnitCode (
					) const { return (m_ZUnitCode); }

				//! Set COORDREFSYS for constructed 3D points.
				void SetCoordRefSys (
					const SPATREF::COORDREFSYS& crs
					) { m_CoordRefSys = crs; }

				void SetComponent (
					const COMPONENT& component
					) { m_Pin = component; }

				//! Set field name for each coordinate of constructed 3D point.
				void SetName (
					const NAME id,
					const RVC::OBJECTNAME& name
					);

				//! Set generic query for constructing 3D point.
				void SetQuery (
					const MISTRING& query
					) { m_Query = query; }

				//! Set unit code for x coordinate of constructed 3D point.
				void SetXUnitCode (
					const UINT8 unitcode
					) { m_XUnitCode = unitcode; }

				//! Set unit code for y coordinate of constructed 3D point.
				void SetYUnitCode (
					const UINT8 unitcode
					) { m_YUnitCode = unitcode; }

				//! Set unit code for z coordinate of constructed 3D point.
				void SetZUnitCode (
					const UINT8 unitcode
					) { m_ZUnitCode = unitcode; }

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				COMPONENT m_Pin;
				RVC::OBJECTNAME m_Names[NAME_Count];
				SPATREF::COORDREFSYS m_CoordRefSys;
				MISTRING m_Query;
				UINT8 m_XUnitCode;
				UINT8 m_YUnitCode;
				UINT8 m_ZUnitCode;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Class describes parameters of Vector source object.
		class VECTOR_PARAMETERS : public SOURCE_PARAMETERS {
			public:

				//! enum describes available types of source components.
				enum TYPE {
					TYPE_Point = 0,
					TYPE_Line = 1
					};

				VECTOR_PARAMETERS (
					);

				~VECTOR_PARAMETERS (
					);

				bool GetComponent (
					const TYPE type,
					COMPONENT& component
					) const;

				const OP::COMPONENT& GetComponentLine (
					) const { return (m_Line); }

				const OP::COMPONENT& GetComponentPoint (
					) const { return (m_Point); }

				void SetComponent (
					const TYPE type,
					const COMPONENT& component
					);

				void SetComponentLine (
					const OP::COMPONENT& component
					) { m_Line = component; }

				void SetComponentPoint (
					const OP::COMPONENT& component
					) { m_Point = component; }

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				COMPONENT m_Point;
				COMPONENT m_Line;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};


		//! Class describes parameters of Shape source object.
		class SHAPE_PARAMETERS : public SOURCE_PARAMETERS {
			public:

				SHAPE_PARAMETERS (
					);

				~SHAPE_PARAMETERS (
					);

				const COMPONENT& GetComponent (
					) const { return (m_Component); }

				void SetComponent (
					const COMPONENT& component
					) { m_Component = component; }

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				COMPONENT m_Component;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};


		//! Class describes parameters of destination object.
		class DESTINATION_PARAMETERS {
			public:

				DESTINATION_PARAMETERS (
					);

				~DESTINATION_PARAMETERS (
					);

				//! Returns true if break lines has to be applyed to destination object.
				bool GetApplyBreaklines (
					) const;

				//! Returns true if clipping has to be applyed to destination object.
				bool GetApplyClipping (
					) const;

				//! Returns break lines OBJITEM
				const RVC::OBJITEM& GetBreaklinesObjItem (
					) const;

				//! Returns clip method: TINCLIPMODE_Include or TINCLIPMODE_Exclude
				INT32 GetClipMethod (
					) const;

				//! Returns clipping OBJITEM
				const RVC::OBJITEM& GetClipObjItem (
					) const;

				//! Returns true if statistics has to be computed on destination object
				bool GetComputeStatistics (
					) const;

				//! Returns posible maximum edge length in destination object if optimization is set
				double GetMaximumEdgeLength (
					) const;

				//! Returns true if optimization is set
				bool GetOptimize (
					) const;

				//! Returns possible minimum edge length in destination object if optimization is set
				double GetMinimumEdgeLength (
					) const;

				double GetZTolerance (
					) const;

				//! Set to true if break lines has to be applyed to destination object.
				void SetApplyBreaklines (
					const bool apply
					) ;

				//! Set to true if clipping has to be applyed to destination object.
				void SetApplyClipping (
					const bool apply
					);

				//! Set break lines OBJITEM
				void SetBreaklinesObjItem (
					const RVC::OBJITEM& ObjItem
					) ;

				//! Set clip method: TINCLIPMODE_Include or TINCLIPMODE_Exclude
				void SetClipMethod (
					const INT32 method 
					);

				//! Set clipping OBJITEM
				void SetClipObjItem (
					const RVC::OBJITEM& ObjItem
					);

				//! Set to true if statistics has to be computed on destination object
				void SetComputeStatistics (
					const bool compute
					);

				//! Set desired maximum edge length in destination object when optimization is set
				void SetMaximumEdgeLength (
					const double length
					);

				//! Set to true for optimization 
				void SetOptimize (
					const bool optimize
					);

				//! Set desired minimum edge length in destination object when optimization is set
				void SetMinimumEdgeLength (
					const double length
					);;

				void SetZTolerance (
					const double tolerance
					); ;


			private:

				#ifndef GENERATING_DOXYGEN_OUTPUT

				bool m_Optimize;
				double m_ZTolerance;
				double m_MinimumEdgeLength;
				double m_MaximumEdgeLength;

				bool m_ApplyBreaklines;
				RVC::OBJITEM m_BreaklinesObjItem;

				bool m_ApplyClipping;
				RVC::OBJITEM m_ClipObjItem;
				INT32 m_ClipMethod;

				bool m_ComputeStatistics;

				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Base class for all triangulation methods.
		class METHOD_PARAMETERS {
			public:

				METHOD_PARAMETERS (
					const METHOD method
					);

				virtual ~METHOD_PARAMETERS (
					);

				//! Return enum of method in use
				METHOD GetMethod (
					) const;

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT

				const METHOD m_Method;

				METHOD_PARAMETERS (); // not implemented

				static METHOD Check (
					const METHOD method
					);

				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Class for Delaunay triangulation method from points.
		class DELAUNAY_PARAMETERS : public METHOD_PARAMETERS {
			public:

				//! Construct with default values.
				DELAUNAY_PARAMETERS (); 

				//! Destructor.
				~DELAUNAY_PARAMETERS ();

				//! Returns tolerance distance between points.
				double GetTolerance (
					) const { return (m_Tolerance); }

				//! Returns desired z scale for points.
				double GetZScale (
					) const { return (m_ZScale); }

				//! Returns desired z offset for points.
				double GetZOffset (
					) const { return (m_ZOffset); }

				//! Returns maximum memory to use for internal buffers in MB.
				double GetMaximumMemory (
					) const { return (m_MaximumMemory); }

				//! Set tolerance distance between points.
				double SetTolerance (
					const double tolerance
					);

				//! Set desired z scale for points.
				double SetZScale (
					const double ZScale
					);

				//! Set desired z offset for points.
				void SetZOffset (
					const double ZOffset
					) { m_ZOffset = ZOffset; }

				//! Set maximum memory to use for internal buffers in MB.
				void SetMaximumMemory (
					const double MaximumMemory
					) { m_MaximumMemory = MaximumMemory; }

			private:

				#ifndef GENERATING_DOXYGEN_OUTPUT

				double m_Tolerance;
				double m_ZScale;
				double m_ZOffset;
				double m_MaximumMemory;

				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Class for Adaptive Densification triangulation method from raster.
		class ADAPTIVEDENSIFICATION_PARAMETERS : public METHOD_PARAMETERS {
			public:

				ADAPTIVEDENSIFICATION_PARAMETERS (
					);

				~ADAPTIVEDENSIFICATION_PARAMETERS (
					);

				//! Returns accuracy distance between surface and triangle plane.
				double GetAccuracy (
					) const { return (m_Accuracy); }

				//! Returns node limit in triangulation.
				INT32 GetNodeLimit (
					) const { return (m_NodeLimit); }

				//! Set accuracy distance between surface and triangle plane.
				double SetAccuracy (
					const double accuracy
					);

				//! Set node limit in triangulation.
				INT32 SetNodeLimit (
					const INT32 NodeLimit
					);

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT

				double m_Accuracy;
				INT32 m_NodeLimit;

				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Generates triangulation
		static ERRVALUE Process (
		   const RVC::OBJITEM& SrcObjItem,
	      RVC::OBJITEM& DestObjItem, 
			const SOURCE_PARAMETERS* SrcParameters,
			const METHOD_PARAMETERS* MethodParameters,
			const DESTINATION_PARAMETERS& DestParameters 
			);

	    //! Generates Delaunay triangulation from geometric(vector, database)
	   static ERRVALUE ProcessDelaunay (
	      const RVC::OBJITEM& SrcObjItem,
	      RVC::OBJITEM& DestObjItem, 
			const SOURCE_PARAMETERS* SrcParameters,
			const DELAUNAY_PARAMETERS& MethodParameters,
			const DESTINATION_PARAMETERS& DestParameters 
			);

	     //! Generates triangulation from raster
	   static ERRVALUE ProcessAdaptiveDensification (
	      const RVC::OBJITEM& SrcObjItem,
	      RVC::OBJITEM& DestObjItem, 
			const SOURCE_PARAMETERS* SrcParameters,
			const ADAPTIVEDENSIFICATION_PARAMETERS& MethodParameters,
			const DESTINATION_PARAMETERS& DestParameters 
			);

	private:

		static ERRVALUE ProcessDestination (
	      RVC::OBJITEM& DestObjItem, 
			const DESTINATION_PARAMETERS& DestParameters 
			);
	       
	};

}  // end namespace OP
}  // end namespace RVC


#endif  // INC_RVC_OPTRIANGULATION_H
