/**
 * \file soa/arcimsserviceinfo.h
 * \brief SOA::ARCIMS::SERVICEINFO class definitions.
 *
 * \if NODOC
 * $Id: arcimsserviceinfo.h_v 1.7 2008/09/02 21:44:26 mju Exp $
 *
 * $Log: arcimsserviceinfo.h_v $
 * Revision 1.7  2008/09/02 21:44:26  mju
 * Fix doxygen warning.
 *
 * Revision 1.1  2007/09/12 17:09:07  dwilliss
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SOA_ARCIMSCAPABILITIES_H
#define	INC_SOA_ARCIMSCAPABILITIES_H

#ifndef	INC_MI32_MISTRINGLIST_H
	#include <mi32/mistringlist.h>
#endif

#ifndef	INC_MI32_SPATREF_H
	#include <mi32/spatref.h>
#endif

#ifndef	INC_MI32_RECT_H
	#include <mi32/rect.h>
#endif

class XMLDOC;
class XMLNODE;
class HTTPMESSAGE;
class SUBREGION2D;

namespace SOA {
namespace ARCIMS {

	class SERVICEINFO;

//=====================================================================================================================

//! Information about a field in an ArcIMS layer
class FIELD {
	public:

		//! Field Type
		enum TYPE {
			TYPE_ROWID = -99,
			TYPE_SHAPE = -98,
			TYPE_Bool = -7,		//!< true or false
			TYPE_BigInt = -5,		//!< INT64 or UINT64
			TYPE_Char = 1,			//!< Any character field
			TYPE_Int = 4,			//!< INT32 or UINT32
			TYPE_SmallInt = 5,	//!< INT16 or UINT16
			TYPE_Float = 6,
			TYPE_Double = 8,
			TYPE_VarChar = 12,		//!< For all practical purposes, the same as TYPE_Char
			TYPE_CBLOB = 13,		//!< Character Large OBject
			TYPE_NVARCHAR = 14,	//!< Variable-length Unicode character data
			TYPE_NCLOB = 15,		//!< Unicode character large object
			TYPE_Date = 91
			};

		//! Constructor.
		FIELD();

		//! Destructor.
		~FIELD();

		//! Get the field name.
		const MISTRING& GetName (
			) const { return m_Name; }

		//! Get the field type
		TYPE GetType (
			) const { return m_Type; }

		//! Return the size of the field in pixels
		INT32 GetSize (
			) const { return m_Size; }

		//! Get the precision (# of decimal places) for floating-point fields
		INT32 GetPrecision (
			) const { return m_Precision; }

		//! Parse values from XMLNODE.
		ERRVALUE ParseXML (
			const XMLNODE* node
			);
	private:
		#ifndef	GENERATING_DOXYGEN_OUTPUT
		// Elements
		MISTRING m_Name;
		TYPE m_Type;
		INT32 m_Size;
		INT32 m_Precision;
		#endif
	};

//=====================================================================================================================

//! The LAYER_INFO from an ArcIMS GET_SERVICE_INFO request
class LAYER {
	public:

		enum TYPE {
			TYPE_PointFeature,
			TYPE_LineFeature,
			TYPE_PolygonFeature,
			TYPE_Image,
			TYPE_Acetate
			};

		//! Constructor.
		LAYER ();

		//! Destructor.
		~LAYER ();

		//! Compute subregion for this layer in specified CRS.
		ERRVALUE ComputeSubRegion (
			const SPATREF::COORDREFSYS& CoordRefSys,
			SUBREGION2D& SubRegion
			) const;

		//! Get the bounding box for the layer.
		//! The bounding box is in the CRS of the service.
		const DRECT2D& GetBoundingBox (
			) const;

		//! Get Layer ID.
		//! The Layer ID is not meant to be shown to the user.
		const MISTRING& GetID (
			) const { return (m_ID); }

		//! Get maximum scale in meters/pixel.
		//! ArcIMS gives scales in different units depending on the CRS used.
		//! For this method, it's always converted to meters/pixel.
		double GetMaxScale (
			) const { return (m_MaxScale); }

		//! Get the maximum scale denominator
		double GetMaxScaleDenominator (
			) const;

		//! Get minimum scale in meters/pixel
		//! ArcIMS gives scales in different units depending on the CRS used.
		//! For this method, it's always converted to meters/pixel.
		double GetMinScale (
			) const { return (m_MinScale); }

		//! Get the minimum scale denominator
		double GetMinScaleDenominator (
			) const;

		//! Get Name.
		const MISTRING& GetName (
			) const { return (m_Name); }

		//! Get the layer type
		TYPE GetType (
			) const { return (m_Type); }

		//! Return whether or not the layer has a MaxScale
		bool HasMaxScale (
			) const { return (m_bHasMaxScale); }

		//! Return whether or not the layer has a MinScale
		bool HasMinScale (
			) const { return (m_bHasMinScale); }

		//! Return the default visibillity of the layer
		bool IsVisible (
			) const { return (m_bVisible); }

		//! Parse values from XMLNODE.
		ERRVALUE ParseXML (
			SERVICEINFO *pParent,
			const XMLNODE *xmlnode
			);

	private:
		#ifndef	GENERATING_DOXYGEN_OUTPUT

		// Attributes
		SERVICEINFO* m_pService;
		TYPE m_Type;
		MISTRING m_Name;
		bool m_bVisible, m_bHasMinScale, m_bHasMaxScale;
		MISTRING m_ID;
		double m_MinScale, m_MaxScale;
		DRECT2D m_Envelope;
		MILIST<FIELD> m_Fields;

		ERRVALUE ParseFCLASS (
			const XMLNODE* node
			);

		#endif
	};

typedef MILIST<LAYER> LAYERLIST;

//=====================================================================================================================

//! ARCIMS "Service" element.
class SERVICEINFO {
	public:

		//! The units of extents, features, etc.
		enum MAPUNITS {
			MAPUNIT_DecimalDegrees,
			MAPUNIT_Feet,
			MAPUNIT_Meters
			};

		//! Constructor.
		SERVICEINFO ();

		//! Increment reference count.
		void AddRef (
			) { ++m_RefCount; }

		//! Find a layer given its id. 
		//! Retuns nll if no such layer found.
		const LAYER* FindLayerByID (
			const MISTRING& id
			) const;

		//! Find a layer given its name. 
		//! Retuns nll if no such layer found.
		const LAYER* FindLayerByName (
			const MISTRING& name
			) const;

		//! Get the default DPI used in minscale/maxscale computations
		double GetDPI (
			) const { return (m_dpi); }

		//! Get the CRS used for reporting features and extents.
		const SPATREF::COORDREFSYS& GetFeatureCRS (
			) const { return (m_FeatureCRS); }

		//! Get the default CRS to be used sending a query to the server.
		//! This is usually the same as the Feature CRS, but not always.
		const SPATREF::COORDREFSYS& GetFilterCRS (
			) const { return (m_FilterCRS); }

		//! Get the initial extents envelope rectangle.
		//! Note, this may or may not be the extents of the full set of data.
		//! It's only supposed to be used as a good default initial extents for
		//! viewing the data.
		const DRECT2D& GetInitialExtent (
			) const { return (m_Envelope); }

		//! Get the list of layers available on the server
		const LAYERLIST& GetLayers (
			) const { return (m_Layers); }

		//! Get the units used in coordinates and minscale/maxscale calculations.
		MAPUNITS GetMapUnit (
			) const { return (m_MapUnit); }

		//! Get maximum image size in pixels (height x width)
		UINT32 GetMaxPixels (
			) const { return (m_MaxPixels); }

		//! Get Title.
		const MISTRING& GetTitle (
			) const { return (m_Title); }

		//! Get online resource.
		const MISTRING& GetURL (
			) const { return (m_URL); }

		//! Get the XML document of the SERVCEINFO reply.
		const XMLDOC& GetXML (
			) const { return (*m_pXMLDoc); }

		//! Set the XML of the SERVICEINFO reply.  
		//! The SERVICEINFO "owns" the document after this and will free it
		//! later
		ERRVALUE SetXML (
			XMLDOC* xmldoc
			);

		//! Decrement reference count, destroy if no more references.
		void Release (
			) {
			if ((--m_RefCount) == 0) delete this;
			}

		//! Set the URL of the service
		void SetURL (
			const MISTRING& url
			);

	private:
		#ifndef	GENERATING_DOXYGEN_OUTPUT
		// Elements
		UINT32 m_RefCount;
		MISTRING m_Title;
		MISTRING m_URL;
		UINT32 m_MaxPixels;
		double m_dpi;
		DRECT2D m_Envelope;
		SPATREF::COORDREFSYS m_FeatureCRS;
		SPATREF::COORDREFSYS m_FilterCRS;
		MAPUNITS m_MapUnit;
		LAYERLIST m_Layers;
		XMLDOC* m_pXMLDoc;

		ERRVALUE ParseEnvironment(const XMLNODE*);
		ERRVALUE ParseProperties(const XMLNODE*);

		//! Destructor.  Must use Release to free one
		~SERVICEINFO ();

		// Unimplemented
		SERVICEINFO (const SERVICEINFO& rhs);
		SERVICEINFO& operator= (const SERVICEINFO& rhs);

		friend class LAYER;	// To shup up the warning that the dtor is private and friendless
		#endif
	};

//=====================================================================================================================

}	// End namespace ARCIMS
}	// End namespace SOA

#endif	// INC_SOA_ARCIMSCAPABILITIES_H
