/**
 * \file imagepixel.h <spatmod/imagepixel.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::PIXEL class definition
 *
 * \if NODOC
 * $Id: imagepixel.h_v 1.9 2012/12/11 17:46:20 mju Exp $
 *
 * $Log: imagepixel.h_v $
 * Revision 1.9  2012/12/11 17:46:20  mju
 * Add FindSampleGroups.
 *
 * Revision 1.8  2007/01/24 21:24:46  mju
 * Add type_GrayscaleAlpha.
 *
 * Revision 1.7  2007/01/02 22:02:17  mju
 * Add exchange().
 *
 * Revision 1.6  2006/11/27 18:04:45  mju
 * Add setType.
 *
 * Revision 1.5  2006/08/04 17:11:59  mju
 * Change vhs to hsv.
 *
 * Revision 1.4  2006/08/03 13:26:00  mju
 * Clarify isSameType behavior.
 *
 * Revision 1.3  2006/05/01 14:44:48  mju
 * Add type_Opacity.
 *
 * Revision 1.2  2005/08/17 14:35:36  mju
 * Forgot bool on inlines.
 *
 * Revision 1.1  2005/08/16 15:10:19  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEPIXEL_H
#define	INC_SPATMOD_IMAGEPIXEL_H

#ifndef	INC_SPATMOD_DATATYPE_H
#include <spatmod/datatype.h>
#endif

namespace SPATMOD {
namespace IMAGE {

class SAMPLE;
class SAMPLEGROUPLIST;

//=====================================================================================================================

//! Container and properties for an image pixel.
//! A pixel is represented by a PIXEL::TYPE and a set of SAMPLEs.
class PIXEL {
	public:

		enum TYPE {
			TYPE_Invalid =				0x0000,
			TYPE_Single =				0x0101,	//!< Single sample, any type not listed below
			TYPE_Multiple =			0x0200,	//!< Multiple samples of combination not listed below
			TYPE_Grayscale =			0x0301,	//!< Single sample of SAMPLE::TYPE_Gray
			TYPE_GrayscaleAlpha =	0x8302,	//!< 2 samples: TYPE_Grayscale, Type_Opacity
			TYPE_Indexed =				0x0401,	//!< Single sample of SAMPLE::TYPE_Index
			TYPE_RGB =					0x0503,	//!< 3 samples: TYPE_Red, TYPE_Green, TYPE_Blue
			TYPE_RGBA =					0x8504,	//!< 4 samples: TYPE_Red, TYPE_Green, TYPE_Blue, TYPE_Opacity
			TYPE_CMY =					0x0603,	//!< 3 samples: TYPE_Cyan, TYPE_Magenta, TYPE_Yellow
			TYPE_CMYA =					0x8604,	//!< 4 samples: TYPE_Cyan, TYPE_Magenta, TYPE_Yellow, TYPE_Opacity
			TYPE_CMYK =					0x0704,	//!< 4 samples: TYPE_Cyan, TYPE_Magenta, TYPE_Yellow, TYPE_Black
			TYPE_CMYKA =				0x8705,	//!< 5 samples: TYPE_Cyan, TYPE_Magenta, TYPE_Yellow, TYPE_Black, TYPE_Opacity
			TYPE_ComplexMP =			0x0802,	//!< 2 samples: TYPE_ComplexMagnitude, TYPE_ComplexPhase
			TYPE_ComplexRI =			0x0902,	//!< 2 samples: TYPE_ComplexReal, TYPE_ComplexImaginary
			TYPE_IHS =					0x0A03,	//!< 3 samples: TYPE_Gray (intensity), TYPE_Hue, TYPE_Saturation
			TYPE_IHSA =					0x8A04,	//!< 4 samples: TYPE_Gray (intensity), TYPE_Hue, TYPE_Saturation, TYPE_Opacity
			TYPE_BHS =					0x0B03,	//!< 3 samples: TYPE_Gray (brightness), TYPE_Hue, TYPE_Saturation
			TYPE_BHSA =					0x8B04,	//!< 4 samples: TYPE_Gray (brightness), TYPE_Hue, TYPE_Saturation, TYPE_Opacity
			TYPE_HSV =					0x0C03,	//!< 3 samples: TYPE_Hue, TYPE_Saturation, TYPE_Gray (value)
			TYPE_HSVA =					0x8C04,	//!< 4 samples: TYPE_Hue, TYPE_Saturation, TYPE_Gray (value), TYPE_Opacity
			TYPE_Opacity =				0x8001,	//!< Single sample of SAMPLE::TYPE_Opacity
			TYPE_MASK_NumSamples =	0x00FF,
			TYPE_MASK_Alpha =			0x8000	//!< Alpha (opacity) included as last sample in pixel
			};

		//! Default constructor, create pixel with TYPE_Invalid and no samples
		PIXEL ();

		//! Construct pixel of specified type with all samples same DATATYPE.
		PIXEL (
			TYPE PixelType,						//!< Pixel type
			DATATYPE DataType,					//!< Sample data type
			UINT16 NumSamples = 0				//!< Number of samples, only used if cannot determine from PixelType
			);

		//! Construct pixel from specified samples.
		//! If PixelType not specified will be determined from sample types.
		PIXEL (
			UINT16 NumSamples,					//!< Number of samples
			const SAMPLE *samples,				//!< Sample array
			TYPE PixelType = TYPE_Invalid		//!< Pixel type
			);

		//! Copy constructor
		PIXEL (
			const PIXEL& rhs
			);

		//! Destructor.
		~PIXEL ();

		//! Assignment
		PIXEL& operator= (
			const PIXEL& rhs
			);

		//! Exchange with another pixel.
		void Exchange (
			PIXEL& other								//!< Pixel to exchange with
			);

		//! Find sample groups.
		void FindSampleGroups (
			SAMPLEGROUPLIST& GroupList,			//!< List returned
			bool IncludeOpacityInRange				//!< Whether to include opacity (if present) in range of samples for each group
			) const;

		//! Get data type of samples.
		//! @return sample DATATYPE or DATATYPE_Invalid if samples not all of same data type.
		DATATYPE GetDataType () const;

		//! Get size of single pixel in bytes.
		UINT32 GetNumBytes () const;

		//! Get number of samples.
		UINT16 GetNumSamples (
			) const { return (m_NumSamples); }

		//! Get pixel type.
		TYPE GetPixelType (
			) const { return (m_PixelType); }

		//! Get reference to specified sample (const).
		const SAMPLE& GetSample (
			UINT16 index								//! Sample index
			) const;

		//! Get reference to specified sample.
		SAMPLE& GetSample (
			UINT16 index								//! Sample index
			);

		//! Get array of SAMPLEs (const).
		const SAMPLE* GetSamples (
			) const { return (m_Samples); }

		//! Get array of SAMPLEs.
		SAMPLE* GetSamples (
			) { return (m_Samples); }

		//! Determine if alpha band is present based on pixel type.
		bool HasAlpha (
			) const { return ((m_PixelType & TYPE_MASK_Alpha) != 0); }

		//! Compare pixels for equality of pixel type and all samples.
		bool IsEqual (
			const PIXEL& rhs
			) const;

		//! Compare pixels for equality of pixel type and all samples (type, datatype, precision).
		//! Does not compare sample values.
		bool IsSameType (
			const PIXEL& rhs
			) const;

		//! Set sample.
		void SetSample (
			UINT16 index,								//! Sample index
			const SAMPLE& Sample						//! Sample value
			);

		//! Set pixel type.
		//! Specified type must be appropriate for number of samples.
		//! Sample types will be changed if required.
		ERRVALUE SetType (
			TYPE PixelType								//! New pixel type
			);

		//! Set all sample values to minimum.
		void SetValueMinimum ();

		//! Set all sample values to maximum.
		void SetValueMaximum ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SAMPLE *m_Samples;
		UINT16 m_NumSamples;
		TYPE m_PixelType;

		void Copy (const PIXEL& rhs);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//! Compare pixels for equality of pixel type and all samples.
inline bool operator== (
	const PIXEL& lhs,
	const PIXEL& rhs
	) { return (lhs.IsEqual(rhs)); }

//! Compare pixels for inequality.
inline bool operator!= (
	const PIXEL& lhs,
	const PIXEL& rhs
	) { return (!lhs.IsEqual(rhs)); }

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	 // INC_SPATMOD_IMAGEPIXEL_H
