/**
 * \file imagestage.h <spatmod/imagestage.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::STAGE definition.
 *
 * \if NODOC
 *
 * $Id: imagestage.h_v 1.18 2012/03/27 17:57:33 mju Exp $
 *
 * $Log: imagestage.h_v $
 * Revision 1.18  2012/03/27 17:57:33  mju
 * Add zoomCoordinates/Rectangle.
 *
 * Revision 1.17  2012/02/02 17:17:46  mju
 * Add enum for getBestZoom.
 *
 * Revision 1.16  2012/01/25 22:36:22  mju
 * Add v_computePixelRange and getBoundingRegion methods.
 *
 * Revision 1.15  2012/01/23 21:46:19  mju
 * Add computePixelRange using section.
 *
 * Revision 1.14  2012/01/23 20:45:55  mju
 * Add getBestZoom.
 *
 * Revision 1.13  2011/03/31 15:07:42  mju
 * Make pullBegin/End/SubSection non-inline so can do logging.
 *
 * Revision 1.12  2007/01/03 23:03:09  mju
 * Add getMetadata.
 *
 * Revision 1.11  2006/10/16 19:44:31  mju
 * Add description for doxygen.
 *
 * Revision 1.10  2006/10/11 14:10:14  mju
 * Move stage_autoptr and stage_array to own hdrs.
 *
 * Revision 1.9  2006/07/27 21:49:25  mju
 * Remove access methods.
 *
 * Revision 1.8  2006/07/14 21:49:54  mju
 * Add access and v_access methods.
 *
 * Revision 1.7  2006/01/26 21:00:34  mju
 * Add computeMultivariateStats.
 *
 * Revision 1.6  2006/01/25 17:34:19  mju
 * Add setSampleStats.
 *
 * Revision 1.5  2006/01/23 21:47:54  mju
 * Add methods for sample statistics.
 *
 * Revision 1.4  2006/01/11 18:48:24  mju
 * Add computePixelRange.
 *
 * Revision 1.2  2005/08/19 14:26:41  mju
 * Remove setTotalSection.
 *
 * Revision 1.1  2005/08/16 15:11:04  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGESTAGE_H
#define	INC_SPATMOD_IMAGESTAGE_H

#ifndef	INC_SPATMOD_IMAGEBASE_H
#include <spatmod/imagebase.h>
#endif

#ifndef	INC_SPATMOD_METADATACONTENT_H
#include <spatmod/metadatacontent.h>
#endif

#ifndef	INC_SPATMOD_METADATAFORMAT_H
#include <spatmod/metadataformat.h>
#endif

#ifndef INC_MI32_DELEGATE_H
#include <mi32/delegate.h>
#endif

class STAT_UNIVARIATE;
class STAT_MULTIVAR;

namespace SPATMOD {
namespace IMAGE {

class SECTION;
class SECTION_ITERATOR;
class BUFFER;

//=====================================================================================================================

//! Base class for image supporting "pull" operations.
class STAGE : public BASE {
	public:

		//! Enumeration for use with GetBestZoom.
		enum BESTZOOM {
			BESTZOOM_Closest,				//!< Closest absolute
			BESTZOOM_ClosestTo1,			//!< Closest to 1 (unity)
			BESTZOOM_Below75Pct,			//!< Highest value below desired and above 75% of desired, if any, otherwise closest to 1
			};

		// Default constructor.
		STAGE ();

		// Destructor.
		virtual ~STAGE ();

		//! Compute multivariate statistics for entire image.
		//! This will read entire image once.
		//! Initialize() must be done before calling this function.
		ERRVALUE ComputeMultivariateStats (
			STAT_MULTIVAR& Stats							//!< Statistics returned
			);

		//! Compute pixel range for image suitable for specified range type.
		//! Will read entire image if necessary.
		//! Initialize() must be done before calling this function.
		ERRVALUE ComputePixelRange (
			RANGETYPE RangeType = RANGETYPE_Exact	//!< Required range type
			);

		//! Compute pixel range for specified section.
		//! The specified "accuracy" value may affect the method used to compute the range.
		//! If RANGETYPE_Typical is specified and the image supports zoom values lower than that
		//! specified, additional zooming may be used to improve performance.
		//! If RANGETYPE_Maximal is used, implementations may use previously computed aggregate
		//! ranges that cover the specified section, possibly up to the entire image extent.
		//! Initialize() must be done before calling this function.
		ERRVALUE ComputePixelRange (
			const SECTION& section,						//!< Image section, including zoom factor if desired
			PIXEL& PixelValueMin,						//!< Minimum pixel values returned
			PIXEL& PixelValueMax,						//!< Maximum pixel values returned
			RANGETYPE accuracy = RANGETYPE_Maximal	//!< Desired range accuracy
			) { return (v_ComputePixelRange(section,PixelValueMin,PixelValueMax,accuracy)); }

		//! Compute statistics for image samples.
		//! Will read entire image if necessary.
		//! Initialize() must be done before calling this function.
		ERRVALUE ComputeSampleStats (
			STATSVALIDITY StatsValidity				//!< Required validity
			);

		//! Get best allowed zoom value for specified zoom.
		double GetBestZoom (
			double ZoomDesired,							//!< Desired zoom value
			BESTZOOM bestzoom								//!< How to determine best zoom
			) const;

		//! Get bounding region for image.
		//! Default implementation returns image rectangular extents.
		ERRVALUE GetBoundingRegion (
			REGION2D& region								//!< Region in image coordinates returned
			) const { return (v_GetBoundingRegion(region)); }

		//! Get metadata as string.
		void GetMetadata (
			MISTRING& result,								//!< Resulting string returned
			METADATACONTENT content,					//!< Requested content
			METADATAFORMAT format						//!< Formatting
			) const;

		//! Get statistics for sample.
		//! @return Pointer to statistics, or NULL if none available
		const STAT_UNIVARIATE* GetSampleStats (
			UINT16 index									//!< Sample index
			) const;

		//! Get validity for sample statistics.
		STATSVALIDITY GetSampleStatsValidity (
			) const { return (m_SampleStatsValidity); }

		//! Initialize stage.
		//! Must be be called after construction and optional parameter settings before stage is inserted into pipeline.
		ERRVALUE Initialize ();

		//! Pull section of image into buffer.
		//! If buffer has already been created it must match the specified section in size,
		//! and have the same pixel properties as the image.  If buffer has not been created
		//! it will be created with the appropriate size and pixel properties.
		ERRVALUE Pull (
			const SECTION& section,				//!< Section to pull data from
			BUFFER& buffer							//!< Buffer to put data into
			);

		//! Begin pull of a specific section.
		//! Should not be called directly except within v_PullBegin by class derived from FILTER.
		//! May be called multiple times.
		ERRVALUE PullBegin (
			const SECTION_ITERATOR& iterator
			);

		//! Finish pulling subsections.
		//! Should not be called directly except within v_PullEnd by class derived from FILTER.
		ERRVALUE PullEnd (
			);

		//! Pull data for specified subsection into buffer.
		//! Should not be called directly except within v_PullSubSection by class derived from FILTER.
		ERRVALUE PullSubSection (
			const SECTION& SubSection,
			BUFFER& buffer
			);

		//! Apply zoom to coordinates as points.
		ERRVALUE ZoomCoordinates (
			DPOINT2D *PtsImage,							//!< Points passed/returned in source/target image coordinates
			UINT32 NumPoints,								//!< Number of points
			double TgtZoom,								//!< Target zoom factor
			double SrcZoom = 1							//!< Source zoom factor
			) const;

		//! Apply zoom to rectangle.
		ERRVALUE ZoomRectangle (
			DRECT2D& RectImage,							//!< Rectangle passed/returned in source/target image coordinates
			double TgtZoom,								//!< Target zoom factor
			double SrcZoom = 1							//!< Source zoom factor
			) const;

	protected:

		//! Get metadata for this stage only.
		void GetMetadataThisOnly (
			MISTRING& result,								//!< Resulting string returned
			METADATACONTENT content,					//!< Requested content
			METADATAFORMAT format						//!< Formatting
			) const;

		//! Set sample statistics array and validity.
		//! Should only be used in implementations where v_Initialize can efficiently determine statistics.
		ERRVALUE SetSampleStats (
			const STAT_UNIVARIATE *SampleStatsArray,
			STATSVALIDITY StatsValidity
			);

		//! Set validity of sample statistics.
		//! Should only be called by override of v_ComputeSampleStats.
		void SetSampleStatsValidity (
			STATSVALIDITY StatsValidity
			);

		//! Compute pixel range for specified image section.
		//! Implementations are not required to honor RANGETYPE_Exact requests if deemed too
		//! inefficient and may instead behave as though RANGETYPE_Maximal is used.
		//! Default implementation returns range for entire image if available and specified
		//! section is large, otherwise the requested section is read and the exact range determined.
		virtual ERRVALUE v_ComputePixelRange (
			const SECTION& section,						//!< Section, including desired zoom value if allowed
			PIXEL& PixelValueMin,						//!< Minimum pixel value returned
			PIXEL& PixelValueMax,						//!< Maximum pixel value returned
			RANGETYPE accuracy							//!< Desired range accuracy
			);

		//! Get bounding region for image.
		//! Default implementation returns image rectangular extents.
		virtual ERRVALUE v_GetBoundingRegion (
			REGION2D& region								//!< Region in image coordinates returned
			) const;

		//! Query image for specified capability.
		virtual bool v_QueryCapability (
			CAPABILITY capability
			) const;

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		STATSVALIDITY m_SampleStatsValidity;
		STAT_UNIVARIATE *m_SampleStatsArray;

		// BASE overrides.
		virtual const char* v_GetTypeName () const;

		// Unimplemented
		STAGE (const STAGE&);
		STAGE& operator= (const STAGE&);
		#endif // GENERATING_DOXYGEN_OUTPUT

		// STAGE Overridables

		//! Compute statistics for image samples.
		//! May read entire image if necessary.
		//! Default implementation will compute at less than 1X zoom if supported unless STATSVALIDITY_Exact used.
		//! Most implementations should override to maximize efficiency.
		virtual ERRVALUE v_ComputeSampleStats (
			STAT_UNIVARIATE *SampleStatsArray,		//!< Array to contain statistics, dimension matches number of samples
			STATSVALIDITY StatsValidity				//!< Required validity
			);

		//! Get metadata as string.
		virtual void v_GetMetadata (
			MISTRING& result,								//!< Resulting string returned
			METADATACONTENT content,					//!< Requested content
			METADATAFORMAT format						//!< Formatting
			) const = 0;

		//! Get metadata for this stage only.
		virtual void v_GetMetadataThisOnly (
			MISTRING& result,								//!< Resulting string returned
			METADATACONTENT content,					//!< Requested content
			METADATAFORMAT format						//!< Formatting
			) const;

		//! Initialize stage before insertion into pipeline.
		//! Must do the following (in order), checking for error after each call.
		//! Validate all class-specific settings.
		//! Open any data sources/files if required.
		//! Call v_Initialize() on base class (if not private in base).
		//! Call SetAllProperties OR call SetPixelProperties(), SetTotalDimensions(), and SetTransToCRS().
		//! Optionally call SetTileDimensions, SetPixelRange and/or other Set... methods needed.
		//! Preprocessing of source stages to determine properties may be done here or the first time v_PullBegin is
		//! called if does not affect base image properties.
		virtual ERRVALUE v_Initialize (
			) = 0;

		//! Begin pull of a specific section.
		//! May be called multiple times.
		//! If any extra processing is done it may be done the first time this is called if not done in v_Initialize.
		virtual ERRVALUE v_PullBegin (
			const SECTION_ITERATOR& iterator		//!< SubSection iterator to be used
			) = 0;

		//! Finish pulling subsections.
		virtual ERRVALUE v_PullEnd (
			) = 0;

		//! Pull data for specified subsection into buffer.
		//! Buffer provided may be larger than needed for subsection, in which case
		//! data not overwritten in buffer may be left unchanged for efficiency.
		virtual ERRVALUE v_PullSubSection (
			const SECTION& SubSection,				//!< Current subsection to read
			BUFFER& buffer								//!< Buffer to read into
			) = 0;

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGESTAGE_H

