/**
 * \file imagetargetj2k.h <spatmod/imagetargetj2k.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::TARGET_J2K definition.
 *
 * \if NODOC
 *
 * $Id: imagetargetj2k.h_v 1.10 2011/10/11 14:50:44 mju Exp $
 *
 * $Log: imagetargetj2k.h_v $
 * Revision 1.10  2011/10/11 14:50:44  mju
 * Remove unneeded hdrs.
 *
 * Revision 1.9  2009/08/13 16:46:15  mju
 * Base factory not nested.
 *
 * Revision 1.8  2009/08/13 15:38:19  mju
 * Add factory.
 *
 * Revision 1.7  2007/08/15 20:18:27  mju
 * Add setting for target "quality".
 *
 * Revision 1.6  2007/06/21 13:27:19  mju
 * Un-nest parms class so can forward-delcare.
 *
 * Revision 1.5  2007/04/11 22:54:43  dfriberg
 * Remove unneeded members
 *
 * Revision 1.4  2007/01/26 23:27:14  dfriberg
 * Add virtual methods to be called when writing one output file per tile
 *
 * Revision 1.3  2006/12/08 18:58:54  mju
 * Add curTileExtent member.
 *
 * Revision 1.2  2006/05/01 13:45:27  mju
 * Include point hdr.
 *
 * Revision 1.1  2006/04/27 20:36:00  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGETARGETJ2K_H
#define	INC_SPATMOD_IMAGETARGETJ2K_H

#ifndef	INC_SPATMOD_IMAGETARGETFILE_H
#include <spatmod/imagetargetfile.h>
#endif

class JP2_TARGET;

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

class TARGETPARM_J2K {
	public:

		//! Methods for embedding georeference in JPEG2000 file.
		enum GEOMETHOD {
			GEOMETHOD_None =			0x00,
			GEOMETHOD_GML =			0x01,
			GEOMETHOD_GeoTIFF =		0x02,
			GEOMETHOD_Automatic =	0x00,		//!< Automatic based on whether georeference is affine or CtrlPoint
			GEOMETHOD_Affine =		0x10,		//!< Always use affine
			};

		//! Progression order.
		// (Values are chosen to match J2K_TARGET::PROGORDER defined in mi32p/j2kimage.h)
		enum PROGORDER {
			PROGORDER_LRCP = 0,
			PROGORDER_RLCP = 1,
			PROGORDER_RPCL = 2,
			PROGORDER_PCRL = 3,
			PROGORDER_CPRL = 4,
			PROGORDER_Auto = 99
			};

		//! Constructor.
		TARGETPARM_J2K ();

		//! Set embedded georeference method.
		//! Default is to use both GML and GeoTIFF embedding with automatic.
		void SetGeoMethod (
			GEOMETHOD GeoMethod
			) { m_GeoMethod = GeoMethod; }
		
		//! Set whether to always include packet length (PLT) markers.
		//! Including packet length markers allows better random access to the J2K file.
		//! Default will include packet length markers only for images > 1M pixels.
		void SetIncludePacketLengthMarkers (
			bool IncludePLT = true
			) { m_IncludePLT = IncludePLT; }

		//! Set whether to always add tile-part headers for resolution-progressive orderings.
		//! This introduces tile-part headers immediately before each resolution level and locates
		//! the packet length information with the header of the tile-part to which the packets belong.
		//! This has the effect of delaying the loading and parsing of packet length identifiers
		//! (hundreds of thousands of packets for a typical 500 MByte image) until an interactive
		//! viewer or browser requests the relevant resolution.
		//! Default is to include only for images > 4M pixels
		void SetIncludeTilePartHeaders (
			bool IncludeTPH = true
			) { m_IncludeTPH = IncludeTPH; }

		//! Set number of detail layers.
		//! Defaults to 1 detail layer if not called.
		void SetNumDetailLayers (
			int NumDetailLayers
			) { m_NumDetailLayers = NumDetailLayers; }

		//! Set number of resolution levels.
		//! Default will automatically determine based on dimensions such that lowest resolution
		//! level will have dimensions < 512x512.
		void SetNumResolutionLevels (
			int numlevels
			) { m_NumResLevels = numlevels; }

		//! Set whether to always use precincts or not.
		//! Default is to only use precincts for images larger than 4M pixels.
		void SetPrecinctUsage (
			bool UsePrecincts
			) { m_UsePrecincts = UsePrecincts; }

		//! Set precinct size for each resolution level.
		//! Last array entry is replicated through remaining levels.
		void SetPrecinctSizes (
			const SIMPLE_ARRAY<LPOINT2D>& PrecinctSizes
			) { m_PrecinctSizes = PrecinctSizes; }

		//! Set packet progression order.
		//! Default is Auto if not called.
		void SetProgressionOrder (
			PROGORDER ProgOrder
			) { m_ProgOrder = ProgOrder; }

		//! Set whether compression is 'reversible' (lossless) or not.
		//! Default is 'irreversible' (lossy) if method not used.
		void SetReversible (
			bool reversible = true
			) { m_Reversible = reversible; }

		//! Set target compression quality.
		void SetTargetQuality (
			double TargetQuality							//!< Relative quality (1-100)
			) { m_TargetQuality = TargetQuality; m_TargetRatio = 0; m_TargetSizeBytes = 0; }

		//! Set target compression ratio.
		void SetTargetRatio (
			UINT16 TargetRatio
			) { m_TargetRatio = TargetRatio; m_TargetQuality = 0; m_TargetSizeBytes = 0; }

		//! Set target codestream size in bytes.
		void SetTargetSize (
			INT64 TargetSizeBytes						//! Target size, 0 to retain maximum quality (not necessarily lossless)
			) { m_TargetSizeBytes = TargetSizeBytes; m_TargetQuality = 0; m_TargetRatio = 0; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<LPOINT2D> m_PrecinctSizes;
		double m_TargetQuality;
		INT64 m_TargetSizeBytes;
		UINT16 m_TargetRatio;
		int m_NumDetailLayers;
		int m_NumResLevels;
		PROGORDER m_ProgOrder;
		GEOMETHOD m_GeoMethod;
		bool m_IncludePLT;
		bool m_IncludeTPH;
		bool m_UsePrecincts;
		bool m_Reversible;
		friend class TARGET_J2K;
		#endif
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(TARGETPARM_J2K::GEOMETHOD)
#endif // GENERATING_DOXYGEN_OUTPUT

//=====================================================================================================================

//! JPEG2000 image file target.
//! Supports DATATYPE of UINT8, SINT8, UINT16, SINT16 only.
//! Supports PIXEL::TYPE of Single, Grayscale, RGB only.
class TARGET_J2K : public TARGET_FILE {
	public:

		//! Factory class for constructing TARGET instance.
		class FACTORY : public FACTORY_TARGET_FILE {
			public:

				//! Construct with specified settings.
				explicit FACTORY (	
					const TARGETPARM_J2K& settings
					);

				//! Copy constructor.
				FACTORY (const FACTORY& rhs);

				//! Destructor.
				virtual ~FACTORY ();

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				TARGETPARM_J2K m_Settings;

				// FACTORY_TARGET_FILE overrides.
				virtual FACTORY_TARGET_FILE * v_Clone () const;
				virtual TARGET_FILE * v_ConstructTarget (STAGE *pInputStage, const FILEPATH& filepath, GEOFORMATS GeoFormats) const;

				// Unimplemented.
				FACTORY& operator= (const FACTORY& rhs);
				#endif
			};

		//! Construct target with specified path and quality.
		TARGET_J2K (
			STAGE *pInputStage,							//!< Stage in pipeline to obtain image data from
			const FILEPATH& filepath,					//!< Path to target file
			const TARGETPARM_J2K& settings,					//!< Settings
			GEOFORMATS GeoFormats = GEOFORMAT_None	//!< Georeference formats
			);

		//! Destructor.
		virtual ~TARGET_J2K ();

		//! Set maximum memory usage.
		void SetMaxMemUsage (
			int MaxMemUsageMB							// Maximum memory usage in MB (minimum of 64)
			) { m_MaxMemUsageMB = MaxMemUsageMB; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		TARGETPARM_J2K m_Settings;
		int m_MaxMemUsageMB;
		JP2_TARGET *m_pFile;

		ERRVALUE Close ();
		ERRVALUE MakeFile (const FILEPATH& filepath, const LRECT2D& extent);

		// TARGET overrides.
		virtual ERRVALUE v_FillTileRows ( INT32 NumColumns, INT32 NumRows);
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PushBegin (const SECTION_ITERATOR& iterator);
		virtual ERRVALUE v_PushEnd ();
		virtual ERRVALUE v_TileBegin (const FILEPATH& FilePath, const LRECT2D& Extents);
		virtual ERRVALUE v_TileEnd ();
		virtual ERRVALUE v_WriteSubSection (const SECTION& SubSection, const BUFFER& buffer);

		// Unimplemented
		TARGET_J2K (const TARGET_J2K&);
		TARGET_J2K& operator= (const TARGET_J2K&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGETARGETJ2K_H

