/**
 * \file spatmod/imagetileset.h
 * \brief SPATMOD::IMAGE::TILESET base definitions
 *
 * \if NODOC
 * $Id: imagetileset.h_v 1.102 2012/11/06 21:00:50 bcolombe Exp $
 *
 * $Log: imagetileset.h_v $
 * Revision 1.102  2012/11/06 21:00:50  bcolombe
 * Add GetCellSize for list item
 *
 * Revision 1.101  2012/10/25 22:57:21  bcolombe
 * Add copyright, orderx/y, offsetx/y
 *
 * Revision 1.100  2012/09/04 15:43:51  bcolombe
 * Add ComputeGeoref that takes tile range
 *
 * Revision 1.99  2012/07/10 15:02:59  bcolombe
 * Add GetObjItem
 *
 * Revision 1.98  2012/07/09 19:18:05  bcolombe
 * Changes to Generate/Scan tileset test
 *
 * Revision 1.97  2012/06/21 22:08:32  bcolombe
 * Changes to filter
 *
 * Revision 1.96  2012/06/20 21:32:05  bcolombe
 * Add type and filter
 *
 * Revision 1.95  2012/05/21 16:09:55  bcolombe
 * *** empty log message ***
 *
 * Revision 1.94  2012/05/17 16:14:13  bcolombe
 * ScanTileset: allow showing all errors
 *
 * Revision 1.93  2012/05/14 15:59:07  bcolombe
 * Add GenerateTilesetTest, ScaneTilesetTest
 *
 * Revision 1.92  2012/05/03 18:02:51  bcolombe
 * Add reference loc to tsd
 *
 * Revision 1.91  2012/04/27 16:35:34  bcolombe
 * Use scale numerator,denominator
 *
 * Revision 1.90  2012/04/26 17:00:37  bcolombe
 * Remove precision, just use scale
 *
 * Revision 1.89  2012/04/20 21:27:08  bcolombe
 * Changes for min/max values
 *
 * Revision 1.88  2012/04/20 17:00:06  bcolombe
 * Add base serverurl for making aux files
 *
 * Revision 1.87  2012/04/03 21:33:56  bcolombe
 * Add new terrain node
 *
 * Revision 1.86  2012/04/03 20:53:12  bcolombe
 * Changes for new terrain format
 *
 * Revision 1.85  2012/03/29 17:53:27  bcolombe
 * Fix for mac
 *
 * Revision 1.84  2012/03/29 17:37:08  bcolombe
 * Add precision
 *
 * Revision 1.83  2012/03/15 15:35:05  bcolombe
 * Add null value, scale to meters
 *
 * Revision 1.82  2012/03/14 22:00:09  bcolombe
 * Changes for 24,32 bit terrain
 *
 * Revision 1.81  2012/03/06 01:26:39  bcolombe
 * Changes for using tilemap
 *
 * Revision 1.80  2012/02/29 15:41:47  bcolombe
 * Add array to hold tilemap sources
 *
 * Revision 1.79  2012/02/27 17:20:38  bcolombe
 * Add changes for using tilemap,pixelmap for finding tiles
 *
 * Revision 1.78  2012/02/20 17:55:37  bcolombe
 * Support new png16 format
 *
 * Revision 1.77  2012/02/13 21:14:13  bcolombe
 * No longer need test criteria for finditem
 *
 * Revision 1.76  2012/02/06 20:49:49  bcolombe
 * Add TestNode
 *
 * Revision 1.75  2012/02/06 20:00:00  bcolombe
 * Add GenerateRegionTileMap, TerrainXML
 *
 * Revision 1.74  2012/02/03 21:56:08  bcolombe
 * Add enums for list
 *
 * Revision 1.73  2012/02/03 21:04:48  bcolombe
 * FindItem: use region
 *
 * Revision 1.72  2012/02/03 15:27:13  bcolombe
 * *** empty log message ***
 *
 * Revision 1.71  2012/02/02 20:31:05  bcolombe
 * LIST Read: add options to include wms, arcims, proprietary
 *
 * Revision 1.70  2012/02/02 18:50:45  bcolombe
 * Add LIST for reading webtilesets xml
 *
 * Revision 1.69  2012/01/30 20:22:09  bcolombe
 * Add pixel range
 *
 * Revision 1.68  2012/01/30 17:23:27  bcolombe
 * Add GeneratePixelRangeMap
 *
 * Revision 1.67  2012/01/23 18:21:46  bcolombe
 * Add option to clear tile info when checking HaveTiles
 *
 * Revision 1.66  2012/01/11 23:27:34  bcolombe
 * Add PNG 16 bit
 *
 * Revision 1.65  2011/12/01 22:23:08  bcolombe
 * Add SetMargin
 *
 * Revision 1.64  2011/09/15 22:00:55  ldyas
 * added margin
 *
 * Revision 1.63  2011/09/02 18:29:00  bcolombe
 * Added 8bit,24bit imgformats
 *
 * Revision 1.62  2011/08/25 19:33:16  ldyas
 * added svg support
 *
 * Revision 1.61  2011/08/03 16:01:47  bcolombe
 * Add method to set tile ranges when known
 *
 * Revision 1.60  2011/06/20 20:27:32  bcolombe
 * Add FindKMLFile
 *
 * Revision 1.59  2011/04/22 19:29:45  bcolombe
 * Add IsOnBorder when computing tile values
 *
 * Revision 1.58  2011/02/04 18:40:01  bcolombe
 * Add ReadTSD for path/string
 *
 * Revision 1.57  2011/02/04 18:23:36  bcolombe
 * Read TSD from objitem
 *
 * Revision 1.56  2011/01/14 23:08:31  bcolombe
 * Add option to not show extents in properties string
 *
 * Revision 1.55  2010/12/22 21:22:46  bcolombe
 * Add MinLevel to CreateAux
 *
 * Revision 1.54  2010/11/01 23:00:41  bcolombe
 * Imgformat changes
 *
 * Revision 1.53  2010/10/26 15:19:20  bcolombe
 * Allow to specify palette use in CreateReducedLevel
 *
 * Revision 1.52  2010/10/18 18:56:07  bcolombe
 * Add method to get description
 *
 * Revision 1.51  2010/09/15 23:16:19  bcolombe
 * Added TSDURL
 *
 * Revision 1.50  2010/08/11 21:45:59  bcolombe
 * Added GetProperties that takes url
 *
 * Revision 1.49  2010/07/26 18:05:15  bcolombe
 * Changes for showing status when zipping
 *
 * Revision 1.48  2010/07/20 15:37:34  bcolombe
 * Add zip option to CreateReducedLevel
 *
 * Revision 1.47  2010/07/09 20:37:47  bcolombe
 * Added GenerateTileMap
 *
 * Revision 1.46  2010/07/07 22:27:08  bcolombe
 * Allow zip/unzip of tileset to use multiple threads
 *
 * Revision 1.45  2010/07/06 22:30:17  bcolombe
 * Add WriteKML
 *
 * Revision 1.44  2010/06/29 15:13:45  bcolombe
 * Added SetPath
 *
 * Revision 1.43  2010/06/28 18:28:52  bcolombe
 * Add UseExtents to ComputeCellSize
 *
 * Revision 1.42  2010/06/23 16:35:18  bcolombe
 * Added m_DateDoc to hold datelist/range node
 *
 * Revision 1.41  2010/06/21 22:59:34  bcolombe
 * Change to daterange
 *
 * Revision 1.40  2010/06/21 17:23:01  bcolombe
 * Added m_GeogExtents
 *
 * Revision 1.39  2010/06/18 20:12:14  bcolombe
 * Added getting of date for locations
 *
 * Revision 1.38  2010/05/18 16:36:59  bcolombe
 * Add CreateReducedLevel that takes source & target locations
 *
 * Revision 1.37  2010/05/03 17:20:13  bcolombe
 * Added CloseZipFiles
 *
 * Revision 1.36  2010/04/29 21:14:33  bcolombe
 * Added zipfile and zipfilepath members
 *
 * Revision 1.35  2010/04/23 22:06:29  bcolombe
 * Add zip/unzip that take targetpath
 *
 * Revision 1.34  2010/04/19 20:36:00  bcolombe
 * Add FindTileFile that returns name
 * For use in finding zip tile
 *
 * Revision 1.33  2010/04/19 16:33:55  bcolombe
 * Added bOverwriteExisting to CreateAuxiliaryFiles
 *
 * Revision 1.32  2010/04/07 19:46:28  bcolombe
 * Added Zip and Unzip
 *
 * Revision 1.31  2010/04/07 17:55:44  bcolombe
 * Add ZipLevel
 *
 * Revision 1.30  2010/04/06 23:21:54  bcolombe
 * Added CacheLocation
 *
 * Revision 1.29  2010/04/02 22:26:49  bcolombe
 * Added CreateReducedLevel, CreateAuxiliaryFiles and GetTarget
 *
 * Revision 1.28  2009/12/30 19:09:12  bcolombe
 * Added new WriteTSD methods
 *
 * Revision 1.27  2009/12/29 23:50:31  bcolombe
 * Added GetLocal/RemoteLocationString
 *
 * Revision 1.26  2009/12/29 17:08:20  bcolombe
 * Removed HasTargetTile, use HasTile
 *
 * Revision 1.25  2009/12/18 18:05:52  bcolombe
 * Added RegisterTSDFileType
 *
 * Revision 1.24  2009/12/18 17:43:10  bcolombe
 * Added STRUCTURE, GetStructure and GetProperties
 *
 * Revision 1.23  2009/12/17 21:27:59  bcolombe
 * Added metadata
 *
 * Revision 1.22  2009/12/14 21:06:29  bcolombe
 * Added WriteTSD that takes MISTRINGLIST of locations
 *
 * Revision 1.21  2009/12/11 22:49:30  bcolombe
 * Added additional HasTiles methods
 *
 * Revision 1.20  2009/12/07 21:10:42  bcolombe
 * Added HasSourceTiles and HasTargetTile
 *
 * Revision 1.19  2009/11/30 20:26:47  bcolombe
 * Added GetHREFStrings and GetPathStrings
 *
 * Revision 1.18  2009/11/05 19:00:00  bcolombe
 * GetHREFStr no longer takes level
 *
 * Revision 1.17  2009/11/04 23:38:52  bcolombe
 * Change to use TSD
 *
 * Revision 1.16  2009/10/22 19:12:46  bcolombe
 * Added ReadTMS and GetHREFStr
 *
 * Revision 1.15  2009/09/03 21:52:35  bcolombe
 * Added GetBaseTileCount
 *
 * Revision 1.14  2009/08/31 18:16:43  bcolombe
 * Added SetBaseTileCount
 *
 * Revision 1.13  2009/08/17 20:38:04  bcolombe
 * Remove previous change
 * WorldWind tileset description was incorrect, it should work the same as Geographic profile
 *
 * Revision 1.12  2009/08/14 19:35:54  bcolombe
 * Added Global_Geographic_Square profile
 *
 * Revision 1.11  2009/08/07 20:49:20  mju
 * Add minLevel parm to writeTMS.
 *
 * Revision 1.10  2009/08/06 15:34:19  mju
 * Add computeExtents with tileidx as lpoint3d.
 *
 * Revision 1.9  2009/08/06 13:33:54  mju
 * Add getFormat.
 *
 * Revision 1.8  2009/08/05 19:18:22  mju
 * Add getCellSizeDisplayStr.
 *
 * Revision 1.7  2009/08/03 19:41:07  mju
 * Return error from range::set.
 *
 * Revision 1.6  2009/07/29 19:38:29  bcolombe
 * Add TileFolder to WriteTMS so tilesets can be given path relative to TMS file
 *
 * Revision 1.5  2009/07/20 18:57:13  mju
 * Add computeTileExtents using crs.
 *
 * Revision 1.4  2009/07/17 21:33:52  mju
 * Add computeExtents.
 *
 * Revision 1.3  2009/07/09 21:30:16  mju
 * More basic methods.
 *
 * Revision 1.1  2009/07/01 20:42:45  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGETILESET_H
#define	INC_SPATMOD_IMAGETILESET_H

#ifndef	INC_SPATMOD_IMAGEFORMAT_H
#include <spatmod/imageformat.h>
#endif

#ifndef	INC_MI32_SPATREF_H
#include <mi32/spatref.h>
#endif

#ifndef	INC_MI32_RECT_H
#include <mi32/rect.h>
#endif

#ifndef INC_MI32_XML_H
#include <mi32/xml.h>
#endif

#ifndef INC_MI32_MISTRINGLIST_H
#include <mi32/mistringlist.h>
#endif

#ifndef INC_MI32_BITSET_H
#include <mi32/bitset.h>
#endif

#ifndef INC_MI32_FILEPATH_H
#include <mi32/filepath.h>
#endif

#ifndef INC_MI32_ZIPFILE_H
#include <mi32/zipfile.h>
#endif

#ifndef INC_MI32_DATETIME_H
#include <mi32/datetime.h>
#endif

#ifndef INC_MI32_COLOR_H
#include <mi32/color.h>
#endif

#ifndef INC_MI32_POLYLINE_H
#include <mi32/polyline.h>
#endif

#ifndef INC_RVC_OBJITEM_H
#include <rvc/objitem.h>
#endif

#ifndef INC_MI32_MITHREAD_H
#include <mi32/mithread.h>
#endif

#ifndef INC_MI32_TEXTID_H
#include <mi32/textid.h>
#endif

#ifndef INC_MGUI_DLGGETOBJECTFILTER_H
#include <mgui/dlggetobjectfilter.h>
#endif

// Forward declarations
#ifndef GENERATING_DOXYGEN_OUTPUT
class REGION2D;
class TRANS2D_AFFINE;
class MISTATUSCONTEXT;
namespace SPATMOD {
	class GEOREFERENCE;
	namespace IMAGE {
		class SAMPLEITERATOR;
		class SOURCE_TILESET;
		}
	}
#endif

namespace SPATMOD {
namespace IMAGE {

class PIXEL;

//=====================================================================================================================

//! Container for tileset definition.
class TILESET {
	public:

		static void TSDFileTypeGetDescription (MISTRING& description, const RVC::OBJITEM& objitem);
		static void TSDFileTypeGetProperties (MISTRING& properties, const FILEPATH& filepath);
		static void TSDFileTypeGetProperties (MISTRING& properties, const RVC::OBJITEM& objitem);
		static void TSDFileTypeGetProperties (MISTRING& properties, const RVC::OBJITEM& objitem, bool ShowExtents);
		static void RegisterTSDFileType ();
		static void SaveAs (const MISTRING& url, const FILEPATH& filepath);
		//! Tileset profile
		enum PROFILE {
			PROFILE_None =						0,
			PROFILE_Local =						1,
			PROFILE_Global_Geographic =			2,
			PROFILE_Global_WebMercator =		3
			};

		enum STRUCTURE {
			STRUCTURE_Hierarchical =			0,
			STRUCTURE_Flat =					1,
			STRUCTURE_WorldWind =				2
			};

		enum TARGET {
			TARGET_Undefined =					0,
			TARGET_GoogleMaps =					1,
			TARGET_GoogleEarth =				2,
			TARGET_BingMaps =					3,
			TARGET_NASAWorldWind =				4
			};

		//! Format of external image files.
		enum IMGFORMAT {
			IMGFORMAT_Invalid =			0x00000000,							//!< Invalid format
			IMGFORMAT_8bit =			0x00000001,							//!< 8 bit
			IMGFORMAT_16bit =			0x00000002,							//!< 16 bit
			IMGFORMAT_16bitA =			0x00000004,							//!< 16 bit
			IMGFORMAT_16bitOLD =		0x00000008,							//!< 16 bit
			IMGFORMAT_24bit =			0x00000010,							//!< 24 bit
			IMGFORMAT_32bit =			0x00000020,							//!< 32 bit
			IMGFORMAT_Terrain =			0x00000040,							//!< Terarin
			IMGFORMAT_PNG =				0x00000080,							//!< PNG file						
			IMGFORMAT_JPEG =			0x00000100,							//!< JPEG file
			IMGFORMAT_TIFF =			0x00000200,							//!< TIFF file
			IMGFORMAT_GIF =				0x00000400,							//!< JP2 file
			IMGFORMAT_JP2 =				0x00000800,							//!< JP2 file
			IMGFORMAT_SVG =				0x00001000,							//!< SVG file
			IMGFORMAT_SVGZ =			(IMGFORMAT_SVG|1),					//!< SVGZ file
			IMGFORMAT_PNG8 =			(IMGFORMAT_PNG|IMGFORMAT_8bit),					//!< 8-bit PNG file
			IMGFORMAT_PNG24 =			(IMGFORMAT_PNG|IMGFORMAT_24bit),					//!< 24-bit PNG file
			IMGFORMAT_PNG16Terrain =	(IMGFORMAT_PNG|IMGFORMAT_16bit|IMGFORMAT_Terrain),	//!< 16-bit Terrain
			IMGFORMAT_PNG16TerrainA =	(IMGFORMAT_PNG|IMGFORMAT_16bitA|IMGFORMAT_Terrain),	//!< 16-bit Terrain
			IMGFORMAT_PNG16TerrainOLD =	(IMGFORMAT_PNG|IMGFORMAT_16bitOLD|IMGFORMAT_Terrain),//!< 16-bit Terrain
			IMGFORMAT_PNG24Terrain =	(IMGFORMAT_PNG|IMGFORMAT_24bit|IMGFORMAT_Terrain),	//!< 24-bit Terrain
			IMGFORMAT_PNG32Terrain =	(IMGFORMAT_PNG|IMGFORMAT_32bit|IMGFORMAT_Terrain),	//!< 32-bit Terrain
			IMGFORMAT_TIFF8 =			(IMGFORMAT_TIFF|IMGFORMAT_8bit),					//!< 8-bit TIFF file
			IMGFORMAT_TIFF24 =			(IMGFORMAT_TIFF|IMGFORMAT_24bit),					//!< 24-bit TIFF file
			IMGFORMAT_Auto_PNG8_JPEG =	(IMGFORMAT_PNG8|IMGFORMAT_JPEG),	//!< Automatically use PNG or JPEG depending on whether need validity/opacity.  Uses 8-bit PNGs
			IMGFORMAT_Auto_PNG24_JPEG =	(IMGFORMAT_PNG24|IMGFORMAT_JPEG)	//!< Automatically use PNG or JPEG depending on whether need validity/opacity.  Uses 24-bit PNGs
			};

		enum TYPE {
			TYPE_None =		0x0000,
			TYPE_Image =	0x0001,
			TYPE_Terrain =	0x0002,
			TYPE_KML =		0x0004,
			TYPE_SVG =		0x0008,
			TYPE_DAE =		0x0010
			};

		//! struct for holding transparent color settings
		struct TransparentColor {
			COLOR color;
			double distance;
			TransparentColor ()
				: distance(0) { }
			};

		class FILTER : public MGUI::DLG_GETOBJECT_FILTER {
			public:
				FILTER ();
				~FILTER ();
				void AddType(TYPE type);
			private:
				TYPE m_Types;
				virtual bool v_IsFileSelectable(const FILEPATH& filepath, int position) const;
			};

		//! webtilesets.xml reader
		class LIST {
			public:
				enum GROUP {
					GROUP_Image = 0,
					GROUP_Terrain,
					GROUP_Server,
					GROUP_Mashup
					};

				enum ITEM {
					ITEM_Image = 0,
					ITEM_Terrain,
					ITEM_TerrainA,
					ITEM_KML,
					ITEM_Server
					};

				//! Default constructor.
				LIST ();

				//! Default destructor.
				~LIST ();

				//! Has list been read already?
				bool IsRead (
					) const { return (m_bIsRead); }

				//! Download and read webtileset.xml file
				ERRVALUE Read (
					bool bIncludeArcIMS = false,
					bool bIncludeWMS = false,
					bool bIncludeProprietary = false
					);

				//! Find group node
				XMLNODE* FindGroup (
					const XMLNODE* node = 0 //!< Parent node to search from.  Default is root.
					) const;

				//! Find group node
				XMLNODE* FindGroup (
					const char* id,			//!< id of node to find
					const XMLNODE* node = 0 //!< Parent node to search from.  Default is root.
					) const;

				//! Find group node
				XMLNODE* FindGroup (
					GROUP group,			//!< id of node to find
					const XMLNODE* node = 0 //!< Parent node to search from.  Default is root.
					) const;

				//! Find group node
				XMLNODE* FindNextGroup (
					const XMLNODE* node //!< Previous sibling node for finding next
					) const;

				//! Find group node
				XMLNODE* FindNextGroup (
					const char* id,			//!< id of node to find
					const XMLNODE* node		//!< Previous sibling node for finding next
					) const;

				//! Find group node
				XMLNODE* FindNextGroup (
					GROUP group,			//!< id of node to find
					const XMLNODE* node		//!< Previous sibling node for finding next
					) const;

				//! Find item node
				XMLNODE* FindItem (
					const XMLNODE* node = 0 //!< Parent node to search from.  Default is root.
					) const;

				//! Find item node
				XMLNODE* FindItem (
					const char* id,			//!< id of node to find
					const XMLNODE* node = 0 //!< Parent node to search from.  Default is root.
					) const;

				//! Find item node
				XMLNODE* FindItem (
					ITEM item,			//!< id of node to find
					const XMLNODE* node = 0 //!< Parent node to search from.  Default is root.
					) const;

				//! Find item node
				XMLNODE* FindNextItem (
					const XMLNODE* node		//!< Previous sibling node for finding next
					) const;

				//! Find item node
				XMLNODE* FindNextItem (
					const char* id,			//!< id of node to find
					const XMLNODE* node		//!< Previous sibling node for finding next
					) const;

				//! Find item node
				XMLNODE* FindNextItem (
					ITEM item,			//!< id of node to find
					const XMLNODE* node		//!< Previous sibling node for finding next
					) const;

				//! Find item node
				//! Will return item matching extents with highest level of detail
				XMLNODE* FindItem (
					const REGION2D& region, //!< Region to test
					const XMLNODE* node //!< Parent node to search from.  Default is root.
					) const;

				//! Find item node
				//! Will return item matching extents with highest level of detail
				XMLNODE* FindItem (
					const char* id,			//!< id of node to find
					const REGION2D& region, //!< Region to test
					const XMLNODE* node //!< Parent node to search from.  Default is root.
					) const;

				//! Find item node
				//! Will return item matching extents with highest level of detail
				XMLNODE* FindItem (
					ITEM item,			//!< id of node to find
					const REGION2D& region, //!< Region to test
					const XMLNODE* node //!< Parent node to search from.  Default is root.
					) const;

				//! Find item node, searching recursively through groups
				//! Will return item matching extents with highest level of detail
				XMLNODE* FindItem (
					ITEM item,			//!< id of node to find
					const REGION2D& region //!< Region to test
					) const;

				//! Return doc's root node
				XMLNODE* GetRootNode (
					) { return (m_doc.GetRootNode()); }

				XMLDOC& GetDoc (
					) { return (m_doc); }

				//! Get cell size of item
				static double GetCellSize (
					const XMLNODE* node //!< Node to get info on
					);

				//! Get max zoom level of item
				static INT32 GetMaxLevel (
					const XMLNODE* node //!< Node to get info on
					);

				//! Get URL of item
				static MISTRING GetURL (
					const XMLNODE* node	//!< Node to get info on
					);

				//! Get OBJITEM for item
				static void GetObjItem (
					RVC::OBJITEM& objitem,	//!< objitem returned
					const XMLNODE* node		//!< Node to get info on
					);

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				bool m_bIsRead;
				XMLDOC m_doc;
				bool TestNode(const XMLNODE* node, const REGION2D& region) const;
				void IMAGE::TILESET::LIST::FindItem(
					ITEM item,
					const REGION2D& region,
					XMLNODE*& retnode,
					const XMLNODE* parentnode
					) const;
				#endif
			};

		//! Container for representing position within tileset.
		class POSN {
			public:
				//! Default constructor.
				POSN ();
				//! Get tile index.
				const LPOINT2D& GetTileIdx (
					) const { return (m_TileIdx); }
				//! Get pixel position within tile.
				const LPOINT2D& GetPixelInTile (
					) const { return (m_PixelInTile); }
				//! Set position.
				void Set (
					const LPOINT2D& TileIdx,		//!< Tile index, typically absolute within tileset
					const LPOINT2D& PixelInTile	//!< Pixel position within tile
					);

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				LPOINT2D m_TileIdx;
				LPOINT2D m_PixelInTile;
				#endif
			};

		//! Range of positions within tileset.
		class RANGE {
			public:
				//! Default constructor.
				RANGE ();
				//! Compute number of X pixels for tile size.
				INT32 ComputeNumPixelsX (INT32 TileSize) const;
				//! Compute number of Y pixels for tile size.
				INT32 ComputeNumPixelsY (INT32 TileSize) const;
				//! Compute number of tiles in X direction.
				INT32 ComputeNumTilesX (
					) const { return (m_BottomRight.GetTileIdx().x - m_TopLeft.GetTileIdx().x + 1); }
				//! Compute number of tiles in Y direction.
				INT32 ComputeNumTilesY (
					) const { return (m_BottomRight.GetTileIdx().y - m_TopLeft.GetTileIdx().y + 1); }
				//! Determine if range contains specified tile index.
				bool Contains (
					const LPOINT2D& TileIdx
					) const;
				//! Get bottom-right position for range.
				const POSN& GetBottomRight (
					) const { return (m_BottomRight); }
				//! Get top-left position for range.
				const POSN& GetTopLeft (
					) const { return (m_TopLeft); }
				//! Set range.
				ERRVALUE Set (
					const POSN& TopLeft,
					const POSN& BottomRight
					);
			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				POSN m_TopLeft;
				POSN m_BottomRight;
				#endif
			};

		//! Default constructor, creates undefined tileset.
		TILESET ();

		//! Destructor.
		~TILESET ();

		//! Add additional tile format and optional extension.
		//! Used for tilesets that use multiple tile formats (ie mixed JPEG and PNG).
		void AddTileFormat (
			const IMAGE::FORMAT& format			//!< Image format
			);

		//! Adjust tileidx by offset and inversion
		void AdjustTileIdx (
			INT32 level,
			const LPOINT2D& BaseTileIdx,
			LPOINT2D& TileIdx
			) const;

		// Clear tileset, resetting all values.
		void Clear ();

		//! Close any zip files left open by FindTileFile
		void CloseZipFiles();

		//! Compute cell size for specified zoom level.
		//! Cell size in profile-specific CRS units returned.
		double ComputeCellSize (
			INT32 level,								//!< Zoom level (0-30)
			bool UseExtents	= false						//!< Use location of data to determine cellsize (for mercator)
			) const;

		//! Compute extents for entire level.
		ERRVALUE ComputeExtents (
			DRECT2D& Extents,							//!< Extents in tileset CRS returned
			INT32 level									//!< Zoom level (0-30)
			) const;

		//! Compute extents for single tile.
		ERRVALUE ComputeExtents (
			DRECT2D& Extents,							//!< Extents in tileset CRS returned
			const LPOINT3D& TileIdx					//!< Absolute tile index
			) const { return (ComputeExtents(Extents,TileIdx.z,TileIdx.x,TileIdx.y,TileIdx.x,TileIdx.y)); }

		//! Compute extents for single tile.
		ERRVALUE ComputeExtents (
			DRECT2D& Extents,							//!< Extents in tileset CRS returned
			INT32 level,								//!< Zoom level (0-30)
			INT32 TileColumn,							//!< Absolute tile column on overall grid
			INT32 TileRow								//!< Absolute tile row on overall grid
			) const { return (ComputeExtents(Extents,level,TileColumn,TileRow,TileColumn,TileRow)); }

		//! Compute extents for range of tiles.
		ERRVALUE ComputeExtents (
			DRECT2D& Extents,							//!< Extents in tileset CRS returned
			INT32 level,								//!< Zoom level (0-30)
			INT32 TileColumn1,						//!< Absolute left tile column on overall grid
			INT32 TileRow1,							//!< Absolute top tile row on overall grid
			INT32 TileColumn2,						//!< Absolute right tile column on overall grid
			INT32 TileRow2								//!< Absolute bottom tile row on overall grid
			) const;

		//! Compute georeference for specified zoom level relative to top-left tile.
		ERRVALUE ComputeGeoreference (
			INT32 level,								//!< Zoom level (0-30)
			GEOREFERENCE& Georeference				//!< Georeference returned
			) const;

		//! Compute georeference for specified zoom level relative to top-left tile of specified range.
		ERRVALUE ComputeGeoreference (
			INT32 level,							//!< Zoom level (0-30)
			const TILESET::RANGE& TileRange,		//!< Tile Range
			GEOREFERENCE& Georeference				//!< Georeference returned
			) const;

		//! Compute maximum zoom level to maintain all of source image detail.
		INT32 ComputeMaxLevel (
			const LPOINT2D& SrcImageSize,			//!< Source image size
			const GEOREFERENCE& SrcGeoreference	//!< Source image georeference
			) const;

		//! Compute zoom level where specified region is smaller than single tile.
		//! Note that actual result may have up to 4 tiles if region overlaps tile boundaries.
		INT32 ComputeMinLevel (
			const REGION2D& MapRegion,				//!< Region in any map coordinates (may differ from tileset CRS)
			INT32 MaxLevel = -1						//!< Maximum level, -1 (default) to use previously 
			) const;

		//! Compute tile range and padding values from extentand zoom level.
		ERRVALUE ComputeTileRange (
			const DRECT2D& Extents,					//!< Extents in tileset CRS
			INT32 level,								//!< Zoom level (0-30)
			TILESET::RANGE& TileRange,				//!< Absolute tile range returned
			bool IsOnBorder	= false					//!< True if extents boundary lies on borders between tiles
			) const;

		//! Compute tile range and padding values from geographic extent and zoom level.
		ERRVALUE ComputeTileRangeGeog (
			const DRECT2D& GeogExtents,			//!< Geographic extents in degrees
			INT32 level,								//!< Zoom level (0-30)
			TILESET::RANGE& TileRange,				//!< Absolute tile range returned
			bool IsOnBorder	= false					//!< True if extents boundary lies on borders between tiles
			) const;

		//! Create using profile and tile size.
		ERRVALUE Create (
			PROFILE profile,							//!< Profile, should not be Local
			INT32 TileSize								//!< Tile size
			);

		//! Create with local profile based on CRS, tile size and cell size.
		ERRVALUE Create (
			const SPATREF::COORDREFSYS& crs,		//!< CRS for local profile
			INT32 TileSize,							//!< Tile size
			double CellSize = 0						//!< Base level cell size, usually at maximum zoom level, 0 if will be set later
			);

		//! Create the auxiliary tileset files (html/xml/kml)
		ERRVALUE CreateAuxiliaryFiles (
			bool bOverwriteExisting = true,						//!< Write over any existing files
			TILESET::TARGET Target = TILESET::TARGET_Undefined,	//!< Target to create files for.  Default will determine target from tileset.
			INT32 MinLevel = -1									//!< Minimum zoom level to include
			);

		//! Create the auxiliary tileset files (html/xml/kml)
		ERRVALUE CreateAuxiliaryFiles (
			const MISTRING& BaseURL,							//!< Server location where html will be
			bool bOverwriteExisting = true,						//!< Write over any existing files
			TILESET::TARGET Target = TILESET::TARGET_Undefined,	//!< Target to create files for.  Default will determine target from tileset.
			INT32 MinLevel = -1									//!< Minimum zoom level to include
			);

		//! Create reduced resolution level
		ERRVALUE CreateReducedLevel (
			INT32 level,											//!< Zoom level to compute.
			IMGFORMAT ImgFormat = IMGFORMAT_Invalid,				//!< ImgFormat to use for output files.  Default will determine from tileset.
			UINT16 QualityJPEG = 75,								//!< Quality of JPEG output
			TILESET::TARGET Target = TILESET::TARGET_Undefined,		//!< Target to create files for.  Default will determine from tileset.	
			MISTRING message = TEXTID_Creating_ResolutionLevelN,	//!< Context message to display
			bool bZipOutput = false									//!< zip tileset
			);

		//! Create reduced resolution level
		ERRVALUE CreateReducedLevel (
			const FILEPATH& SourcePath,								//!< Location of source tiles
			const FILEPATH& TargetPath,								//!< Location to output tiles
			INT32 level,											//!< Zoom level to compute.
			IMGFORMAT ImgFormat = IMGFORMAT_Invalid,				//!< ImgFormat to use for output files.  Default will determine from tileset.
			UINT16 QualityJPEG = 75,								//!< Quality of JPEG output
			TILESET::TARGET Target = TILESET::TARGET_Undefined,		//!< Target to create files for.  Default will determine from tileset.
			MISTRING message = TEXTID_Creating_ResolutionLevelN,	//!< Context message to display
			bool bZipOutput = false									//!< zip tileset
			);

		//! Find KML file.
		//! @return False if not found.
		bool FindKMLFile (
			const FILEPATH& PathBaseFolder,			//!< Path to base folder
			INT32 level,							//!< Zoom level
			const LPOINT2D& TileIdx,				//!< Absolute tile index within defined tile range for zoom level
			FILEPATH& PathTileFile,					//!< Path to tile file returned
			MISTRING& TileName						//!< Name of tile found
			);

		//! Find tile file and format.
		//! @return Pointer to format for tile file if found, NULL if not found.
		const FORMAT* FindTileFile (
			const FILEPATH& PathBaseFolder,		//!< Path to base folder
			INT32 level,								//!< Zoom level
			const LPOINT2D& TileIdx,				//!< Absolute tile index within defined tile range for zoom level
			FILEPATH& PathTileFile					//!< Path to tile file returned
			) const;

		//! Find tile file and format.  Uses sourcepath read from TSD
		//! @return Pointer to format for tile file if found, NULL if not found.
		const FORMAT* FindTileFile (
			INT32 level,								//!< Zoom level
			const LPOINT2D& TileIdx,				//!< Absolute tile index within defined tile range for zoom level
			FILEPATH& PathTileFile					//!< Path to tile file returned
			);

		//! Find tile file and format.
		//! @return Pointer to format for tile file if found, NULL if not found.
		const FORMAT* FindTileFile (
			const FILEPATH& PathBaseFolder,			//!< Path to base folder
			INT32 level,							//!< Zoom level
			const LPOINT2D& TileIdx,				//!< Absolute tile index within defined tile range for zoom level
			FILEPATH& PathTileFile,					//!< Path to tile file returned
			MISTRING& TileName						//!< Name of tile found
			);

		//! Find tile file and format.  Uses sourcepath read from TSD
		//! @return Pointer to format for tile file if found, NULL if not found.
		const FORMAT* FindTileFile (
			INT32 level,							//!< Zoom level
			const LPOINT2D& TileIdx,				//!< Absolute tile index within defined tile range for zoom level
			FILEPATH& PathTileFile,					//!< Path to tile file returned
			MISTRING& TileName						//!< Name of tile found
			);

		//! Generate path for tile file.
		ERRVALUE GenerateTileFilePath (
			const FILEPATH& PathBaseFolder,			//!< Path to base folder
			INT32 level,							//!< Zoom level
			const LPOINT2D& TileIdx,				//!< Absolute tile index within defined tile range for zoom level
			FILEPATH& PathTileFile,					//!< Path to tile file returned
			const FORMAT *pFormat = 0,				//!< Optional format, if 0 will use first one set for tileset
			bool MakeFolder = false					//!< Create folder if does not exist
			) const;
		
		//! Generate pixel range map
		ERRVALUE GeneratePixelRangeMap (
			INT32 level = -1
			);

		//! Generate region tile map
		ERRVALUE GenerateRegionTileMap (
			);

		//! Generate tile map
		ERRVALUE GenerateTileMap (
			INT32 level = -1
			);

		//! Make tileset test table
		ERRVALUE GenerateTilesetTest ();

		//! Scan tileset test table
		ERRVALUE ScanTilesetTest ();

		//! Generate xml snippet for terrain tileset
		ERRVALUE GenerateTerrainXML (
			);

		//! Get base cell size for "local" profile.
		//! @return Base cell size specified in Create(), undefined if not using local profile.
		double GetBaseCellSize (
			) const { return (m_BaseCellSize); }

		//! Get level corresponding to base cell size for "local" profile.
		//! @return Base level, undefined if not using local profile.
		INT32 GetBaseLevel (
			) const { return (m_BaseLevel); }

		//! Get number of tiles at base level in y direction.
		INT32 GetBaseTileCount (
			) const { return (m_BaseTileCount); }

		//! Get display string for cell size.
		MISTRING GetCellSizeDisplayStr (
			INT32 level									//!< Zoom level (0-30)
			) const;

		//! Get copyright
		const MISTRING& GetCopyright (
			) const { return (m_copyright); }

		//! Get Coordinate Reference System.
		const SPATREF::COORDREFSYS& GetCRS (
			) const { return (m_crs); }

		//! Get node containing list of dates to use
		XMLNODE* GetDateList () const { return (m_DateList); }

		//! Get node containing date range to use
		XMLNODE* GetDateRange () const { return (m_DateRange); }

		//! Get extents in tileset CRS.
		const DRECT2D& GetExtentsInCRS (
			) const { return (m_ExtentsInCRS); }

		//! Get location to cache if defined in TSD
		void GetCacheLocationPath (
			FILEPATH& CacheLocationPath
			);

		//! Determine format from tileset
		void GetImgFormat (
			IMGFORMAT& ImgFormat
			);

		//! Determine format from tileset
		void GetImgFormat (
			IMGFORMAT& ImgFormat,
			PIXEL& PixelProperties
			);

		//! Get location string for local tileset, TSD must have been read.
		//! Will return level specific location if one exists.
		//! If using a list of locations, will iterate through list and return next location
		MISTRING GetLocalLocationString (
			INT32 level = -1			//!< Optional level to get location for
			);

		//! Get list of all local location strings, TSD must have been read.
		const MISTRINGLIST& GetLocalLocationStrings () const {return m_LocalLocationStrings;}

		//! Get list of all level specific local location strings, TSD must have been read.
		//! List size is number of levels, starts at minlevel.  Contains empty string if no level specific location
		const MISTRINGLIST& GetLocalLocationLevelStrings () const {return m_LocalLocationLevelStrings;}

		//! Get location path for local tileset, TSD must have been read
		//! Will return level specific location if one exists.
		//! If using a list of locations, will iterate through list and return next location
		void GetLocalLocationPath (
			FILEPATH& LocationPath,
			INT32 level = -1			//!< Optional level to get location for
			);

		//! Get margin (svg tileset)
		double GetMargin (
			) const { return m_Margin; }

		//! Get location string for online tileset, TSD must have been read
		//! Will return level specific location if one exists.
		//! If using a list of locations, will iterate through list and return next location
		MISTRING GetRemoteLocationString (
			INT32 level = -1			//!< Optional level to get location for
			);

		//! Get list of all remote location strings, TSD must have been read.
		const MISTRINGLIST& GetRemoteLocationStrings () const {return m_RemoteLocationStrings;}

		//! Get list of all level specific remote location strings, TSD must have been read.
		//! List size is number of levels, starts at minlevel.  Contains empty string if no level specific location
		const MISTRINGLIST& GetRemoteLocationLevelStrings () const {return m_RemoteLocationLevelStrings;}

		//! Get maximum zoom level defined.
		INT32 GetMaxLevel (
			) const { return (m_MaxLevel); }

		//! Get minimum zoom level defined.
		INT32 GetMinLevel (
			) const { return (m_MinLevel); }

		//! Get metadata
		const MISTRING& GetMetadata (
			) const { return (m_metadata); }

		//! Get objitem of tsd
		const RVC::OBJITEM& GetObjItem() {return m_objitem;}

		//! Get path of TSD if it has been read or written
		void GetPath (
			FILEPATH& path
			) { path = m_TSDPath; }

		double GetPixelValueMin (
			) const { return (m_PixelValueMin); }

		double GetPixelValueMax (
			) const { return (m_PixelValueMax); }

		//! Get profile.
		PROFILE GetProfile (
			) const { return (m_profile); }

		//! Get details about tileset as string
		void GetProperties (
			MISTRING& properties,
			bool ShowExtents = true
			);

		//! Get reference location
		const MISTRING& GetReferenceLocation (
			) const { return (m_Reference); }

		//! Get scale to meters
		double GetScale (
			) const { return (m_Scale); }

		INT32 GetScaleNumerator (
			) const { return (m_ScaleNumerator); }

		INT32 GetScaleDenominator (
			) const { return (m_ScaleDenominator); }

		//! Get prescale offset
		INT64 GetPrescaleOffset (
			) const { return (m_PrescaleOffset); }

		//! Get postscale offset
		INT64 GetPostscaleOffset (
			) const { return (m_PostscaleOffset); }

		//! Get structure.
		STRUCTURE GetStructure ();

		//! Get target tileset.
		TARGET GetTarget ();

		//! Get type
		TYPE GetType () {return m_type;}

		//! Get preferred tile format.
		const FORMAT& GetTileFormat (
			bool HasNullCells
			) const;

		//! Get list of tile formats used.
		const MILIST<FORMAT>& GetTileFormatList (
			) const { return (m_TileFormatList); }

		//! Get tile range at specified zoom level.
		ERRVALUE GetTileRange (
			INT32 level,								//!< Zoom level, must be in previously specified range
			TILESET::RANGE& TileRange				//!< Tile range returned
			) const;

		//! Get tile size.
		INT32 GetTileSize (
			) const { return (m_TileSize); }

		//! Get title;
		const MISTRING& GetTitle (
			) const { return (m_title); }

		//! Get top left tile for specified zoom level.
		ERRVALUE GetTopLeftTile (
			INT32 level,							//!< Zoom level, must be in previously specified range
			LPOINT2D& TileIdx						//!< Top left tile index returned
			) const;

		//! Get list of transparent colors for tileset
		void GetTransparentColorList(MILIST<TransparentColor>& TransparentColors) {TransparentColors = m_TransparentColorsList;}

		//! Return time when tsd was last modified
		const DATETIME& GetTSDLastMod () {return m_TSDLastMod;}

		//! Determine if TSD contains date to use in location
		//! 0 if no date
		//! 1 if date list
		//! 2 if date range or [d] specified in location
		INT32 HasDate () const;

		//! Determine if specified tile has at least 1 source tile from next level down
		bool HasSourceTiles (
			const FILEPATH& PathBaseFolder,
			INT32 level,
			const LPOINT2D& TileIdx
			);

		//! Determine if specified tile has at least 1 source tile from next level down.  Uses sourcepath read from TSD
		bool HasSourceTiles (
			INT32 level,
			const LPOINT2D& TileIdx
			);

		//! Determine if specified row has any source tiles from next level down
		bool HasSourceTiles (
			const FILEPATH& PathBaseFolder,
			INT32 level,
			INT32 row
			);

		//! Determine if specified row has any source tiles from next level down.  Uses sourcepath read from TSD
		bool HasSourceTiles (
			INT32 level,
			INT32 row
			);

		//! Determine if specified tile exists
		bool HasTile (
			const FILEPATH& PathBaseFolder,
			INT32 level,
			const LPOINT2D& TileIdx
			);

		//! Determine if specified tile exists.  Uses sourcepath read from TSD
		bool HasTile (
			INT32 level,
			const LPOINT2D& TileIdx
			);

		//! Determine if specified row has any tiles
		bool HasTiles (
			const FILEPATH& PathBaseFolder,
			INT32 level,
			INT32 row,
			bool bClear = false
			);

		//! Determine if specified row has any tiles.  Uses sourcepath read from TSD
		bool HasTiles (
			INT32 level,
			INT32 row,
			bool bClear = false
			);

		//! Read and parse TSD from objitem
		ERRVALUE ReadTSD (
			const RVC::OBJITEM& objitem
			);

		//! Read and parse TSD from filepath
		ERRVALUE ReadTSD (
			const FILEPATH& filepath
			);

		//! Read and parse TSD from URL
		ERRVALUE ReadTSD (
			const MISTRING& URL
			);

		//! Read and parse TSD from xml
		ERRVALUE ReadTSD (
			const XMLDOC& doc
			);

		//! Set base cell size for local profile.
		void SetBaseCellSize (
			double BaseCellSize
			) { m_BaseCellSize = BaseCellSize; }

		//! Set number of tiles at base level in y direction.
		//! Default is 1.  For case of geographic profile, default would be 1x2 tiles at level 0.  Mercator would be 1x1.
		//! If TileCount was 5, then number of tiles would be 5x10 for geographic and 5x5 for mercator.
		void SetBaseTileCount (
			INT32 TileCount
			) { m_BaseTileCount = TileCount; }

		//! Set copyright.
		void SetCopyright (
			const MISTRING& copyright
			) { m_copyright = copyright; }

		//! Set spatial and zoom level extents.
		ERRVALUE SetExtents (
			const DRECT2D& ExtentsInCRS,			//!< Extents in tileset CRS at maximum zoom level
			INT32 MinLevel,							//!< Minimum zoom level (0-30)
			INT32 MaxLevel,								//!< Maximum zoom level, should not exceed base level if local profile
			bool IsOnBorder	= false					//!< True if extents boundary lies on borders between tiles
			);

		//! Set margin (for svg tileset)
		void SetMargin (
			double margin
			) { m_Margin = margin; }

		//! Set tile range for max level.  Will compute range for all levels.
		ERRVALUE SetMaxTileRange (
			const TILESET::RANGE& TileRange		//!< Tile Range at max level
			);

		//! Set metadata.
		void SetMetadata (
			const MISTRING& metadata
			) { m_metadata = metadata; }

		//! Set path of TSD
		void SetPath (
			const FILEPATH& path
			) { m_TSDPath = path; }

		//! Set pixel range values to be written to ttsd
		void SetPixelValueRange(
			double PixelValueMin,
			double PixelValueMax
			) { m_PixelValueMin = PixelValueMin;  m_PixelValueMax = PixelValueMax; }

		//! Manually set list of urls.  Used in bing maps.
		void SetRemoteLocationStrings (const MISTRINGLIST& RemoteLocationStrings) {m_RemoteLocationStrings = RemoteLocationStrings;}

		//! Set scale to meters
		void SetScale (
			double Scale
			) { m_Scale = Scale; }

		void SetScaleRational (
			INT32 Numerator,
			INT32 Denominator
			) { m_ScaleNumerator = Numerator; m_ScaleDenominator = Denominator; m_Scale = (double)Numerator/(double)Denominator; }

		//! Set prescale offset
		void SetPrescaleOffset (
			INT64 offset
			) { m_PrescaleOffset = offset; }

		//! Set Postscale offset
		void SetPostscaleOffset (
			INT64 offset
			) { m_PostscaleOffset = offset; }

		//! Set tile format and optional extension.
		//! Will clear all previous tile formats.
		void SetTileFormat (
			const IMAGE::FORMAT& format			//!< Image format
			);

		//! Set tile size.
		void SetTileSize (
			INT32 TileSize
			) { m_TileSize = TileSize; }

		//! Set title.
		void SetTitle (
			const MISTRING& title
			) { m_title = title; }

		//! Set type
		void SetType (
			TYPE type
			) { m_type = type; }

		//! Set URL of TSD
		void SetURL (
			const MISTRING& URL
			) { m_TSDURL = URL; }

		//! Write KML file for a Google Earth tile
		ERRVALUE WriteKML (
			const FILEPATH& PathBaseFolder,			//!< Path to base folder
			INT32 level,							//!< Zoom level
			const LPOINT2D& TileIdx					//!< Absolute tile index within defined tile range for zoom level
			);

		//! Write descriptive file in TSD format.
		ERRVALUE WriteTSD (
			const FILEPATH& filepath,				//!< TSD path
			const FILEPATH& PathBaseFolder,			//!< Location of tile folder
			INT32 MinLevel = -1						//!< Minimum zoom level to include
			);

		//! Write descriptive file in TSD format.
		ERRVALUE WriteTSD (
			const FILEPATH& filepath,				//!< TSD path
			const MISTRINGLIST& LocationList,		//!< Location list of tile folders
			INT32 MinLevel = -1						//!< Minimum zoom level to include
			);

		//! Write descriptive file in TSD format.
		ERRVALUE WriteTSD (
			const FILEPATH& filepath,				//!< TSD path
			const FILEPATH& PathBaseFolder,			//!< Location of tile folder
			const MISTRINGLIST& LocalLevelLocationList,	//!< Location list of level specific tile folders
			const MISTRINGLIST& RemoteLevelLocationList,//!< Location list of level specific tile folders
			INT32 MinLevel = -1						//!< Minimum zoom level to include
			);

		//! Write descriptive file in TSD format.
		ERRVALUE WriteTSD (
			const FILEPATH& filepath,				//!< TSD path
			const MISTRINGLIST& LocationList,		//!< Location list of tile folders
			const MISTRINGLIST& LocalLevelLocationList,	//!< Location list of level specific tile folders
			const MISTRINGLIST& RemoteLevelLocationList,//!< Location list of level specific tile folders
			INT32 MinLevel = -1						//!< Minimum zoom level to include
			);

		//! Write descriptive file in TSD format.  Uses same locations set by ReadTSD
		ERRVALUE WriteTSD (
			const FILEPATH& filepath,				//!< TSD path
			INT32 MinLevel = -1						//!< Minimum zoom level to include
			);

		//! Unzip tileset
		ERRVALUE Unzip (
			const FILEPATH& TileFolderBasePath,		//!< Location of tile folder
			bool bDeleteAfterUnzip = false			//!< Delete zip files after unzipping
			);

		//! Unzip tileset to location
		ERRVALUE Unzip (
			const FILEPATH& TileFolderBasePath,		//!< Location of tile folder
			const FILEPATH& TargetPath,				//!< Target Location
			bool bDeleteAfterUnzip = false			//!< Delete zip files after unzipping
			);

		//! Zip tileset
		ERRVALUE Zip (
			const FILEPATH& TileFolderBasePath,		//!< Location of tile folder
			bool bDeleteAfterZip = false			//!< Delete source tiles after zipping
			);

		//! Zip tileset
		ERRVALUE Zip (
			const FILEPATH& TileFolderBasePath,		//!< Location of tile folder
			const FILEPATH& TargetPath,				//!< Target Location
			bool bDeleteAfterZip = false			//!< Delete source tiles after zipping
			);

		//! Zip contents of zoom level
		ERRVALUE ZipLevel (
			const FILEPATH& TileFolderBasePath,		//!< Location of tile folder
			INT32 level,							//!< Zoom level to zip
			bool bDeleteAfterZip = false			//!< Delete source tiles after zipping
			);

		//! Zip contents of zoom level
		ERRVALUE ZipLevel (
			const FILEPATH& TileFolderBasePath,		//!< Location of tile folder
			const FILEPATH& TargetPath,				//!< Target Location
			INT32 level,							//!< Zoom level to zip
			bool bDeleteAfterZip = false,			//!< Delete source tiles after zipping
			MITHREAD::QUEUE* Queue = NULL			//!< Thread queue for zipping
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		RVC::OBJITEM m_objitem;
		PROFILE m_profile;
		TYPE m_type;
		INT32 m_TileSize;
		SPATREF::COORDREFSYS m_crs;
		double m_BaseCellSize;						// Base cell size used in local profile
		INT32 m_BaseLevel;							// Level corresponding to base cell size for local profile
		INT32 m_BaseTileCount;
		INT32 m_MaxLevel;
		INT32 m_MinLevel;
		bool m_bInvertX;
		bool m_bInvertY;
		INT32 m_OffsetX;
		INT32 m_OffsetY;
		INT32 m_BitsetTargetLevel;
		INT32 m_BitsetSourceLevel;
		INT32 m_BitsetTargetRow;
		INT32 m_BitsetSourceRow;
		INT32 m_LastLocationIndex;
		double m_PixelValueMin;
		double m_PixelValueMax;
		double m_Scale;
		INT32 m_ScaleNumerator;
		INT32 m_ScaleDenominator;
		INT64 m_PrescaleOffset;
		INT64 m_PostscaleOffset;
		MISTRING m_title;
		MISTRING m_copyright;
		MISTRING m_metadata;
		MISTRING m_Reference;
		DRECT2D m_ExtentsInCRS;
		DRECT2D m_GeogExtents;
		XMLNODE* m_DateList;
		XMLNODE* m_DateRange;
		XMLDOC m_DateDoc;
		MILIST<TransparentColor> m_TransparentColorsList;
		FILEPATH m_TSDPath;
		MISTRING m_TSDURL;
		DATETIME m_TSDLastMod;

		MISTATUSCONTEXT *m_Context;
		INT32 m_JobCount;

		MISTRINGLIST m_RemoteLocationStrings;
		MISTRINGLIST m_LocalLocationStrings;
		MISTRINGLIST m_RemoteLocationLevelStrings;
		MISTRINGLIST m_LocalLocationLevelStrings;
		MISTRING m_CacheLocationString;

		MILIST<IMAGE::FORMAT> m_TileFormatList;
		TILESET::RANGE *m_TileRangeArray;
		BITSET m_TargetTileExistsBitset;
		BITSET m_SourceTilesExistBitset;

		ZIPFILE m_LastZipFile;
		FILEPATH m_LastZipFilePath;

		double m_Margin; //currently only used for svgs a margin outside of which the data may render inspite of tile size (required for html rendering)
		
		ERRVALUE ComputeTileValues (
			const DPOINT2D& pt,
			INT32 level,
			TILESET::POSN& TilePosn,
			bool IsBottomRight,
			bool IsOnBorder
			) const;

		ERRVALUE ComputeTileValuesGeog (
			const DPOINT2D& ptGeog,
			INT32 level,
			TILESET::POSN& TilePosn,
			bool IsBottomRight,
			bool IsOnBorder
			) const;

		ERRVALUE CreateTilesetMap (
			const RVC::OBJITEM& objitem,
			INT32 level,
			const TRANS2D_AFFINE& SrcTransAffine,
			const MISTRING& name,
			const MISTRING& description,
			const IMGFORMAT& ImgFormat
			);

		bool FilterDelegate16ToRGBA (
			const SPATMOD::IMAGE::SAMPLEITERATOR *ittgt,
			const SPATMOD::IMAGE::SAMPLEITERATOR *itsrc,
			bool valid
			);

		MISTRING GetRelativePathString (
			const FILEPATH& TSDPath,		//!< Location of TSD file
			const FILEPATH& PathBaseFolder	//!< Location of Tile folder
			) const;

		MISTRING GetNextLocationString (
			const MISTRINGLIST& LocationList
			);

		MISTRING GetLocationString (
			INT32 level,
			const MISTRINGLIST& LocationStrings,
			const MISTRINGLIST& LocationLevelStrings
			);

		void OnJobDone (MITHREAD::ID id);
		void SetupTargetBitsets (const FILEPATH& PathBaseFolder, INT32 level, INT32 row, bool bClearRow = false);
		void SetupSourcesBitsets (const FILEPATH& PathBaseFolder, INT32 level, INT32 row);
		MISTRING GetPathStructureComment (const MISTRING& BasePath) const;
		ERRVALUE WriteTSDBase (MISTRING& string);
		ERRVALUE WriteTSDZoomLevels (MISTRING& string, INT32 MinLevel);
		ERRVALUE WriteTSDZoomLevels (MISTRING& string, INT32 MinLevel, const MISTRINGLIST& LocalLevelLocationList, const MISTRINGLIST& RemoteLevelLocationList);
		ERRVALUE WriteTSDLocations (MISTRING& string, const FILEPATH& filepath, const FILEPATH& PathBaseFolder) const;
		ERRVALUE WriteTSDLocations (MISTRING& string, const MISTRINGLIST& LocationList) const;

		#endif // GENERATING_DOXYGEN_OUTPUT

	};

#ifndef	GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(TILESET::TYPE)
#endif

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGETILESET_H
